// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include <iostream>
#include "Eth_IPv4.h"
#include "DHCP_ARP.h"
#include "NPRV_L1_L2_radio.h"
#include <math.h>
#include <SFML/System.hpp>
#include "NPRV_SDR_if.h"
#include "NPRV_sat_emul.h"

#ifdef NPRV_windows
#include "Win_tuntap_if.h"
#pragma warning(disable: 4996)
#endif

#ifdef NPRV_linux
#include <stdio.h>
#include <string.h>
#include "Linux_tuntap_if.h"
#endif

sf::Mutex mutex_tuntap_TX;
sf::Mutex mutex_TX_IPv4_struct;

typedef struct {
	unsigned char data[1600];
	int size;
	unsigned int dest_ID;
	unsigned char* data_remain;
	int seg_counter;
} TX_IP_packet;

static TX_IP_packet TX_IP_buffers[1000];

static unsigned char temp_Eth_buffer[1600] = { 0 };//temporary buffer waiting for ARP entry
static uint32_t temp_Eth_buff_IP;
static int temp_Eth_buff_size = 0;

static unsigned int IP_ident = 01; 

const uint8_t CONF_virtual_MAC_SPY_RX[8] = { 0x4E, 0x50, 0x52, 0x57, 0xA3, 0xB1 };

FILE* NPRV_file_debug_IPv4;

uint32_t IP_char2int(unsigned char* IP_char) {
	uint32_t result;
	result = 0x1000000 * IP_char[0] + 0x10000 * IP_char[1] + 0x100 * IP_char[2] + IP_char[3];
	return result;
}

void IP_int2char(uint32_t IP_int, unsigned char* IP_char) {
	IP_char[0] = (IP_int & 0xFF000000) >> 24;
	IP_char[1] = (IP_int & 0x00FF0000) >> 16;
	IP_char[2] = (IP_int & 0x0000FF00) >> 8;
	IP_char[3] = (IP_int & 0x000000FF);
}

void IP_int2txt(uint32_t IP_int, char* IP_char) {
	unsigned char IP_temp[20];
	IP_int2char(IP_int, IP_temp);
	sprintf(IP_char, "%u.%u.%u.%u", IP_temp[0], IP_temp[1], IP_temp[2], IP_temp[3]);
}


// ********* TX data structure and interaction ****************

static TX_IP_packet* TX_IP_empty_slots[1000];
static int indx_IP_empty_slots = 999;
static int TX_IPv4_buffer_filling = 0;
static TX_IP_packet* TX_IP_pkt_FIFOs[NPRV_MAX_CLIENT_NB][100];
static unsigned int TX_IP_FIFO_RD_point[NPRV_MAX_CLIENT_NB];
static unsigned int TX_IP_FIFO_WR_point[NPRV_MAX_CLIENT_NB];

//#define NPRV_IPv4_BUFF_DEBUG

// just checks if TX data are available
void NPRV_IPv4_init_TX_struct(void) {
	int i;
	for (i = 0; i < 1000; i++) {
		TX_IP_empty_slots[i] = &TX_IP_buffers[i];
	}
	for (i = 0; i < NPRV_MAX_CLIENT_NB; i++) {
		TX_IP_FIFO_RD_point[i] = 0;
		TX_IP_FIFO_WR_point[i] = 0;
	}
	TX_IPv4_buffer_filling = 0;
	indx_IP_empty_slots = 999;
#ifdef NPRV_IPv4_BUFF_DEBUG
	NPRV_file_debug_IPv4 = fopen("user_data/NPRV_file_debug_IPv4.txt", "w");
#endif
}

int NPRV_IPv4_push_TX_pckt(unsigned char* in_pckt, int size, unsigned int dest_ID) {
	TX_IP_packet* loc_pckt;
	int FIFO_client_filling;
	FIFO_client_filling = TX_IP_FIFO_WR_point[dest_ID] - TX_IP_FIFO_RD_point[dest_ID];
	//printf("FIFO client %i filling %i\r\n", dest_ID, FIFO_client_filling);
	if ( (indx_IP_empty_slots > 10) && (FIFO_client_filling < 97) ){ // room available //0
		mutex_TX_IPv4_struct.lock();
		loc_pckt = TX_IP_empty_slots[indx_IP_empty_slots];
		indx_IP_empty_slots--;
		loc_pckt->data[0] = 0;//for future use dest_MAC_IP
		loc_pckt->data[1] = 0;
		memcpy((loc_pckt->data)+2, in_pckt, size);
		loc_pckt->data_remain = loc_pckt->data;
		loc_pckt->size = size + (size % 2) + 2;//round to upper pair value ; +2 because of dest_MAC_IP
		loc_pckt->dest_ID = dest_ID;
		loc_pckt->seg_counter = 0;
		TX_IP_pkt_FIFOs[dest_ID][TX_IP_FIFO_WR_point[dest_ID] % 100] = loc_pckt;
		TX_IP_FIFO_WR_point[dest_ID]++;
		TX_IPv4_buffer_filling = TX_IPv4_buffer_filling + loc_pckt->size;
#ifdef NPRV_IPv4_BUFF_DEBUG
		fprintf(NPRV_file_debug_IPv4, "RX:%i Now:%i\r\n", size, TX_IPv4_buffer_filling);
#endif
		mutex_TX_IPv4_struct.unlock();
		return 1;
	}
	else {
		return 0;
	}
}

int NPRV_IPv4_get_TX_filling(void) {
	int answer;
	//printf("IPv4 filling: frames:%d bytes:%d\r\n", 999 - indx_IP_empty_slots, TX_IPv4_buffer_filling);
	answer = TX_IPv4_buffer_filling;
	return answer;
}

int NPRV_IPv4_TX_data_available(void) {
	int answer;
	if (indx_IP_empty_slots < 999) { answer = 1; }
	else { answer = 0; }
	if ((CONF_is_NPRV_HUB == 0) && (G_NPRV_MyClientID >= NPRV_MAX_CLIENT_NB)) { answer = 0; }
	return answer;
}

//prepare the frame, except the Header
int NPRV_IPv4_get_TXr_frame(unsigned char* tx_frame, int avail_size) {
	static unsigned char packet_counter[NPRV_MAX_CLIENT_NB] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	static unsigned int client_round = 0;
	unsigned int dest_ID;
	int size_frame = 0;
	int size_IP_data;
	int client_count = 0;
	unsigned int data_available;
	int needed_size;
	TX_IP_packet* loc_pckt;

	mutex_TX_IPv4_struct.lock();
	do {
		client_count++;
		client_round = (client_round + 1) % NPRV_MAX_CLIENT_NB;
		data_available = (TX_IP_FIFO_WR_point[client_round] - TX_IP_FIFO_RD_point[client_round]) & 0xFFFF;
	} while ((data_available == 0) && (client_count < NPRV_MAX_CLIENT_NB));
	if (data_available > 0) {
		//printf("IPv4 R Frame TX");
		avail_size = avail_size - (avail_size % 2);
		dest_ID = client_round;
		loc_pckt = TX_IP_pkt_FIFOs[client_round][TX_IP_FIFO_RD_point[client_round] % 100];
		if (loc_pckt->seg_counter > 0) {//following data
			tx_frame[0] = (dest_ID & 0x1F00) >> 8;//dest ID
			tx_frame[1] = dest_ID & 0xFF;
			tx_frame[2] = (0x02 << 4) /*protocol*/
				+ (packet_counter[client_round] & 0x0F);//protocol IPv4 & pckt counter
			needed_size = loc_pckt->size + 6;
			if (needed_size > avail_size) {
				tx_frame[3] = loc_pckt->seg_counter & 0x7F;//last segment & seg counter
				size_IP_data = avail_size - 6;
				memcpy(tx_frame + 4, loc_pckt->data_remain, size_IP_data);
				loc_pckt->seg_counter++;
				loc_pckt->size = loc_pckt->size - size_IP_data;
				loc_pckt->data_remain = loc_pckt->data_remain + size_IP_data;//pointer
				TX_IPv4_buffer_filling = TX_IPv4_buffer_filling - size_IP_data;
#ifdef NPRV_IPv4_BUFF_DEBUG
				fprintf(NPRV_file_debug_IPv4, "TX follow:%i Now:%i\r\n", size_IP_data, TX_IPv4_buffer_filling);
#endif
			}
			else {
				tx_frame[3] = 0x80 + (loc_pckt->seg_counter & 0x7F);//last segment & seg counter
				size_IP_data = loc_pckt->size;
				memcpy(tx_frame + 4, loc_pckt->data_remain, size_IP_data);
				TX_IP_FIFO_RD_point[dest_ID]++;
				indx_IP_empty_slots++;
				TX_IP_empty_slots[indx_IP_empty_slots] = loc_pckt;
				TX_IPv4_buffer_filling = TX_IPv4_buffer_filling - size_IP_data;
#ifdef NPRV_IPv4_BUFF_DEBUG
				fprintf(NPRV_file_debug_IPv4, "TX    last:%i Now:%i\r\n", size_IP_data, TX_IPv4_buffer_filling);
#endif
			}
		}
		else {//initial data
			packet_counter[client_round]++;
			tx_frame[0] = (dest_ID & 0x1F00) >> 8;//dest ID
			tx_frame[1] = dest_ID & 0xFF;
			tx_frame[2] = (0x02 << 4) /*protocol*/
				+ (packet_counter[client_round] & 0x0F);//protocol IPv4 & pckt counter
			
			needed_size = loc_pckt->size + 6;//6 = 4 (for header inside) + 2 (checksum)
			if (needed_size > avail_size) {//need to split in several segments
				tx_frame[3] = 0x00; // Last seg & seg_counter
				size_IP_data = avail_size - 6; // 6 = 4 (for header inside) + 2 (checksum)
				memcpy(tx_frame + 4, loc_pckt->data, size_IP_data);
				loc_pckt->seg_counter++;
				loc_pckt->size = loc_pckt->size - size_IP_data;
				loc_pckt->data_remain = loc_pckt->data_remain + size_IP_data;//pointer
				TX_IPv4_buffer_filling = TX_IPv4_buffer_filling - size_IP_data;
#ifdef NPRV_IPv4_BUFF_DEBUG
				fprintf(NPRV_file_debug_IPv4, "TX   first:%i Now:%i\r\n", size_IP_data, TX_IPv4_buffer_filling);
#endif
			}
			else {//ok to send all packet
				tx_frame[3] = 0x80; // Last seg & seg_counter
				size_IP_data = loc_pckt->size;
				memcpy(tx_frame + 4, loc_pckt->data, size_IP_data);
				TX_IP_FIFO_RD_point[dest_ID]++;
				indx_IP_empty_slots++;
				TX_IP_empty_slots[indx_IP_empty_slots] = loc_pckt;
				TX_IPv4_buffer_filling = TX_IPv4_buffer_filling - size_IP_data;
#ifdef NPRV_IPv4_BUFF_DEBUG
				fprintf(NPRV_file_debug_IPv4, "TX   single:%i Now:%i\r\n", size_IP_data, TX_IPv4_buffer_filling);
#endif
			}
			
			
		}
		size_frame = (size_IP_data + 4);//for internal header
		size_frame = size_frame + (size_frame % 2);
		NPRV_code_checksum(tx_frame, size_frame);
		size_frame = size_frame + 2;//for checksum
		// Radio frame construction (header is built inside "decide_new_TXframe")
		mutex_TX_IPv4_struct.unlock();
		//printf("pkt:%u seg:%u\r\n", (tx_frame[2] & 0x0F), (tx_frame[3] & 0x7F));
		return size_frame;
	}
	else { 
		mutex_TX_IPv4_struct.unlock();
		return 0;
	}
}

#define NPRV_IPV4_RX_SLOT_CNT 128
static int pkt_reconstr_occupied[NPRV_IPV4_RX_SLOT_CNT] = { 0 };
static unsigned int pkt_reconstr_TXID[NPRV_IPV4_RX_SLOT_CNT] = { 0 };
static unsigned int pkt_reconstr_RXID[NPRV_IPV4_RX_SLOT_CNT] = { 0 };
static uint32_t pkt_reconstr_Time[NPRV_IPV4_RX_SLOT_CNT] = { 0 };
static unsigned char pkt_reconstr_PktCounter[NPRV_IPV4_RX_SLOT_CNT] = { 0 };

static int pkt_reconstr_SegSize[NPRV_IPV4_RX_SLOT_CNT][128] = { 0 };//[slots][seg_nb]
static int pkt_reconstr_SegIsLast[NPRV_IPV4_RX_SLOT_CNT][128] = { 0 };//[slots][seg_nb]
static unsigned char pkt_reconstr_SegData[NPRV_IPV4_RX_SLOT_CNT][128][200] = { 0 };//[slots][seg_nb]

void NPR_IPv4_init_RX_reconstr(void) {
	int i,j;
	for (i = 0; i < NPRV_IPV4_RX_SLOT_CNT; i++) {
		pkt_reconstr_occupied[i] = 0;
		pkt_reconstr_Time[i] = 0;
		for (j = 0; j < 128; j++) {
			pkt_reconstr_SegSize[i][j] = 0;
			pkt_reconstr_SegIsLast[i][j] = 0;
		}
	}
}

void NPR_IPv4_RX_free_reconstr_slot(int slot) {
	int i;
	pkt_reconstr_occupied[slot] = 0;
	for (i = 0; i < NPRV_IPV4_RX_SLOT_CNT; i++) {
		pkt_reconstr_SegSize[slot][i] = 0;
	}
}

int NPRV_IPv4_find_empty_reconstr_slot(uint32_t time) {
	int i, j;
	int found = 0xFFFF;
	int delta_T;
	for (i = 0; i < NPRV_IPV4_RX_SLOT_CNT; i++) {
		delta_T = time - pkt_reconstr_Time[i];
		delta_T = abs(delta_T);
		if ((pkt_reconstr_occupied[i] == 1) && (delta_T > 5800000)) {//620000
			NPR_IPv4_RX_free_reconstr_slot(i);
			//printf("FREE IPv4 slot\r\n");
			//NPRV_HMI_new_TXT_message();
		}
		if (pkt_reconstr_occupied[i] == 0) {
			found = i;
		}
	}
	return found;
}

void NPR_IPv4_RX_attempt_reconstr_pkt(int slot) {
	static unsigned char IPv4_RX_buffer[1600] = { 0 };
	int seg_count = 0;
	int end_pkt = 0;
	int missing_segment = 0;
	int loc_size = 0;
	int i;
	int seg_size;
	//checking if IPv4 frame is complete, with all segments
	while ((end_pkt == 0) && (seg_count < 126) && (missing_segment == 0)) {
		seg_size = pkt_reconstr_SegSize[slot][seg_count];
		if (seg_size == 0) {
			missing_segment = 1;
		} else {
			loc_size = loc_size + seg_size;
			end_pkt = pkt_reconstr_SegIsLast[slot][seg_count];
		}
		seg_count++;
	}
	// if complete, reconstruct the frame
	if ((end_pkt == 1) && (missing_segment == 0) && (loc_size > 32) && (loc_size < 1600)) {
		loc_size = 0;
		for (i = 0; i < seg_count; i++) {
			seg_size = pkt_reconstr_SegSize[slot][i];
			memcpy(IPv4_RX_buffer + 14 + loc_size - 2, pkt_reconstr_SegData[slot][i], seg_size);// -2 because of dest_MAC_IP field 2 bytes
			loc_size = loc_size + seg_size;
		}
		//printf("***** received_full_packet IPv4 ********\r\n"); 
		IPv4_to_LAN(IPv4_RX_buffer, 14 + loc_size - 2, pkt_reconstr_TXID[slot], pkt_reconstr_RXID[slot]);
		NPR_IPv4_RX_free_reconstr_slot(slot);
	}
}

void NPRV_IPv4_fill_reconstr_RX(
	unsigned char* data, int seg_size, 
	unsigned int TX_ID, unsigned int RX_ID,
	unsigned char is_last_seg, unsigned char seg_counter, unsigned char pkt_counter, uint32_t loc_time
) {
	int i;
	int matching_slot = 0xFFFF;
	int new_slot;
	int delta_T;
	for (i = 0; i < NPRV_IPV4_RX_SLOT_CNT; i++) {
		delta_T = loc_time - pkt_reconstr_Time[i];
		delta_T = abs(delta_T);
		if ((pkt_reconstr_occupied[i] == 1) && 
			(pkt_reconstr_TXID[i] == TX_ID) && (pkt_reconstr_RXID[i] == RX_ID) &&
			(pkt_reconstr_PktCounter[i] == pkt_counter) // && (delta_T < 5800000)620000	
			) { 
			matching_slot = i;
		}
	}
	//we found a previous matching entry
	if (matching_slot != 0xFFFF) {
		if (loc_time < pkt_reconstr_Time[matching_slot]) {
			pkt_reconstr_Time[matching_slot] = loc_time;
		}
		pkt_reconstr_SegSize[matching_slot][seg_counter] = seg_size;
		pkt_reconstr_SegIsLast[matching_slot][seg_counter] = is_last_seg;
		memcpy(pkt_reconstr_SegData[matching_slot][seg_counter], data, seg_size);
		NPR_IPv4_RX_attempt_reconstr_pkt(matching_slot);
	}
	else {//no previous entry found
		new_slot = NPRV_IPv4_find_empty_reconstr_slot(loc_time);
		if (new_slot == 0xFFFF) {
			printf("ERROR NO IPV4 reconstr slot found\r\n");
		}
		else {
			pkt_reconstr_occupied[new_slot] = 1;
			pkt_reconstr_TXID[new_slot] = TX_ID;
			pkt_reconstr_RXID[new_slot] = RX_ID;
			pkt_reconstr_Time[new_slot] = loc_time;
			pkt_reconstr_PktCounter[new_slot] = pkt_counter;

			pkt_reconstr_SegSize[new_slot][seg_counter] = seg_size;
			pkt_reconstr_SegIsLast[new_slot][seg_counter] = is_last_seg;
			memcpy(pkt_reconstr_SegData[new_slot][seg_counter], data, seg_size);
			NPR_IPv4_RX_attempt_reconstr_pkt(new_slot);
		}
	}
}

void NPRV_IPv4_rFrame_interpret(unsigned char* data_RX, int rframe_size, uint32_t sync_time) {//size of data field
	static unsigned char IPv4_RX_buffer[NPRV_MAX_CLIENT_NB][1600] = { 0 };
	static int size_received[NPRV_MAX_CLIENT_NB] = { 0 };
	static unsigned char prev_seg_counter[NPRV_MAX_CLIENT_NB] = { 0 };
	static unsigned char curr_pkt_counter[NPRV_MAX_CLIENT_NB] = { 0 };
	unsigned int dest_ID, source_ID;
	unsigned char is_last_seg, seg_counter, pkt_counter;
	int seg_size;
	int i;
	unsigned char* data_field;
#ifdef NPRV_WITH_HEADER_FEC_RS
	data_field = data_RX + 12;
#else
	data_field = data_RX + 8;
#endif
	source_ID = ((data_RX[3] & 0x1F) << 8) + data_RX[4];
	dest_ID = ((data_field[0] & 0x1F) << 8) + data_field[1];//[6] [7]
	pkt_counter = (data_field[2] & 0x0F);//[8]
	is_last_seg = (data_field[3] & 0x80) >> 7;//[9]
	seg_counter = data_field[3] & 0x7F;//[9]

	seg_size = rframe_size - 6;// -6: 4 bytes of header after and 2 bytes checksum
	
	NPRV_IPv4_fill_reconstr_RX(data_field + 4, seg_size, source_ID, dest_ID, is_last_seg, seg_counter, pkt_counter, sync_time);
}

int NPRV_check_IPv4_TXID_match(uint32_t source_IP_addr, unsigned int TXID) {
	int answer = 0;
	uint32_t last_IP;
	// !!!A revoir intgralement avec nouvelle donne!!!Notamment routage IP 
	if (TXID == 0) {
		//last_IP = CONF_radio_IP_start + CONF_radio_IP_size;
		//if ((source_IP_addr < CONF_radio_IP_start) || (source_IP_addr >= last_IP)) {
		//	answer = 1;
		//}
		answer = 1;
	} 
	else if (TXID < NPRV_MAX_CLIENT_NB) {
	//if (TXID < NPRV_MAX_CLIENT_NB) {
		last_IP = CONF_radio_addr_table_IP_begin[TXID] + CONF_radio_addr_table_IP_size[TXID];
		if ((source_IP_addr >= CONF_radio_addr_table_IP_begin[TXID]) && (source_IP_addr < last_IP)) {
			answer = 1;
		}
	}

	return answer;
}

void IPv4_to_LAN(unsigned char* RX_eth_frame, int RX_size, unsigned int TXID, unsigned int RXID) {
	uint32_t dest_IP_addr;
	uint32_t source_IP_addr;
	uint32_t LAN_dest_IP;
	int local_size;
	int i;
	int dest_MAC_found;
	int is_inside_subnet = 0;
	int is_inside_my_IP_range = 0;
	//int is_inside_radio_range = 0;
	int LAN_TX_need = 0;
	int IPv4_TXID_match = 0;
	dest_IP_addr = IP_char2int(RX_eth_frame + 30);
	source_IP_addr = IP_char2int(RX_eth_frame + 26);
	IPv4_TXID_match = NPRV_check_IPv4_TXID_match(source_IP_addr, TXID);
	//printf("source IP %08X TXID %i match %i\r\n", source_IP_addr, TXID, 1);
	//printf("Dest IPv4:%08X\r\n", dest_IP_addr);
	local_size = (RX_eth_frame[16] << 8) + RX_eth_frame[17];
	//now we check if packet is really for our part of the subnet
	if ((dest_IP_addr >= LAN_conf_applied.DHCP_range_start) && (dest_IP_addr < (LAN_conf_applied.DHCP_range_start + LAN_conf_applied.DHCP_range_size))) {
		is_inside_my_IP_range = 1;
	}
	if (CONF_is_NPRV_HUB == 0) { // clients behaviour
		// Checks if dest IP is inside local range
		if ((is_inside_my_IP_range) && (IPv4_TXID_match == 1)) {
			LAN_TX_need = 1;
			LAN_dest_IP = dest_IP_addr;
		}
	}
	if ((LAN_TX_need) && (local_size < 1510)) {
		//printf("IPv4 Eth TX\r\n");
		for (i = 0; i < 6; i++) {
			RX_eth_frame[i + 6] = CONF_modem_MAC[i];
		}
		RX_eth_frame[12] = 0x08; // Ethertype IPv4
		RX_eth_frame[13] = 0x00;
		//printf("Looking for IP:%08X\r\n", LAN_dest_IP);
		dest_MAC_found = lookfor_MAC_from_IP(RX_eth_frame, LAN_dest_IP);
		if (dest_MAC_found) {
			mutex_tuntap_TX.lock();
			tuntap_send_packet(RX_eth_frame, local_size + 14);
			mutex_tuntap_TX.unlock();
			//printf("IP SENT TO LAN: ");
			/*
			for (i = 0; i < local_size + 14; i++) {
				printf("%02X:", RX_eth_frame[i]);
			}
			*/
			//printf("\r\n");
		}
		else {
			//printf("MAC NOT FOUND\r\n");
			temp_Eth_buff_IP = LAN_dest_IP;
			temp_Eth_buff_size = local_size + 14;
			memcpy(temp_Eth_buffer, RX_eth_frame, temp_Eth_buff_size);
		}
	}
	if ((CONF_IPv4_SPY >= 1) && 
		(RXID != G_NPRV_MyClientID) && (RXID < NPRV_MAX_CLIENT_NB) &&
		(TXID != G_NPRV_MyClientID) &&
		(local_size < 1510)) {
		for (i = 0; i < 6; i++) {
			RX_eth_frame[i] = CONF_virtual_MAC_SPY_RX[i];
		}
		for (i = 0; i < 6; i++) {
			RX_eth_frame[i + 6] = CONF_modem_MAC[i];
		}
		RX_eth_frame[12] = 0x08; // Ethertype IPv4
		RX_eth_frame[13] = 0x00;
		mutex_tuntap_TX.lock();
		tuntap_send_packet(RX_eth_frame, local_size + 14);
		mutex_tuntap_TX.unlock();
	}
}

void flush_temp_Eth_buffer(uint32_t loc_IP) {
	if (temp_Eth_buff_size > 0) {
		if (loc_IP == temp_Eth_buff_IP) {
			lookfor_MAC_from_IP(temp_Eth_buffer, loc_IP); //puts MAC inside Eth
			mutex_tuntap_TX.lock();
			tuntap_send_packet(temp_Eth_buffer, temp_Eth_buff_size);
			mutex_tuntap_TX.unlock();
		}
		temp_Eth_buff_size = 0; 
	}
}


int Eth_RX_dequeue(void) {
	int answer = 0;
	unsigned char RX_data[2600] = { 0 };//1600
	unsigned int RX_port = 0;
	unsigned char RX_proto;
	uint32_t RX_dest_IP = 0;
	unsigned char source_MAC[8];
	unsigned char dest_MAC[8];

	static int more_to_read = 0;

	int RX_size = 0;
	int mac_size = 0;
	int i;
	unsigned int ethertype;
	int Eth_dest_broadcast;
	int Eth_dest_unicast; 
	
	do {
		RX_size = tuntap_recv_packet(RX_data);
		if (RX_size > 0) {
			answer = 1;
			mac_size = RX_size;
			Eth_dest_broadcast = 1;
			Eth_dest_unicast = 1;
			for (i = 0; i < 6; i++) {
				dest_MAC[i] = RX_data[i];
				source_MAC[i] = RX_data[i+6];
				if (dest_MAC[i] != 0xFF) { Eth_dest_broadcast = 0; }
				if (dest_MAC[i] != CONF_modem_MAC[i]) { Eth_dest_unicast = 0; }
			}

			ethertype = (RX_data[12] << 8) + RX_data[13];


			if (Eth_dest_broadcast || Eth_dest_unicast) {
				//printf("size:%i Ethertype:%04X\r\n", RX_size, ethertype);

				//printf("source:%02X:%02X:%02X:%02X:%02X:%02X\r\n", source_MAC[0], source_MAC[1], source_MAC[2], source_MAC[3], source_MAC[4], source_MAC[5]);
				//printf("dest:%02X:%02X:%02X:%02X:%02X:%02X\r\n", dest_MAC[0], dest_MAC[1], dest_MAC[2], dest_MAC[3], dest_MAC[4], dest_MAC[5]);
				//printf("ethertype:%04X\r\n", ethertype);
				if (ethertype == 0x0806) { //ARP packet received
					//printf("RX ARP\r\n");

					ARP_RX_packet_treatment(RX_data, mac_size);
				}

				if ( (ethertype == 0x0800)  ) { // IPv4 packet // && (Eth_dest_unicast)
					//printf("RX IPv4\r\n");
					IPv4_RX_dequeue(RX_data+14, mac_size-14, Eth_dest_unicast);
				}
			}
		}
	} while (RX_size > 0);
	return answer;
}

void IPv4_RX_dequeue(unsigned char* data, int size, int is_unicast) {
	//unsigned long int dest_IP;
	uint32_t dest_IP;
	unsigned int source_port;
	unsigned int dest_port;
	unsigned char IPv4_proto;

	dest_IP = IP_char2int(data+16);
	IPv4_proto = data[9];
	source_port = (data[20] << 8) + data[21];
	dest_port = (data[22] << 8) + data[23];
	//printf("dest_IP:%08X proto:%04X\r\n", dest_IP, IPv4_proto);
	if (G_NPRV_MyIPConfigured == 1) {
		if ((dest_IP == LAN_conf_applied.LAN_modem_IP) && (IPv4_proto == 0x01)) {
			//printf("ICMP for me\r\n");
			ICMP_reply(data, size);
		}
		else if ((IPv4_proto == 0x11) && (source_port == 0x0044) && (dest_port == 0x0043)) {
			//printf("DHCP\r\n");
			if (LAN_conf_applied.DHCP_server_active == 1) {
				DHCP_server(data + 20, size - 20);
			}
		}
		else if (is_unicast == 1) {
			//printf("IPv4 to radio\r\n");
			IPv4_to_radio(data, size);
		}
	}
}

void IPv4_to_radio(unsigned char* data, int size) {
	int i;
	uint32_t dest_IP_addr, source_IP_addr;
	int dest_is_inside_subnet = 0;
	int dest_is_inside_my_client_range = 0;
	int source_is_inside_my_client_range = 0;
	//int is_outside_NPRV_range = 0;
	unsigned int dest_client_ID;
	int radio_tx_need = 0;

	dest_IP_addr = IP_char2int(data + 16);
	source_IP_addr = IP_char2int(data + 12);

	if ((dest_IP_addr >= LAN_conf_applied.DHCP_range_start) && (dest_IP_addr < (LAN_conf_applied.DHCP_range_start + LAN_conf_applied.DHCP_range_size))) {
		dest_is_inside_my_client_range = 1;
	}
	if ((source_IP_addr >= LAN_conf_applied.DHCP_range_start) && (source_IP_addr < (LAN_conf_applied.DHCP_range_start + LAN_conf_applied.DHCP_range_size))) {
		source_is_inside_my_client_range = 1;
		//printf("Source IP: %8X OK:%i\r\n", source_IP_addr, source_is_inside_my_client_range);
		//NPRV_HMI_new_TXT_message();
		//NPRV_emul_HMI_new_TXT_message();
	}
	if ((dest_IP_addr & LAN_conf_applied.LAN_subnet_mask) == (LAN_conf_applied.LAN_modem_IP & LAN_conf_applied.LAN_subnet_mask)) {
		dest_is_inside_subnet = 1;
	}
	if ((CONF_is_NPRV_HUB == 0) && (dest_IP_addr != LAN_conf_applied.LAN_modem_IP)) { // TDMA Client
		
		if ((dest_is_inside_subnet == 1) && 
			(dest_is_inside_my_client_range == 0) && 
			(source_is_inside_my_client_range == 1))//!!! remove & change condition if routing allowed at client side
		{
			dest_client_ID = lookfor_client_ID_from_IP(dest_IP_addr);
			if ((dest_client_ID < NPRV_MAX_CLIENT_NB) && (dest_client_ID != G_NPRV_MyClientID)) {
				radio_tx_need = 1;
				//printf("%i.%i.%i.%iRadio TX to a CLIENT %i\r\n", data[16], data[17], data[18], data[19], dest_client_ID);
				//NPRV_HMI_new_TXT_message();
			}
		}
		
		// outside subnet and IP gateway active -> to VSAT HUB
		if ((dest_is_inside_subnet == 0) && (LAN_conf_applied.LAN_def_route_activ != 0)) {  
			//!!! A changer quand j'introduirais le MAC-IP virtuelle et multi routage
			dest_client_ID = 0;// 0 is the VSAT HUB
			radio_tx_need = 1;
			//printf("%i.%i.%i.%iRadio TX to gateway\r\n", data[16], data[17], data[18], data[19]);
			//NPRV_HMI_new_TXT_message();
		}
	}
	if ((radio_tx_need == 1) && (G_NPRV_MyRadioState == 2)) {
		//printf("final TX decision client ID %i\r\n", dest_client_ID);
		NPRV_IPv4_push_TX_pckt(data, size, dest_client_ID);
	}
}

void UDP_to_LAN (unsigned char* data_UDP, int size_UDP, uint32_t dest_IP, unsigned char* dest_MAC, unsigned int source_port, unsigned int dest_port) {
	unsigned char Eth_data[1600] = { 0 };
	int size_out;
	int i;
	unsigned int loc_UDP_length, loc_IP_size; 
	unsigned int IPv4_chksum;

	memcpy(Eth_data+42, data_UDP, size_UDP);
	size_out = size_UDP + 42;

	Eth_data[40] = 0x00; //!!! checksum
	Eth_data[41] = 0x00; //!!! checksum

	loc_UDP_length = size_UDP + 8;
	Eth_data[38] = (loc_UDP_length & 0xFF00) >> 8;
	Eth_data[39] = loc_UDP_length & 0xFF;
	
	Eth_data[36] = (dest_port & 0xFF00) >> 8;
	Eth_data[37] = dest_port & 0xFF;

	Eth_data[34] = (source_port & 0xFF00) >> 8;
	Eth_data[35] = source_port & 0xFF;

	IP_int2char(dest_IP, Eth_data + 30);
	IP_int2char(LAN_conf_applied.LAN_modem_IP, Eth_data + 26);

	Eth_data[24] = 0x00;// checksum
	Eth_data[25] = 0x00;

	Eth_data[23] = 0x11;//IP protocol
	Eth_data[22] = 0x80;//TTL

	Eth_data[20] = 0x40;//don't fragment
	Eth_data[21] = 0x00;

	Eth_data[18] = (IP_ident & 0xFF00) >> 8;
	Eth_data[19] = IP_ident & 0xFF;
	IP_ident++;

	loc_IP_size = size_UDP + 28;
	Eth_data[16] = (loc_IP_size & 0xFF00) >> 8;
	Eth_data[17] = loc_IP_size & 0xFF;

	Eth_data[15] = 0x00; //type_service
	Eth_data[14] = 0x45; //header length
	//checksum IP!!! 

	IPv4_chksum = 0 + (Eth_data[14] << 8) + Eth_data[15];
	//printf("chksum:%06X\r\n", IPv4_chksum);
	IPv4_chksum = IPv4_chksum + (Eth_data[16] << 8) + Eth_data[17];
	//printf("chksum:%06X\r\n", IPv4_chksum);
	IPv4_chksum = IPv4_chksum + (Eth_data[18] << 8) + Eth_data[19];
	//printf("chksum:%06X\r\n", IPv4_chksum);
	IPv4_chksum = IPv4_chksum + (Eth_data[20] << 8) + Eth_data[21];
	IPv4_chksum = IPv4_chksum + (Eth_data[22] << 8) + Eth_data[23];

	IPv4_chksum = IPv4_chksum + (Eth_data[26] << 8) + Eth_data[27];
	IPv4_chksum = IPv4_chksum + (Eth_data[28] << 8) + Eth_data[29];
	IPv4_chksum = IPv4_chksum + (Eth_data[30] << 8) + Eth_data[31];
	IPv4_chksum = IPv4_chksum + (Eth_data[32] << 8) + Eth_data[33];
	//printf("chksum:%06X\r\n", IPv4_chksum);
	IPv4_chksum = IPv4_chksum + ((IPv4_chksum >> 16) & 0xFF);
	//printf("chksum add:%06X\r\n", IPv4_chksum);
	IPv4_chksum = IPv4_chksum & 0xFFFF;
	IPv4_chksum = 0xFFFF - IPv4_chksum;
	//printf("chksum:%06X\r\n", IPv4_chksum);
	Eth_data[24] = (IPv4_chksum & 0xFF00) >> 8;
	Eth_data[25] = IPv4_chksum & 0xFF;

	for (i = 0; i < 6; i++) {				// Source MAC
		Eth_data[i + 6] = CONF_modem_MAC[i];
		Eth_data[i] = dest_MAC[i];
	}
	Eth_data[12] = 0x08;
	Eth_data[13] = 0x00;
	mutex_tuntap_TX.lock();
	tuntap_send_packet(Eth_data, size_out);
	mutex_tuntap_TX.unlock();
}

void ICMP_reply(unsigned char* req_IP_pckt, int raw_size) {
	unsigned char answ_eth_frame[1600] = { 0 };
	unsigned char* answ_IP_pckt = answ_eth_frame + 14;
	unsigned char type_mess;
	uint32_t IP_source; 
	unsigned int chksum_temp;
	unsigned int IPv4_chksum;
	unsigned int loc_IP_size;
	int i;
	type_mess = req_IP_pckt[20];
	IP_source = IP_char2int(req_IP_pckt + 12);

	if (req_IP_pckt[20] == 0x08) {
		//printf("trying to answer\r\n");
		answ_IP_pckt[20] = 0x00;
		answ_IP_pckt[21] = 0x00;
		chksum_temp = (req_IP_pckt[22] << 8) + req_IP_pckt[23];
		chksum_temp = chksum_temp + 0x0800;
		chksum_temp = chksum_temp + ((chksum_temp >> 16) & 0xFF);;
		answ_IP_pckt[22] = (chksum_temp & 0xFF00) >> 8;;
		answ_IP_pckt[23] = chksum_temp & 0xFF;

		for (i=24; i<raw_size; i++) {
			answ_IP_pckt[i] = req_IP_pckt[i];
		}

		IP_int2char(IP_source, answ_IP_pckt +16);
		IP_int2char(LAN_conf_applied.LAN_modem_IP, answ_IP_pckt + 12);

		answ_IP_pckt[9] = 0x01;//ICMP protocol
		answ_IP_pckt[8] = 0x80;//TTL

		answ_IP_pckt[6] = 0x40;//don't fragment
		answ_IP_pckt[7] = 0x00;

		answ_IP_pckt[4] = (IP_ident & 0xFF00) >> 8;
		answ_IP_pckt[5] = IP_ident & 0xFF;
		IP_ident++;

		loc_IP_size = raw_size;
		answ_IP_pckt[2] = (loc_IP_size & 0xFF00) >> 8;
		answ_IP_pckt[3] = loc_IP_size & 0xFF;

		answ_IP_pckt[1] = 0x00; //type_service
		answ_IP_pckt[0] = 0x45; //header length

		IPv4_chksum = 0 + (answ_IP_pckt[0] << 8) + answ_IP_pckt[1];
		//printf("chksum:%06X\r\n", IPv4_chksum);
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[2] << 8) + answ_IP_pckt[3];
		//printf("chksum:%06X\r\n", IPv4_chksum);
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[4] << 8) + answ_IP_pckt[5];
		//printf("chksum:%06X\r\n", IPv4_chksum);
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[6] << 8) + answ_IP_pckt[7];
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[8] << 8) + answ_IP_pckt[9];
		//do not count the chksum slot
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[12] << 8) + answ_IP_pckt[13];
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[14] << 8) + answ_IP_pckt[15];
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[16] << 8) + answ_IP_pckt[17];
		IPv4_chksum = IPv4_chksum + (answ_IP_pckt[18] << 8) + answ_IP_pckt[19];
		//printf("chksum:%06X\r\n", IPv4_chksum);
		IPv4_chksum = IPv4_chksum + ((IPv4_chksum >> 16) & 0xFF);
		//printf("chksum add:%06X\r\n", IPv4_chksum);
		IPv4_chksum = IPv4_chksum & 0xFFFF;
		IPv4_chksum = 0xFFFF - IPv4_chksum;
		//printf("chksum:%06X\r\n", IPv4_chksum);
		answ_IP_pckt[10] = (IPv4_chksum & 0xFF00) >> 8;
		answ_IP_pckt[11] = IPv4_chksum & 0xFF;
		
		IPv4_to_LAN(answ_eth_frame, loc_IP_size + 14, 0, 0xFFFF);
	}
}


unsigned int lookfor_client_ID_from_IP(uint32_t IP_addr) {
	unsigned int i, i_found;
	uint32_t last_IP;
	i_found = 0xFFFF;

	for (i = 0; i < NPRV_MAX_CLIENT_NB; i++) {//one of the clients
		last_IP = CONF_radio_addr_table_IP_begin[i] + CONF_radio_addr_table_IP_size[i];
		if ((IP_addr >= CONF_radio_addr_table_IP_begin[i]) && (IP_addr < last_IP) &&
			(CONF_radio_addr_table_status[i] == 1) && //2023_01_07
			(i != G_NPRV_MyClientID))
		{
			i_found = i;
		}
	}
	//printf("lookfor client ID from IP %i\r\n", i_found);
	return i_found;
}
