// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include "NPRV_HMI.h"
#include "Eth_IPv4.h"
#include <math.h>
#include <SFML/System.hpp>

#ifdef NPRV_linux
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#endif

#pragma warning(disable: 4996)
#define NPRV_HEADER_BYTE_SIZE_HMI 12

sf::Mutex NPRV_HMI_mutex_stats_err;

static int HMI_TOG = 0;
static int BER_err_count_BPSK[2] = { 0 };
static int BER_tot_word_BPSK[2] = { 0 };
static int BER_err_count_QPSK[2] = { 0 };
static int BER_tot_word_QPSK[2] = { 0 };
static int BER_err_count_BPSK_hub[2] = { 0 };
static int BER_tot_word_BPSK_hub[2] = { 0 };
static int BER_err_count_QPSK_hub[2] = { 0 };
static int BER_tot_word_QPSK_hub[2] = { 0 };
static int BER_err_count_BPSK_me[2] = { 0 };
static int BER_tot_word_BPSK_me[2] = { 0 };
static int BER_err_count_QPSK_me[2] = { 0 };
static int BER_tot_word_QPSK_me[2] = { 0 };

void NPRV_BER_stat_save(int errors, int words, int modul_type) {
	if (modul_type == NPRV_BPSK) {
		BER_err_count_BPSK[HMI_TOG] = BER_err_count_BPSK[HMI_TOG] + errors;
		BER_tot_word_BPSK[HMI_TOG] = BER_tot_word_BPSK[HMI_TOG] + words;
	}
	if (modul_type == NPRV_QPSK) {
		BER_err_count_QPSK[HMI_TOG] = BER_err_count_QPSK[HMI_TOG] + errors;
		BER_tot_word_QPSK[HMI_TOG] = BER_tot_word_QPSK[HMI_TOG] + words;
	}
}

void NPRV_HMI_save_BER_frames(int modul, int BER_header, int BER_DataField, int byte_size, unsigned int TX_ID) {
	if (TX_ID == 0) {
		NPRV_HMI_mutex_stats_err.lock();
		BER_err_count_BPSK_hub[HMI_TOG] = BER_err_count_BPSK_hub[HMI_TOG] + BER_header;
		BER_tot_word_BPSK_hub[HMI_TOG] = BER_tot_word_BPSK_hub[HMI_TOG] + NPRV_HEADER_BYTE_SIZE_HMI;
		if (modul == NPRV_BPSK) {
			BER_err_count_BPSK_hub[HMI_TOG] = BER_err_count_BPSK_hub[HMI_TOG] + BER_DataField;
			BER_tot_word_BPSK_hub[HMI_TOG] = BER_tot_word_BPSK_hub[HMI_TOG] + byte_size;
		}
		if (modul == NPRV_QPSK) {
			BER_err_count_QPSK_hub[HMI_TOG] = BER_err_count_QPSK_hub[HMI_TOG] + BER_DataField;
			BER_tot_word_QPSK_hub[HMI_TOG] = BER_tot_word_QPSK_hub[HMI_TOG] + byte_size;
		}
		NPRV_HMI_mutex_stats_err.unlock();
	}
	if ((TX_ID == G_NPRV_MyClientID) && 
		((G_NPRV_MyClientID < NPRV_MAX_CLIENT_NB) || (G_NPRV_MyClientID == 0x1FFA))) 
	{
		NPRV_HMI_mutex_stats_err.lock();
		BER_err_count_BPSK_me[HMI_TOG] = BER_err_count_BPSK_me[HMI_TOG] + BER_header;
		BER_tot_word_BPSK_me[HMI_TOG] = BER_tot_word_BPSK_me[HMI_TOG] + NPRV_HEADER_BYTE_SIZE_HMI;
		if (modul == NPRV_BPSK) {
			BER_err_count_BPSK_me[HMI_TOG] = BER_err_count_BPSK_me[HMI_TOG] + BER_DataField;
			BER_tot_word_BPSK_me[HMI_TOG] = BER_tot_word_BPSK_me[HMI_TOG] + byte_size;
		}
		if (modul == NPRV_QPSK) {
			BER_err_count_QPSK_me[HMI_TOG] = BER_err_count_QPSK_me[HMI_TOG] + BER_DataField;
			BER_tot_word_QPSK_me[HMI_TOG] = BER_tot_word_QPSK_me[HMI_TOG] + byte_size;
		}
		NPRV_HMI_mutex_stats_err.unlock();
	}
}

void NPRV_get_BER_stat(float* BER_answer_BPSK, float* BER_answer_QPSK) {
	if ((BER_tot_word_BPSK[0] + BER_tot_word_BPSK[1]) > 0) {
		*BER_answer_BPSK = 100 * (float)(BER_err_count_BPSK[0] + BER_err_count_BPSK[1]) / (float)(BER_tot_word_BPSK[0] + BER_tot_word_BPSK[1]);
	}
	else {
		*BER_answer_BPSK = 200;
	}
	if ((BER_tot_word_QPSK[0] + BER_tot_word_QPSK[1]) > 0) {
		*BER_answer_QPSK = 100 * (float)(BER_err_count_QPSK[0] + BER_err_count_QPSK[1]) / (float)(BER_tot_word_QPSK[0] + BER_tot_word_QPSK[1]);
	}
	else {
		*BER_answer_QPSK = 200;
	}
	BER_err_count_BPSK[HMI_TOG] = 0;
	BER_tot_word_BPSK[HMI_TOG] = 0 ;
	BER_err_count_QPSK[HMI_TOG] = 0;
	BER_tot_word_QPSK[HMI_TOG] = 0;
}



static int FRAME_err_count_BPSK[2] = { 0 };
static int FRAME_tot_count_BPSK[2] = { 0 };
static int FRAME_err_count_QPSK[2] = { 0 };
static int FRAME_tot_count_QPSK[2] = { 0 };

static int FRAME_err_count_BPSK_me[2] = { 0 };
static int FRAME_tot_count_BPSK_me[2] = { 0 };
static int FRAME_err_count_QPSK_me[2] = { 0 };
static int FRAME_tot_count_QPSK_me[2] = { 0 };
static int FRAME_err_count_BPSK_hub[2] = { 0 };
static int FRAME_tot_count_BPSK_hub[2] = { 0 };
static int FRAME_err_count_QPSK_hub[2] = { 0 };
static int FRAME_tot_count_QPSK_hub[2] = { 0 };
static int FRAME_TOT_count_QPSK_me = { 0 };
static int FRAME_TOT_count_BPSK_me = { 0 };
static int FRAME_TOT_count_QPSK_hub = { 0 };
static int FRAME_TOT_count_BPSK_hub = { 0 };

static int FRAME_TX_tot_BPSK[2] = { 0 };
static int FRAME_TX_tot_QPSK[2] = { 0 };
static int FRAME_TX_TOT_BPSK = 0;
static int FRAME_TX_TOT_QPSK = 0;

void NPRV_save_stat_TX(int modulation) {
	if (modulation == NPRV_BPSK) {
		FRAME_TX_tot_BPSK[HMI_TOG]++;
		FRAME_TX_TOT_BPSK++;
	}
	if (modulation == NPRV_QPSK) {
		FRAME_TX_tot_QPSK[HMI_TOG]++;
		FRAME_TX_TOT_QPSK++;
	}
}

void NPRV_RX_discovery_state_save(uint32_t rand_RX) {
	uint32_t my_rand_nb;
	my_rand_nb =
		(G_NPRV_my_random_nb[0] << 24) +
		(G_NPRV_my_random_nb[1] << 16) +
		(G_NPRV_my_random_nb[2] << 8) +
		(G_NPRV_my_random_nb[3]);
	if (rand_RX == my_rand_nb) {
		FRAME_tot_count_BPSK_me[HMI_TOG] = FRAME_tot_count_BPSK_me[HMI_TOG] + 1;
		FRAME_TOT_count_BPSK_me++;
	}
}

void NPRV_FrameErr_stat_save(int chcksum_OK, int modulation, unsigned int TX_ID) {
	NPRV_HMI_mutex_stats_err.lock();
	if (modulation == NPRV_BPSK) {
		FRAME_err_count_BPSK[HMI_TOG] = FRAME_err_count_BPSK[HMI_TOG] + (1 - chcksum_OK);
		FRAME_tot_count_BPSK[HMI_TOG] = FRAME_tot_count_BPSK[HMI_TOG] + 1;
	}
	if (modulation == NPRV_QPSK) {
		FRAME_err_count_QPSK[HMI_TOG] = FRAME_err_count_QPSK[HMI_TOG] + (1 - chcksum_OK);
		FRAME_tot_count_QPSK[HMI_TOG] = FRAME_tot_count_QPSK[HMI_TOG] + 1;
	}
	if (TX_ID == 0) {
		if (modulation == NPRV_BPSK) {
			FRAME_err_count_BPSK_hub[HMI_TOG] = FRAME_err_count_BPSK_hub[HMI_TOG] + (1 - chcksum_OK);
			FRAME_tot_count_BPSK_hub[HMI_TOG] = FRAME_tot_count_BPSK_hub[HMI_TOG] + 1;
			FRAME_TOT_count_BPSK_hub++;

		}
		if (modulation == NPRV_QPSK) {
			FRAME_err_count_QPSK_hub[HMI_TOG] = FRAME_err_count_QPSK_hub[HMI_TOG] + (1 - chcksum_OK);
			FRAME_tot_count_QPSK_hub[HMI_TOG] = FRAME_tot_count_QPSK_hub[HMI_TOG] + 1;
			FRAME_TOT_count_QPSK_hub++;
		}
	}
	if ((TX_ID == G_NPRV_MyClientID) && 
		((G_NPRV_MyClientID < NPRV_MAX_CLIENT_NB) || (G_NPRV_MyClientID == 0x1FFA))) {
		if (modulation == NPRV_BPSK) {
			FRAME_err_count_BPSK_me[HMI_TOG] = FRAME_err_count_BPSK_me[HMI_TOG] + (1 - chcksum_OK);
			FRAME_tot_count_BPSK_me[HMI_TOG] = FRAME_tot_count_BPSK_me[HMI_TOG] + 1;
			FRAME_TOT_count_BPSK_me++;
		}
		if (modulation == NPRV_QPSK) {
			FRAME_err_count_QPSK_me[HMI_TOG] = FRAME_err_count_QPSK_me[HMI_TOG] + (1 - chcksum_OK);
			FRAME_tot_count_QPSK_me[HMI_TOG] = FRAME_tot_count_QPSK_me[HMI_TOG] + 1;
			FRAME_TOT_count_QPSK_me++;
		}
	}
	NPRV_HMI_mutex_stats_err.unlock();
}

void NPRV_get_FrameErr_stat(float* err_2FSK, float* err_4FSK) {
	if ((FRAME_tot_count_BPSK[0] + FRAME_tot_count_BPSK[1]) > 0) {
		(*err_2FSK) = 100 * (float)(FRAME_err_count_BPSK[0] + FRAME_err_count_BPSK[1]) / (float)(FRAME_tot_count_BPSK[0] + FRAME_tot_count_BPSK[1]);
	}
	else { (*err_2FSK) = 200; }
	if ((FRAME_tot_count_QPSK[0] + FRAME_tot_count_QPSK[1]) > 0) {
		(*err_4FSK) = 100 * (float)(FRAME_err_count_QPSK[0] + FRAME_err_count_QPSK[1]) / (float)(FRAME_tot_count_QPSK[0] + FRAME_tot_count_QPSK[1]);
	}
	else { (*err_4FSK) = 200; }
	FRAME_err_count_BPSK[HMI_TOG] = 0;
	FRAME_tot_count_BPSK[HMI_TOG] = 0;
	FRAME_err_count_QPSK[HMI_TOG] = 0;
	FRAME_tot_count_QPSK[HMI_TOG] = 0;
}

static float NPRV_HMI_ampl_RX_hub = 0;
static float NPRV_HMI_ampl_RX_me = 0;
void NPRV_HMI_store_amp_RX_frame(uint32_t TX_ID, float amplitude, int channel) {
	const float normalisation_chan_type[20] = { 1, 1.4142, 0.7071 };//square root of 2
	int chan_type;
	float amplitude_2;
	chan_type = NPRV_channels_SR_std[channel];
	amplitude_2 = amplitude * normalisation_chan_type[chan_type];
	if (TX_ID == 0) {
		NPRV_HMI_ampl_RX_hub = 0.9 * NPRV_HMI_ampl_RX_hub + 0.1 * amplitude_2;
	}
	if ((TX_ID == G_NPRV_MyClientID) && 
		((G_NPRV_MyClientID < NPRV_MAX_CLIENT_NB) || (G_NPRV_MyClientID == 0x1FFA))) {
		if (NPRV_HMI_ampl_RX_me == 0) {
			NPRV_HMI_ampl_RX_me = amplitude_2;
		} else {
			NPRV_HMI_ampl_RX_me = 0.9 * NPRV_HMI_ampl_RX_me + 0.1 * amplitude_2;
		}
	}
}

/*
int NPRV_HMI_print_radio_stats(void) {
	float err_me_BPSK, err_me_QPSK, err_hub_BPSK, err_hub_QPSK;
	float BER_me_BPSK, BER_me_QPSK, BER_hub_BPSK, BER_hub_QPSK;
	if (FRAME_tot_count_BPSK_me > 0) {
		err_me_BPSK = 100 * (float)FRAME_err_count_BPSK_me / (float)FRAME_tot_count_BPSK_me;
	}
	else { err_me_BPSK = 999.99; }
	if (FRAME_tot_count_QPSK_me > 0) {
		err_me_QPSK = 100 * (float)FRAME_err_count_QPSK_me / (float)FRAME_tot_count_QPSK_me;
	}
	else { err_me_QPSK = 999.99; }
	
	if (FRAME_tot_count_BPSK_hub > 0) {
		err_hub_BPSK = 100 * (float)FRAME_err_count_BPSK_hub / (float)FRAME_tot_count_BPSK_hub;
	}
	else { err_hub_BPSK = 999.99; }
	if (FRAME_tot_count_QPSK_hub > 0) {
		err_hub_QPSK = 100 * (float)FRAME_err_count_QPSK_hub / (float)FRAME_tot_count_QPSK_hub;
	}
	else { err_hub_QPSK = 999.99; }
	//BER_err_count_QPSK_me = 0;
	//static int BER_tot_word_QPSK_me
	//BER_me_BPSK
	if (BER_tot_word_BPSK_me > 0) {

	}

	printf("                +-------------+-------------+\r\n");
	printf("                |  from Hub   |   from ME   |\r\n");
	printf("                +------+------+------+------+\r\n");
	printf("                | BPSK | QPSK | BPSK | QPSK |\r\n");
	printf("----------------+------+------+------+------+\r\n");
	printf("  frame TX count|      |      |%6i|%6i|\r\n", FRAME_TX_TOT_BPSK, FRAME_TX_TOT_QPSK);
	printf("  frame TX rate |      |      | %3i/s| %3i/s|\r\n", FRAME_TX_tot_BPSK, FRAME_TX_tot_QPSK);
	//printf("frame RX count |%3i/s%6i|%3i/s%6i|%3i/s%6i|%3i/s%6i|\r\n", 
	//	FRAME_tot_count_BPSK_hub, FRAME_TOT_count_BPSK_hub, 
	//	FRAME_tot_count_QPSK_hub, FRAME_TOT_count_QPSK_hub,
	//	FRAME_tot_count_BPSK_me, FRAME_TOT_count_BPSK_me,
	//	FRAME_tot_count_QPSK_me, FRAME_TOT_count_QPSK_me);
	printf("  frame RX count|%6i|%6i|%6i|%6i|\r\n", FRAME_TOT_count_BPSK_hub, FRAME_TOT_count_QPSK_hub, FRAME_TOT_count_BPSK_me, FRAME_TOT_count_QPSK_me);
	printf("  frame RX rate | %3i/s| %3i/s| %3i/s| %3i/s|\r\n", FRAME_tot_count_BPSK_hub, FRAME_tot_count_QPSK_hub, FRAME_tot_count_BPSK_me, FRAME_tot_count_QPSK_me);
	printf("----------------+------+------+------+------+\r\n");
	printf("  Frame err rate|%6.2f|%6.2f|%6.2f|%6.2f|\r\n", err_hub_BPSK, err_hub_QPSK, err_me_BPSK, err_me_QPSK);
	FRAME_err_count_BPSK_hub = 0;
	FRAME_tot_count_BPSK_hub = 0;
	FRAME_err_count_QPSK_hub = 0;
	FRAME_tot_count_QPSK_hub = 0;
	FRAME_err_count_BPSK_me = 0;
	FRAME_tot_count_BPSK_me = 0;
	FRAME_err_count_QPSK_me = 0;
	FRAME_tot_count_QPSK_me = 0;
	FRAME_TX_tot_BPSK = 0;
	FRAME_TX_tot_QPSK = 0;
	return 11;
} */

int NPRV_HMI_print_radio_stats(void) {
	char err_me_BPSK[8], err_me_QPSK[8], err_hub_BPSK[8], err_hub_QPSK[8];
	char BER_me_BPSK[8], BER_me_QPSK[8], BER_hub_BPSK[8], BER_hub_QPSK[8];
	char TA_text[18];
	char ampl_TXT[18];
	NPRV_HMI_mutex_stats_err.lock();
	if ((FRAME_tot_count_BPSK_me[0] + FRAME_tot_count_BPSK_me[1]) > 0) {
		//err_me_BPSK = 100 * (float)FRAME_err_count_BPSK_me / (float)FRAME_tot_count_BPSK_me;
		sprintf(err_me_BPSK, "%6.2f", (100 * (float)(FRAME_err_count_BPSK_me[0] + FRAME_err_count_BPSK_me[1]) / (float)(FRAME_tot_count_BPSK_me[0] + FRAME_tot_count_BPSK_me[1])));
	}
	else { sprintf(err_me_BPSK, "  --  "); }
	if ((FRAME_tot_count_QPSK_me[0] + FRAME_tot_count_QPSK_me[1]) > 0) {
		//err_me_QPSK = 100 * (float)FRAME_err_count_QPSK_me / (float)FRAME_tot_count_QPSK_me;
		//sprintf(err_me_QPSK, "%6.2f", (100 * (float)FRAME_err_count_QPSK_me / (float)FRAME_tot_count_QPSK_me));
		sprintf(err_me_QPSK, "%6.2f", (100 * (float)(FRAME_err_count_QPSK_me[0] + FRAME_err_count_QPSK_me[1]) / (float)(FRAME_tot_count_QPSK_me[0] + FRAME_tot_count_QPSK_me[1])));
	}
	else { sprintf(err_me_QPSK, "  --  "); }

	if ((FRAME_tot_count_BPSK_hub[0] + FRAME_tot_count_BPSK_hub[1]) > 0) {
		//err_hub_BPSK = 100 * (float)FRAME_err_count_BPSK_hub / (float)FRAME_tot_count_BPSK_hub;
		//sprintf(err_hub_BPSK, "%6.2f", (100 * (float)FRAME_err_count_BPSK_hub / (float)FRAME_tot_count_BPSK_hub));
		sprintf(err_hub_BPSK, "%6.2f", (100 * (float)(FRAME_err_count_BPSK_hub[0] + FRAME_err_count_BPSK_hub[1]) / (float)(FRAME_tot_count_BPSK_hub[0] + FRAME_tot_count_BPSK_hub[1])));
	}
	else { sprintf(err_hub_BPSK, "  --  "); }
	if ((FRAME_tot_count_QPSK_hub[0] + FRAME_tot_count_QPSK_hub[1]) > 0) {
		//err_hub_QPSK = 100 * (float)FRAME_err_count_QPSK_hub / (float)FRAME_tot_count_QPSK_hub;
		//sprintf(err_hub_QPSK, "%6.2f", (100 * (float)FRAME_err_count_QPSK_hub / (float)FRAME_tot_count_QPSK_hub));
		sprintf(err_hub_QPSK, "%6.2f", (100 * (float)(FRAME_err_count_QPSK_hub[0] + FRAME_err_count_QPSK_hub[1]) / (float)(FRAME_tot_count_QPSK_hub[0] + FRAME_tot_count_QPSK_hub[1])));
	}
	else { sprintf(err_hub_QPSK, "  --  "); }
	//BER_err_count_QPSK_me = 0;
	//static int BER_tot_word_QPSK_me
	//BER_me_BPSK
	if ((BER_tot_word_BPSK_me[0] + BER_tot_word_BPSK_me[1])> 0) {
		//sprintf(BER_me_BPSK, "%6.2f", (100 * (float)BER_err_count_BPSK_me / (float)BER_tot_word_BPSK_me));
		sprintf(BER_me_BPSK, "%6.2f", (100 * (float)(BER_err_count_BPSK_me[0] + BER_err_count_BPSK_me[1]) / (float)(BER_tot_word_BPSK_me[0] + BER_tot_word_BPSK_me[1])));
	} else { sprintf(BER_me_BPSK, "  --  "); }
	if ((BER_tot_word_QPSK_me[0] + BER_tot_word_QPSK_me[1]) > 0) {
		//sprintf(BER_me_QPSK, "%6.2f", (100 * (float)BER_err_count_QPSK_me / (float)BER_tot_word_QPSK_me));
		sprintf(BER_me_QPSK, "%6.2f", (100 * (float)(BER_err_count_QPSK_me[0] + BER_err_count_QPSK_me[1]) / (float)(BER_tot_word_QPSK_me[0] + BER_tot_word_QPSK_me[1])));
	}
	else { sprintf(BER_me_QPSK, "  --  "); }
	if ((BER_tot_word_BPSK_hub[0] + BER_tot_word_BPSK_hub[1]) > 0) {
		//sprintf(BER_hub_BPSK, "%6.2f", (100 * (float)BER_err_count_BPSK_hub / (float)BER_tot_word_BPSK_hub));
		sprintf(BER_hub_BPSK, "%6.2f", (100 * (float)(BER_err_count_BPSK_hub[0] + BER_err_count_BPSK_hub[1]) / (float)(BER_tot_word_BPSK_hub[0] + BER_tot_word_BPSK_hub[1])));
	}
	else { sprintf(BER_hub_BPSK, "  --  "); }
	if ((BER_tot_word_QPSK_hub[0] + BER_tot_word_QPSK_hub[1]) > 0) {
		//sprintf(BER_hub_QPSK, "%6.2f", (100 * (float)BER_err_count_QPSK_hub / (float)BER_tot_word_QPSK_hub));
		sprintf(BER_hub_QPSK, "%6.2f", (100 * (float)(BER_err_count_QPSK_hub[0] + BER_err_count_QPSK_hub[1]) / (float)(BER_tot_word_QPSK_hub[0] + BER_tot_word_QPSK_hub[1])));
	}
	else { sprintf(BER_hub_QPSK, "  --  "); }
	//TA_text G_NPRV_my_TimeAdvance
	if ((G_NPRV_my_TimeAdvance > 10000) && (G_NPRV_my_TimeAdvance < 2000000)) {
		sprintf(TA_text, "  %8.3f ms", (float)G_NPRV_my_TimeAdvance / 1000);
	}
	else { sprintf(TA_text,"      ??     "); } 
	if (NPRV_HMI_ampl_RX_me != 0) {
		sprintf(ampl_TXT, "   %7.1f   ", NPRV_HMI_ampl_RX_me);
	} 
	else {sprintf(ampl_TXT,"      ??     "); }
	printf("                 +-------------+-------------+\r\n");
	printf("                 |  from Hub   |   from ME   |\r\n");
	printf("-----------------+-------------+-------------+\r\n");
	printf("       SDR-Gain  |  RX:%5.1f dB|  TX:%5.1f dB|\r\n", NPRV_current_RX_gain, NPRV_current_TX_gain);
	//printf("  amplitude RX  |   %7.1f   |   %7.1f   |\r\n", NPRV_HMI_ampl_RX_hub, NPRV_HMI_ampl_RX_me);
	printf("   amplitude RX  |   %7.1f   |%s|\r\n", NPRV_HMI_ampl_RX_hub, ampl_TXT);
	//G_NPRV_my_TimeAdvance
	printf("  Timing-Advance |             |%s|\r\n", TA_text);
	printf(" Freq Offset(MHz)| RX:%8.3f | TX:%8.3f |\r\n", (NPRV_RX_LO_freq - NPRV_freq_offset_RX_tuned)/1e6, NPRV_freq_offset_TX_tuned/1e6);
	printf("-----------------+------+------+------+------+\r\n");
	printf("                 | BPSK | QPSK | BPSK | QPSK |\r\n");
	printf("-----------------+------+------+------+------+\r\n");
	printf("   frame TX count|  --  |  --  |%6i|%6i|\r\n", FRAME_TX_TOT_BPSK, FRAME_TX_TOT_QPSK);
	printf("   frame TX rate |  --  |  --  | %3i/s| %3i/s|\r\n", 
	//printf("   frame TX rate |  --  |  --  |%4.1f/s|%4.1f/s|\r\n",
		(FRAME_TX_tot_BPSK[0] + FRAME_TX_tot_BPSK[1]) / 2, 
		(FRAME_TX_tot_QPSK[0] + FRAME_TX_tot_QPSK[1]) / 2);
	//printf("frame RX count |%3i/s%6i|%3i/s%6i|%3i/s%6i|%3i/s%6i|\r\n", 
	//	FRAME_tot_count_BPSK_hub, FRAME_TOT_count_BPSK_hub, 
	//	FRAME_tot_count_QPSK_hub, FRAME_TOT_count_QPSK_hub,
	//	FRAME_tot_count_BPSK_me, FRAME_TOT_count_BPSK_me,
	//	FRAME_tot_count_QPSK_me, FRAME_TOT_count_QPSK_me);
	printf("   frame RX count|%6i|%6i|%6i|%6i|\r\n", FRAME_TOT_count_BPSK_hub, FRAME_TOT_count_QPSK_hub, FRAME_TOT_count_BPSK_me, FRAME_TOT_count_QPSK_me);
	printf("   frame RX rate | %3i/s| %3i/s| %3i/s| %3i/s|\r\n", 
		(FRAME_tot_count_BPSK_hub[0] + FRAME_tot_count_BPSK_hub[1]) / 2,
		(FRAME_tot_count_QPSK_hub[0] + FRAME_tot_count_QPSK_hub[1]) / 2,
		(FRAME_tot_count_BPSK_me[0] + FRAME_tot_count_BPSK_me[1]) / 2,
		(FRAME_tot_count_QPSK_me[0] + FRAME_tot_count_QPSK_me[1]) / 2);
	printf("-----------------+------+------+------+------+\r\n");
	//printf("  Frame err rate|%6.2f|%6.2f|%6.2f|%6.2f|\r\n", err_hub_BPSK, err_hub_QPSK, err_me_BPSK, err_me_QPSK);
	printf(" Word err rate(%%)|%s|%s|%s|%s|\r\n", BER_hub_BPSK, BER_hub_QPSK, BER_me_BPSK, BER_me_QPSK);
	printf("Frame err rate(%%)|%s|%s|%s|%s|\r\n", err_hub_BPSK, err_hub_QPSK, err_me_BPSK, err_me_QPSK);
	//FRAME_TX_TOT_BPSK, FRAME_TX_TOT_QPSK FRAME_TOT_count_BPSK_hub, FRAME_TOT_count_QPSK_hub, FRAME_TOT_count_BPSK_me, FRAME_TOT_count_QPSK_me
	if (FRAME_TX_TOT_BPSK > 900000) { FRAME_TX_TOT_BPSK = 0; }
	if (FRAME_TX_TOT_QPSK > 900000) { FRAME_TX_TOT_QPSK = 0; }
	if (FRAME_TOT_count_BPSK_hub > 900000) { FRAME_TOT_count_BPSK_hub = 0; }
	if (FRAME_TOT_count_QPSK_hub > 900000) { FRAME_TOT_count_QPSK_hub = 0; }
	if (FRAME_TOT_count_BPSK_me > 900000) { FRAME_TOT_count_BPSK_me = 0; }
	if (FRAME_TOT_count_QPSK_me > 900000) { FRAME_TOT_count_QPSK_me = 0; }
	HMI_TOG = !HMI_TOG;
	//printf("HMI TOG:%i\r\n", HMI_TOG);
	FRAME_err_count_BPSK_hub[HMI_TOG] = 0;
	FRAME_tot_count_BPSK_hub[HMI_TOG] = 0;
	FRAME_err_count_QPSK_hub[HMI_TOG] = 0;
	FRAME_tot_count_QPSK_hub[HMI_TOG] = 0;
	FRAME_err_count_BPSK_me[HMI_TOG] = 0;
	FRAME_tot_count_BPSK_me[HMI_TOG] = 0;
	FRAME_err_count_QPSK_me[HMI_TOG] = 0;
	FRAME_tot_count_QPSK_me[HMI_TOG] = 0;
	FRAME_TX_tot_BPSK[HMI_TOG] = 0;
	FRAME_TX_tot_QPSK[HMI_TOG] = 0;
	BER_err_count_BPSK_hub[HMI_TOG] = 0;
	BER_tot_word_BPSK_hub[HMI_TOG] = 0;
	BER_err_count_QPSK_hub[HMI_TOG] = 0;
	BER_tot_word_QPSK_hub[HMI_TOG] = 0;
	BER_err_count_BPSK_me[HMI_TOG] = 0;
	BER_tot_word_BPSK_me[HMI_TOG] = 0;
	BER_err_count_QPSK_me[HMI_TOG] = 0;
	BER_tot_word_QPSK_me[HMI_TOG] = 0;

	NPRV_HMI_mutex_stats_err.unlock();
	return 17;
}

int NPRV_HMI_print_who(void) {
	int i;
	int count_txt_lines = 1;
	uint32_t last_IP;
	unsigned char IP_char[6];
	char* loc_callsign;
	printf("WHO:\r\n");
	if ((CONF_is_NPRV_HUB == 0) && (G_NPRV_MyRadioState == 2)) {
		IP_int2char(LAN_conf_applied.DHCP_range_start, IP_char);
		printf("ME %i \"%s\" IP: %u.%u.%u.%u to ", G_NPRV_MyClientID, G_NPRV_my_Callsign, IP_char[0], IP_char[1], IP_char[2], IP_char[3]);
		IP_int2char(LAN_conf_applied.DHCP_range_start + LAN_conf_applied.DHCP_range_size - 1, IP_char);
		printf("%u.%u.%u.%u            \r\n", IP_char[0], IP_char[1], IP_char[2], IP_char[3]);
		count_txt_lines++;
	}
	for (i = 0; i < NPRV_MAX_CLIENT_NB; i++) {
		if ((CONF_radio_addr_table_status[i] == 1) && 
			((CONF_is_NPRV_HUB == 1) || (i != G_NPRV_MyClientID)))//2023_01_07
		{
			//CONF_client_callsigns_table_hub RX_who_callsigns
			if (CONF_is_NPRV_HUB == 1) {
				loc_callsign = CONF_client_callsigns_table_hub[i];
			} else {
				loc_callsign = RX_who_callsigns[i];
			}
			//if (CONF_is_NPRV_HUB == 1) {
			if (1) {
				IP_int2char(CONF_radio_addr_table_IP_begin[i], IP_char);
				//loc_callsign = CONF_client_callsigns_table_hub[i];
			}
			else {
				IP_int2char(RX_who_IP_range_start[i], IP_char);
				//loc_callsign = RX_who_callsigns[i];
			}
			printf("  %i \"%s\" IP: %u.%u.%u.%u to ", i, loc_callsign, IP_char[0], IP_char[1], IP_char[2], IP_char[3]);
			//if (CONF_is_NPRV_HUB) {
			if (1) {
				last_IP = CONF_radio_addr_table_IP_begin[i] + CONF_radio_addr_table_IP_size[i] - 1;
			}
			else {
				last_IP = RX_who_IP_range_start[i] + RX_who_IP_range_size[i] - 1;
			}
			IP_int2char(last_IP, IP_char);
			printf("%u.%u.%u.%u Static:%i  \r\n", IP_char[0], IP_char[1], IP_char[2], IP_char[3], RX_who_static_IP[i]);
			count_txt_lines++;
		}
	}
	return count_txt_lines;
}