// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include <SFML/System.hpp>
#include <SoapySDR/Device.h>
#include <SoapySDR/Formats.h>
#include <SoapySDR/Version.h>
#include <math.h>
#include "NPRV_SDR_if.h"
#include "NPRV_mod_demod.h"
#include "NPRV_coder_decoder.h"
#include "NPRV_L1_L2_radio.h"
#include "Eth_IPv4.h"
#include "DHCP_ARP.h"
#include "NPRV_HMI.h"
#include "NPRV_signaling.h"
#include "NPRV_NVRAM.h"
#include "NPRV_config.h"
#include "NPRV_display.h"

#ifdef NPRV_linux 
#include <signal.h>
#endif

#ifdef NPRV_windows
#include <windows.h>
#include <stdio.h>
#endif

#pragma warning(disable: 4996)

#define M_PI 3.14159265358979



#define NPRV_INI_WAIT_TIME 2500 /*2500*/

//#define NPRV_SDR_DISPLAY_DATA

static float NPRV_SDR_gain_TX_mini;
static float NPRV_SDR_gain_TX_maxi;
static float NPRV_SDR_gain_RX_mini;
static float NPRV_SDR_gain_RX_maxi;

static short int my_buff_rx[4][70000] = { 0 };//short
static short int my_buff_tx[70000] = { 0 };//short
static short int my_buff_filtered[70000] = { 0 };
static float data_for_display[270000] = { 0 };
SoapySDRDevice* NPRV_glob_sdr;
SoapySDRStream* NPRV_SDR_rxStream;
SoapySDRStream* NPRV_SDR_txStream;

static float delta_T_ReadStream_HMI_mean;
static float delta_T_ReadStream_HMI_max;
static int count_read_stream = 0;
static float delta_T_Demod_HMI_mean;
static float delta_T_Demod_HMI_max;
static int count_demod = 0;
static float delta_T_WriteStream_HMI_mean;
static float delta_T_WriteStream_HMI_max;
static int count_write_stream = 0;
static float delta_T_main_HMI_mean;
static float delta_T_main_HMI_max;
static int count_main = 0;

//static float NPRV_freq_offset_RX_tuned = 0;
//static float NPRV_freq_offset_TX_tuned = 0;

float G_NPRV_my_amplitude_client = 0;//migre sdr if
float G_NPRV_amplitude_ref_hub = 0;//migre sdr if
int G_NPRV_RX_gain_stabilised = 0;//migre sdr if
float GLOB_ampl_stat_per_ID[NPRV_MAX_CLIENT_NB] = { 0 };//migre sdr if

static int Kill_requested = 0;

static FILE* FID_out_SDR_save_raw = 0;
static FILE* FID_out_SDR_save_filtered = 0;
static FILE* FID_out_SDR_save_data_display = 0;

void termination_handler(int signum)
{
	Kill_requested = 1;
}

#ifdef NPRV_windows
BOOL WINAPI CtrlHandler(DWORD fdwCtrlType)
{
	switch (fdwCtrlType)
	{
		// Handle the CTRL-C signal.
	case CTRL_C_EVENT:
		printf("Ctrl-C event\n\n");
		Kill_requested = 1;
		return TRUE;

		// CTRL-CLOSE: confirm that the user wants to exit.
	case CTRL_CLOSE_EVENT:
		printf("Ctrl-Close event\n\n");
		Kill_requested = 1;
		return TRUE;

		// Pass other signals to the next handler.
	case CTRL_BREAK_EVENT:
		printf("Ctrl-Break event\n\n");
		Kill_requested = 1;
		return FALSE;

	case CTRL_LOGOFF_EVENT:
		printf("Ctrl-Logoff event\n\n");
		Kill_requested = 1;
		return FALSE;

	case CTRL_SHUTDOWN_EVENT:
		printf("Ctrl-Shutdown event\n\n");
		Kill_requested = 1;
		return FALSE;

	default:
		return FALSE;
	}
}
#endif

int NPRV_SDR_open(void) {
	size_t length;

	//enumerate devices
	SoapySDRKwargs *results = SoapySDRDevice_enumerate(NULL, &length);
	printf("c'est moi v47\r\n");
	for (size_t i = 0; i < length; i++)
	{
		printf("Found device #%d: ", (int)i);
		for (size_t j = 0; j < results[i].size; j++)
		{
			printf("%s=%s, ", results[i].keys[j], results[i].vals[j]);
		}
		printf("\n");
	}
	SoapySDRKwargsList_clear(results, length);

	SoapySDRKwargs args = {};
	if (CONF_SDR_HW_type == 0) {
		SoapySDRKwargs_set(&args, "driver", "plutosdr");//rtlsdr plutosdr lime
	}
	if (CONF_SDR_HW_type == 1) {
		SoapySDRKwargs_set(&args, "driver", "lime");//rtlsdr plutosdr lime
		//bufferLength
		//SoapySDRKwargs_set(&args, "bufferLength", "6000");//rtlsdr plutosdr lime
	}
	//SoapySDRKwargs_set(&args, "hostname", "192.168.2.3");//rtlsdr plutosdr
	SoapySDRKwargs_set(&args, "hostname", CONF_SDR_PLUTO_IP);//rtlsdr plutosdr
	//SoapySDRKwargs_set(&args, "remote", "192.168.0.24:1234");//remote
	//SoapySDRKwargs_set(&args, "soapy", "0");//remote
	//SoapySDRKwargs_set(&args, "driver", "remote");//remote
	//SoapySDRKwargs_set(&args, "remote:driver", "plutosdr");//remote
	
	NPRV_glob_sdr = SoapySDRDevice_make(&args);
	SoapySDRKwargs_clear(&args);

	if (NPRV_glob_sdr == NULL)
	{
		printf("SoapySDRDevice_make fail: %s\n", SoapySDRDevice_lastError());
		return EXIT_FAILURE;
	}
}

void NPRV_SDR_close(void) {
	SoapySDRDevice_unmake(NPRV_glob_sdr);
}

int NPRV_SDR_configure(void) {
	int answer;
	float answ_float;
	
	size_t length;
	//query device info
	char** names = SoapySDRDevice_listAntennas(NPRV_glob_sdr, SOAPY_SDR_RX, 0, &length);
	printf("Rx antennas: ");
	for (size_t i = 0; i < length; i++) printf("%s, ", names[i]);
	printf("\n");
	SoapySDRStrings_clear(&names, length);

	SoapySDRArgInfo* loc_arginfo_types = SoapySDRDevice_getStreamArgsInfo(NPRV_glob_sdr, SOAPY_SDR_TX, 0, &length);
	printf("\r\nlength:%i %i\r\n\r\n", length, loc_arginfo_types->numOptions);

	names = SoapySDRDevice_listGains(NPRV_glob_sdr, SOAPY_SDR_RX, 0, &length);
	printf("Rx gains: ");
	for (size_t i = 0; i < length; i++) printf("%s, ", names[i]);
	printf("\n");
	SoapySDRStrings_clear(&names, length);

	names = SoapySDRDevice_listGains(NPRV_glob_sdr, SOAPY_SDR_TX, 0, &length);
	printf("Tx gains: ");
	for (size_t i = 0; i < length; i++) printf("%s, ", names[i]);
	printf("\n");
	SoapySDRStrings_clear(&names, length);

	//answer = SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 2);
	answ_float = SoapySDRDevice_getGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0);
	printf("Gain TX: %f\r\n", answ_float);

	names = SoapySDRDevice_listSensors(NPRV_glob_sdr, &length);
	printf("List sensors: ");
	for (size_t i = 0; i < length; i++) printf("%s, ", names[i]);
	printf("\n");
	SoapySDRStrings_clear(&names, length);
	
	answer = SoapySDRDevice_setGainMode(NPRV_glob_sdr, SOAPY_SDR_RX, 0, 0);
	printf("j'ai r�ussi � r�gler l'AGC :%i\r\n", answer);
	
	answer = SoapySDRDevice_getGainMode(NPRV_glob_sdr, SOAPY_SDR_RX, 0);
	printf("Gain Mode: %i \r\n", answer);
	answer = SoapySDRDevice_getGainMode(NPRV_glob_sdr, SOAPY_SDR_RX, 0);
	answer = SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_RX, 0, 34);//60 52
	
	//answer = SoapySDRDevice_hasDCOffsetMode(NPRV_glob_sdr, SOAPY_SDR_RX, 0);
	//answer = SoapySDRDevice_hasIQBalance(NPRV_glob_sdr, SOAPY_SDR_RX, 0);
	//printf("HAS DC OFFSET MODE : %i\r\n", answer); 

	SoapySDRRange *ranges = SoapySDRDevice_getFrequencyRange(NPRV_glob_sdr, SOAPY_SDR_RX, 0, &length);
	printf("Rx freq ranges: ");
	for (size_t i = 0; i < length; i++) printf("[%g Hz -> %g Hz], ", ranges[i].minimum, ranges[i].maximum);
	printf("\n");
	//free(ranges);

	*ranges = SoapySDRDevice_getGainRange(NPRV_glob_sdr, SOAPY_SDR_TX, 0);
	printf("TX gain ranges: ");
	for (size_t i = 0; i < length; i++) printf("[%g  -> %g ], ", ranges[i].minimum, ranges[i].maximum);
	printf("\n");
	//free(ranges);

	*ranges = SoapySDRDevice_getGainRange(NPRV_glob_sdr, SOAPY_SDR_RX, 0);
	printf("RX gain ranges: ");
	for (size_t i = 0; i < length; i++) printf("[%g  -> %g ], ", ranges[i].minimum, ranges[i].maximum);
	printf("\n");
	//free(ranges);

	if (SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_TX, 0, CONF_center_freq_TX + 30e6, NULL) != 0)//1245.7e6
	{
		printf("setFrequency fail: %s\n", SoapySDRDevice_lastError());
	}
	//sf::sleep(sf::milliseconds(2500));
	if (SoapySDRDevice_setSampleRate(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 1e6) != 0)//1e6
	{
		printf("setSampleRate fail: %s\n", SoapySDRDevice_lastError());
	}
	//sf::sleep(sf::milliseconds(2500));
	if (SoapySDRDevice_setSampleRate(NPRV_glob_sdr, SOAPY_SDR_RX, 0, 1e6) != 0)//1e6
	{
		printf("setSampleRate fail: %s\n", SoapySDRDevice_lastError());
	}
	//sf::sleep(sf::milliseconds(2500));
	if (SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_TX, 0, CONF_center_freq_TX + NPRV_freq_offset_TX_tuned - NPRV_channels_freq[0]*1e6, NULL) != 0)//1245.7e6
	{
		printf("setFrequency fail: %s\n", SoapySDRDevice_lastError());
	}
	if (SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_RX, 0, CONF_center_freq_RX + NPRV_freq_offset_RX_tuned - NPRV_channels_freq[0] * 1e6, NULL) != 0)//437.0e6,
	{
		printf("setFrequency fail: %s\n", SoapySDRDevice_lastError());
	}
	
	//answer = SoapySDRDevice_writeSetting(NPRV_glob_sdr, "quadrature_tracking_aen", "1");
	//printf("arbitrary setting quadrature. Success: %i\r\n", answer);
	return 1;
}

void NPRV_SDR_re_configure(void) {
	if (CONF_TX_test >= 1) {
		NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; //max power
		SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, NPRV_current_TX_gain);
	}
	SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_TX, 0, CONF_center_freq_TX + NPRV_freq_offset_TX_tuned - NPRV_channels_freq[0] * 1e6, NULL);
	//printf("\r\nFREQ TX: %f = %f+%f\r\n\r\n", CONF_center_freq_TX + NPRV_freq_offset_TX_tuned, CONF_center_freq_TX , NPRV_freq_offset_TX_tuned);
	SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_RX, 0, CONF_center_freq_RX + NPRV_freq_offset_RX_tuned - NPRV_channels_freq[0] * 1e6, NULL);
}

void NPRV_SDR_reconfigure_RX_Freq(void) {
	if (CONF_SDR_HW_type != 1) {// !!! 2023_01_23 inhibition for LIME because of task slipping
		SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_RX, 0, CONF_center_freq_RX + NPRV_freq_offset_RX_tuned - NPRV_channels_freq[0] * 1e6, NULL);
	}
}

void NPRV_SDR_reconfigure_TX_Freq(void) {
	if (CONF_SDR_HW_type != 1) {// !!! 2023_01_23 inhibition for LIME because of task slipping
		SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_TX, 0, CONF_center_freq_TX + NPRV_freq_offset_TX_tuned - NPRV_channels_freq[0] * 1e6, NULL);
	}
}

void NPRV_SDR_init_gain(int set_gain_txrx) {
	if (CONF_SDR_HW_type == 0) {//Adalm Pluto
		NPRV_SDR_gain_TX_mini = 0;
		NPRV_SDR_gain_TX_maxi = 89;
		NPRV_SDR_gain_RX_mini = 0;
		NPRV_SDR_gain_RX_maxi = 73;
	}
	if (CONF_SDR_HW_type == 1) {//Lime SDR mini
		NPRV_SDR_gain_TX_mini = -12;
		NPRV_SDR_gain_TX_maxi = 64;
		NPRV_SDR_gain_RX_mini = -12;
		NPRV_SDR_gain_RX_maxi = 61;
	}

	if (CONF_TX_test > 0) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }
	if ((CONF_is_NPRV_HUB == 0) && (set_gain_txrx == 1)) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }//no previous NVRAM value
	if (NPRV_current_TX_gain > NPRV_SDR_gain_TX_maxi) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_maxi; }
	if (NPRV_current_TX_gain < NPRV_SDR_gain_TX_mini) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_mini; }
}

int NPRV_SDR_start_stream(void) {
	size_t TX_stream_MTU;
	SoapySDRKwargs args = {};

	SoapySDRKwargs_set(&args, "bufferLength", "65536");
#if (SOAPY_SDR_API_VERSION & 0x00FF0000) == 0x00080000
	printf("SOAPY API VERSION 8.x\r\n");
	NPRV_SDR_rxStream = SoapySDRDevice_setupStream(NPRV_glob_sdr, SOAPY_SDR_RX, SOAPY_SDR_CS16, NULL, 0, &args);//CF32 CS16
	NPRV_SDR_txStream = SoapySDRDevice_setupStream(NPRV_glob_sdr, SOAPY_SDR_TX, SOAPY_SDR_CS16, NULL, 0, &args);//CF32
#endif
#if (SOAPY_SDR_API_VERSION & 0x00FF0000) == 0x00070000
	printf("SOAPY API VERSION 7.x\r\n");
	SoapySDRDevice_setupStream(NPRV_glob_sdr, &NPRV_SDR_rxStream, SOAPY_SDR_RX, SOAPY_SDR_CS16, NULL, 0, &args);
	SoapySDRDevice_setupStream(NPRV_glob_sdr, &NPRV_SDR_txStream, SOAPY_SDR_TX, SOAPY_SDR_CS16, NULL, 0, &args);
#endif
	SoapySDRKwargs_clear(&args);

	if (NPRV_SDR_rxStream == NULL)
	{
		printf("setupStream fail: %s\n", SoapySDRDevice_lastError());
		SoapySDRDevice_unmake(NPRV_glob_sdr);
		return EXIT_FAILURE;
	}

	TX_stream_MTU = SoapySDRDevice_getStreamMTU(NPRV_glob_sdr, NPRV_SDR_txStream);
	printf("TX stream MTU : %i\r\n", TX_stream_MTU);

	TX_stream_MTU = SoapySDRDevice_getStreamMTU(NPRV_glob_sdr, NPRV_SDR_rxStream);
	printf("RX stream MTU : %i\r\n", TX_stream_MTU);

	SoapySDRDevice_activateStream(NPRV_glob_sdr, NPRV_SDR_rxStream, SOAPY_SDR_HAS_TIME & SOAPY_SDR_RX, 0, 0); //start streaming
	SoapySDRDevice_activateStream(NPRV_glob_sdr, NPRV_SDR_txStream, SOAPY_SDR_HAS_TIME & SOAPY_SDR_TX, 0, 0); //start streaming
	return 1;
}

void NPRV_SDR_stop_stream(void) {
	SoapySDRDevice_deactivateStream(NPRV_glob_sdr, NPRV_SDR_rxStream, 0, 0); //stop streaming
	SoapySDRDevice_deactivateStream(NPRV_glob_sdr, NPRV_SDR_txStream, 0, 0); //stop streaming
	SoapySDRDevice_closeStream(NPRV_glob_sdr, NPRV_SDR_rxStream);
	SoapySDRDevice_closeStream(NPRV_glob_sdr, NPRV_SDR_txStream);
}

int NPRV_SDR_read_stream(void* buffer_out, size_t length) {
	void *buffs_rx[] = { buffer_out };
	int flags = SOAPY_SDR_RX & SOAPY_SDR_HAS_TIME;
	long long int timeNs;
	return SoapySDRDevice_readStream(NPRV_glob_sdr, NPRV_SDR_rxStream, buffs_rx, length, &flags, &timeNs, 100000);//1000000
}

int NPRV_SDR_write_stream(void* buffer_in, size_t length) {
	void *buffs_tx[] = { buffer_in }; //array of buffers
	int flags = SOAPY_SDR_TX & SOAPY_SDR_HAS_TIME;
	return SoapySDRDevice_writeStream(NPRV_glob_sdr, NPRV_SDR_txStream, buffs_tx, length, &flags, 0, 100000);
}

void NPRV_SDR_calib_IQ(void) {
	FILE* fid_loc_out;
	int i, j, k;
	int size_loc = 32768;
	float ampl_loc, ampl_loc_max, loc_I, loc_Q, ampl_mean;
	NPRV_SDR_open();
	NPRV_SDR_configure();
	NPRV_SDR_start_stream();
	fid_loc_out = fopen("user_data/SDR_samples.cs16", "wb");
	
	ampl_mean = 0;
	for (k = -50; k < 500; k = k + 10) {
		if (k == -50) { SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 1045.7e6, NULL); }
		if (k == -40) { SoapySDRDevice_setFrequency(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 1245.7e6, NULL); }
		for (j = 0; j < size_loc; j++) {
			my_buff_tx[2 * j] = 20;//20
			my_buff_tx[2 * j + 1] = -30;//-30
		}
		printf("\r\nI INPUT = %i\r\n", k);
		for (i = 0; i < 90; i++) {

			NPRV_SDR_read_stream(my_buff_rx[0], size_loc);
			fwrite(my_buff_rx[0], sizeof(short int), size_loc * 2, fid_loc_out);
			NPRV_SDR_write_stream(my_buff_tx, size_loc);
			ampl_loc_max = 0;
			for (j = 0; j < size_loc; j++) {
				loc_I = my_buff_rx[0][2 * j];
				loc_Q = my_buff_rx[0][2 * j + 1];
				ampl_loc = sqrt(loc_I*loc_I + loc_Q * loc_Q);
				if (ampl_loc > ampl_loc_max) { ampl_loc_max = ampl_loc; }

			}
			ampl_mean = ampl_mean + ampl_loc_max;
			//printf("step:%i ampl:%f\r\n", i, ampl_loc_max);
		}
		printf("AMPL MEAN: %f\r\n", ampl_mean / 90);
		ampl_mean = 0;
	}
	
	NPRV_SDR_stop_stream();
	NPRV_SDR_close();
	fclose(fid_loc_out);
}


void NPRV_store_ampl_RX_frame(unsigned int TX_ID, int channel, float amplitude_in) {
	const float normalisation_chan_type[20] = { 1, 1.4142, 0.7071 };//square root of 2
	int chan_type;
	float amplitude2;
	chan_type = NPRV_channels_SR_std[channel];
	amplitude2 = amplitude_in * normalisation_chan_type[chan_type];
	
	if (TX_ID < NPRV_MAX_CLIENT_NB) { GLOB_ampl_stat_per_ID[TX_ID] = 0.1 * (float)(amplitude2)+0.9*GLOB_ampl_stat_per_ID[TX_ID]; }
	if ((CONF_is_NPRV_HUB == 0) && 
		(TX_ID == 0) && 
		(G_NPRV_RX_gain_stabilised > 1)) 
	{
		if (G_NPRV_amplitude_ref_hub == 0) { G_NPRV_amplitude_ref_hub = amplitude2; }
		else { G_NPRV_amplitude_ref_hub = (0.1 * amplitude2) + (0.9 * G_NPRV_amplitude_ref_hub); }
	}
	if ((CONF_is_NPRV_HUB == 0) && 
		(TX_ID == G_NPRV_MyClientID) && 
		(TX_ID < NPRV_MAX_CLIENT_NB) && (G_NPRV_RX_gain_stabilised > 1)) 
	{
		if (G_NPRV_my_amplitude_client == 0) { G_NPRV_my_amplitude_client = amplitude2; }
		else { G_NPRV_my_amplitude_client = (0.3 * amplitude2) + (0.7 * G_NPRV_my_amplitude_client); }
		NPRV_SDR_store_RX_ME_event();
	}
}

sf::Mutex mutex_main2demod[8][10]; //[step][chan]
sf::Mutex mutex_demod2main[8][10]; //[step][chan]

void NPRV_threaded_demod(int chan) {

	unsigned int step_demod = 3;
	int length = 32768;//!!! Generalise with global param

	sf::Time elapsed_time;
	int i = 0;
	int delta_T_loc = 0;
	int delta_T_max_loc = 0;
	int delta_T_30_loc = 0;
	int T_capture = 0;
	int T_capture_prev = 0;
	int T_capture_prev30 = 0;

	if (CONF_RX_SAVE_to_FILE == 1) {
		FID_out_SDR_save_filtered = fopen("./user_data/FILTERED_samples.cs16", "wb");//.cs16 fid_loc_out3 
#ifdef NPRV_SDR_DISPLAY_DATA
		FID_out_SDR_save_data_display = fopen("./user_data/data_for_display", "wb");//.cs16 fid_loc_out2
#endif
	}
	mutex_demod2main[0][chan].lock();
	mutex_demod2main[3][chan].lock();
	//}
	printf("Tread nb %i launched\r\n", chan);
	NPRV_init_display();
	sf::sleep(sf::milliseconds(NPRV_INI_WAIT_TIME));//500
	while (Kill_requested==0) {
		mutex_main2demod[step_demod % 4][chan].lock();
		//printf("Thread demod %i\r\n", step_demod % 4);
#ifdef NPRV_SDR_DISPLAY_DATA
		for (i = 0; i < 269000; i++) {
			data_for_display[i] = 0;
		}
#endif
		NPRV_demod_total(my_buff_rx[step_demod % 4], my_buff_filtered, (data_for_display + 128), length);
		NPRV_SDR_autoGain_RX();

		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if ((delta_T_loc > 0) && (delta_T_loc < 15)) {
			sf::sleep(sf::milliseconds(25 - delta_T_loc));
		}
		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if (delta_T_loc > delta_T_max_loc) { delta_T_max_loc = delta_T_loc; }
		T_capture_prev = T_capture;
		if (CONF_RX_SAVE_to_FILE == 1) {
			if (FID_out_SDR_save_filtered != 0) {
				fwrite(my_buff_filtered, sizeof(short int), length * 2, FID_out_SDR_save_filtered);//32768 * 2 // fid_loc_out3
			}
#ifdef NPRV_SDR_DISPLAY_DATA
			if (FID_out_SDR_save_data_display != 0) {
				fwrite(data_for_display, sizeof(float), length * 8, FID_out_SDR_save_data_display);//32768 * 2 // fid_loc_out2
		}
#endif
		}
		if ((step_demod % 30) == 0) {
			delta_T_30_loc = T_capture - T_capture_prev30;
			T_capture_prev30 = T_capture;
			delta_T_Demod_HMI_mean = (float)delta_T_30_loc / 30;
			delta_T_Demod_HMI_max = delta_T_max_loc;
			delta_T_max_loc = 0;
			//printf("DELTA_T DEMOD %f Maxi %f\r\n", delta_T_Demod_HMI_mean, delta_T_Demod_HMI_max);
		}
		mutex_main2demod[step_demod % 4][chan].unlock();
		mutex_demod2main[(step_demod + 0) % 4][chan].unlock(); //+1
		mutex_demod2main[(step_demod + 2) % 4][chan].lock();//+3
		step_demod++;
		count_demod++;
		
	} 
	mutex_demod2main[(step_demod + 1) % 4][chan].unlock();// unlock before exit
	NPRV_dipl_close_all_windows();
	if (CONF_RX_SAVE_to_FILE == 1) {
		if (FID_out_SDR_save_filtered != 0) {
			fclose(FID_out_SDR_save_filtered);
			FID_out_SDR_save_filtered = 0;
		}
#ifdef NPRV_SDR_DISPLAY_DATA
		//fclose(fid_loc_out2);
		if (FID_out_SDR_save_data_display != 0) {
			fclose(FID_out_SDR_save_data_display);
			FID_out_SDR_save_data_display = 0;
		}
#endif
	}
}

void NPRV_threaded_read_stream(void) {
	int length = 32768;//!!!Generalize with global param
	int i;
	sf::Time elapsed_time;
	int delta_T_loc = 0;
	int delta_T_max_loc = 0;
	int delta_T_30_loc = 0;
	int T_capture = 0;
	int T_capture_prev = 0;
	int T_capture_prev30 = 0;
	
	unsigned int step_read = 0;
	mutex_main2demod[0][0].lock();
	mutex_main2demod[1][0].lock();
	sf::sleep(sf::milliseconds(NPRV_INI_WAIT_TIME));
	while (Kill_requested==0) {
		mutex_demod2main[(step_read + 2) % 4][0].lock();
		//printf("Thread read %i\r\n", step_read % 4);
		NPRV_SDR_read_stream(my_buff_rx[step_read % 4], length);//32768

		mutex_demod2main[(step_read + 2) % 4][0].unlock();
		mutex_main2demod[step_read % 4][0].unlock();
		if ((step_read > 9) && (CONF_RX_SAVE_to_FILE == 1) && (FID_out_SDR_save_raw != 0)) {
			fwrite(my_buff_rx[step_read % 4], sizeof(short int), length * 2, FID_out_SDR_save_raw);//32768 * 2  fid_loc_out
		}
		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if ((delta_T_loc > 0) && (delta_T_loc < 15)) {
			sf::sleep(sf::milliseconds(25 - delta_T_loc));
		}
		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if (delta_T_loc > delta_T_max_loc) { delta_T_max_loc = delta_T_loc; }
		T_capture_prev = T_capture;
		if ((step_read % 30) == 0) {
			delta_T_30_loc = T_capture - T_capture_prev30;
			T_capture_prev30 = T_capture; 
			delta_T_ReadStream_HMI_mean = (float)delta_T_30_loc / 30;
			delta_T_ReadStream_HMI_max = delta_T_max_loc;
			delta_T_max_loc = 0;
			//printf("DELTA_T READ %f Maxi %f\r\n", delta_T_ReadStream_HMI_mean, delta_T_ReadStream_HMI_max);
		}
		mutex_main2demod[(step_read + 2) % 4][0].lock();
		step_read++;
		count_read_stream++;
	}
	mutex_main2demod[(step_read + 1) % 4][0].unlock();//+1 instead of +2 ; unlock before exit
}

void NPRV_threaded_ModWriteStream(void) {
	int length = 32768;//!!! Generalize with global param
	sf::Time elapsed_time;
	int i = 0;
	int j;
	int delta_T_loc = 0;
	int delta_T_max_loc = 0;
	int delta_T_30_loc = 0;
	int T_capture = 0;
	int T_capture_prev = 0;
	int T_capture_prev30 = 0;
	static int TX_need[20] = { 0 };
	int TX_need_synth = 0;
	int TX_need_synth_prev = 0;
	sf::sleep(sf::milliseconds(NPRV_INI_WAIT_TIME + 0));
	while (Kill_requested==0) {
		//printf("Thread write %i\r\n", i%4);
		if (CONF_NPRV_filter_TX == 1) {
			TX_need[i % 16] = NPRV_mod_total_2(my_buff_tx, length);
		}
		else {
			TX_need[i % 16] = NPRV_mod_total(my_buff_tx, length);
		}

		if ((TX_need[(i-3) % 16] == 1) || (TX_need[(i - 2) % 16] == 1)) {
			TX_need_synth = 1;
		}
		else {
			TX_need_synth = 0;
		}
		if ((TX_need_synth != TX_need_synth_prev) && (TX_need_synth == 1)) {
			//SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 75);
		}
		if ((TX_need_synth != TX_need_synth_prev) && (TX_need_synth == 0)) {
			//SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 45);
		}
		TX_need_synth_prev = TX_need_synth;
		NPRV_SDR_write_stream(my_buff_tx, length);//32768

		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if ((delta_T_loc > 0) && (delta_T_loc < 15)) {
			sf::sleep(sf::milliseconds(25 - delta_T_loc));
		}
		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if (delta_T_loc > delta_T_max_loc) { delta_T_max_loc = delta_T_loc; }
		T_capture_prev = T_capture;

		if ((i % 30) == 0) {
			delta_T_30_loc = T_capture - T_capture_prev30;
			T_capture_prev30 = T_capture;
			delta_T_WriteStream_HMI_mean = (float)delta_T_30_loc / 30;
			delta_T_WriteStream_HMI_max = delta_T_max_loc;
			delta_T_max_loc = 0;
			//printf("DELTA_T Write %f Maxi %f\r\n", delta_T_WriteStream_HMI_mean, delta_T_WriteStream_HMI_max);
		}
		i++;
		count_write_stream++;
	}
}

void NPRV_SDR_Start_stop_recording(void) {
	if (CONF_RX_SAVE_to_FILE == 1) {
		printf("\r\n\r\n*********** START RECORDING ***********\r\n");
		printf("FID before FOPEN:%X\r\n", FID_out_SDR_save_raw);
		FID_out_SDR_save_raw = fopen("user_data/SDR_samples.cs16", "wb");
		FID_out_SDR_save_filtered = fopen("./user_data/FILTERED_samples.cs16", "wb");//.cs16
		
#ifdef NPRV_SDR_DISPLAY_DATA
		FID_out_SDR_save_data_display = fopen("./user_data/data_for_display", "wb");//.cs16
#endif
	}
	if (CONF_RX_SAVE_to_FILE == 0) {
		printf("\r\n\r\n*********** STOP RECORDING ***********\r\n");
		fclose(FID_out_SDR_save_raw);
		FID_out_SDR_save_raw = 0; 
		fclose(FID_out_SDR_save_filtered);
		FID_out_SDR_save_filtered = 0;
#ifdef NPRV_SDR_DISPLAY_DATA
		fclose(FID_out_SDR_save_data_display);
#endif
	}
}

static int NPRV_do_not_overwrite_txt = 0;

void NPRV_HMI_new_TXT_message(void) {
	NPRV_do_not_overwrite_txt = 1;
}

int NPRV_SDR_main_loop(void) {
	
	float alpha = 0;
	unsigned int step_main = 0;
	int current_time, previous_time;
	int current_time2, previous_time2;
	double current_gain = 0; 
	double current_bandwidth = 0;
	int delta_t_loc, delta_t_max;
	int delta_t_mean, delta_t_count = 0, delta_t_tot = 0;
	sf::Time elapsed_time;
	int count_written = 0;
	int data_from_file = 0;
	int size_loc = 32768;
	int size_read = 0;
	int delta_T_loc = 0;
	int delta_T_max_loc = 0;
	int delta_T_30_loc = 0;
	int T_capture = 0;
	int T_capture_prev = 0;
	int T_capture_prev30 = 0;
	int T_capture_prev_TX_auto_gain = 0;
	static int TX_need[20] = { 0 };
	int TX_need_synth = 0;
	int TX_need_synth_prev = 0;
	float err_frame_BPSK, err_frame_QPSK;
	static int count_frame_prev[30];
	int count_txt_lines = 0;
	int count_txt_lines_prev = 0;
	int k;
	float BER_BPSK, BER_QPSK;
	int loc_error;
	//static char status_text[5][20] = { "NOT-CONNECTED", " ", "CONNECTED        " };
	sf::Thread thread_demod_0(&NPRV_threaded_demod, 0);

	sf::Thread thread_ReadStream(&NPRV_threaded_read_stream);
	sf::Thread thread_ModWriteStream(&NPRV_threaded_ModWriteStream); 

	FILE* fid_loc_in = 0;

	if (CONF_RX_SAVE_to_FILE == 1) {
		FID_out_SDR_save_raw = fopen("user_data/SDR_samples.cs16", "wb");//.cs16  fid_loc_out
	}
	if (data_from_file == 1) {
	}
	else {

		NPRV_SDR_init_gain(0);//do not init gain TX & RX
		NPRV_SDR_TX_gain_capability_change(1, CONF_my_RFChan_TX_capab, 0);
		NPRV_SDR_open();
		printf("end SDR opened\r\n");
		NPRV_SDR_configure();
		printf("end SDR configure\r\n");
	}

	//current_bandwidth = SoapySDRDevice_getBandwidth(NPRV_glob_sdr, SOAPY_SDR_RX, 0);
	printf("RF bandwidth:%f\r\n", current_bandwidth);

	size_t i;
	for (i = 0; i < 70000; i++) {
		my_buff_tx[i] = 0;
	}
	for (i = 0; i < 269000; i++) {
		data_for_display[i] = 0;
	}

	if (data_from_file == 0) { NPRV_SDR_start_stream(); }
	printf("SDR stream started\r\n");
	
	int j;
	float value;
	int ret = 0;
	long long timeNs;
	float estimated_time;
	uint32_t Tstart2; 
	int flags; //flags set by receive operation

	previous_time = 0;

	delta_t_loc = 0;
	delta_t_max = 0;
	//NPRV_mod_demod_init_param();
	
	NPRV_demod_set_channels();
	NPRV_mod_init();
	NPRV_coder_init();
	size_loc = 32768;//32768 16384 8192

	NPRV_IPv4_init_TX_struct();

	if ((CONF_TX_test == 1) || (CONF_TX_test == 3)) {
		for (i = 0; i < CONF_NPRV_channel_count; i++) {
			NPRV_tx_plan_new_burst(i, 800000 + i*6000, 50000, 0);
		}
	}
	else if (CONF_TX_test == 2) {
		NPRV_TX_test_periodic_call();
	}
	
	
	printf("CONFIG MULTITHREAD\r\n");
	printf("Channel nb: %i\r\n", CONF_NPRV_channel_count);
	
	
#ifdef NPRV_linux 
	struct sigaction new_action, old_action;
	new_action.sa_handler = termination_handler;
	sigemptyset(&new_action.sa_mask);
	new_action.sa_flags = 0;

	sigaction(SIGINT, NULL, &old_action);
	if (old_action.sa_handler != SIG_IGN)
		sigaction(SIGINT, &new_action, NULL);
	sigaction(SIGHUP, NULL, &old_action);
	if (old_action.sa_handler != SIG_IGN)
		sigaction(SIGHUP, &new_action, NULL);
	sigaction(SIGTERM, NULL, &old_action);
	if (old_action.sa_handler != SIG_IGN)
		sigaction(SIGTERM, &new_action, NULL);
#endif /*linux*/

#ifdef NPRV_windows
	SetConsoleCtrlHandler(CtrlHandler, TRUE); //catch events

	// terminal mode windows
	loc_error = 0;
	HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
	if (hOut == INVALID_HANDLE_VALUE) { loc_error = 1; }	
	DWORD dwMode = 0;
	if (loc_error == 0) {
		if (!GetConsoleMode(hOut, &dwMode)) { loc_error = 1; }
	} 
	dwMode |= ENABLE_VIRTUAL_TERMINAL_PROCESSING;
	if (loc_error == 0) {
		if (!SetConsoleMode(hOut, dwMode)) { loc_error = 1; }
	}
#endif /*windows*/

	thread_demod_0.launch();
#ifdef NPRV_linux 
	thread_ReadStream.launch();
	thread_ModWriteStream.launch();
#endif /*linux*/

#ifdef NPRV_windows
	mutex_main2demod[0][0].lock();
	mutex_main2demod[1][0].lock();
#endif /*windows*/
	
	sf::sleep(sf::milliseconds(NPRV_INI_WAIT_TIME));//500
	i = 0;

	

	while (1) { //BIG LOOP
		Eth_RX_dequeue();
		flags = SOAPY_SDR_HAS_TIME & SOAPY_SDR_RX;
				
		if (data_from_file == 0) {

			NPR_SDR_setautogainRX();//!!!
			if ((CONF_is_NPRV_HUB == 0) || (i == 0)) {
				NPR_SDR_setautogainTX();
			}
		}
		
#ifdef NPRV_windows
		//heritage of read thread
		mutex_demod2main[(count_main + 2) % 4][0].lock();
		if (data_from_file == 1) {
		} else if (Kill_requested == 0) {
			NPRV_SDR_read_stream(my_buff_rx[count_main % 4], size_loc);
		}
		mutex_demod2main[(count_main + 2) % 4][0].unlock();
		mutex_main2demod[count_main % 4][0].unlock();
		if ((count_main > 9) && (CONF_RX_SAVE_to_FILE == 1) && (FID_out_SDR_save_raw != 0)) {
			fwrite(my_buff_rx[count_main % 4], sizeof(short int), size_loc * 2, FID_out_SDR_save_raw);//32768 * 2 fid_loc_out
		}
		mutex_main2demod[(count_main + 2) % 4][0].lock();
		//heritage of write thread
		if (CONF_NPRV_filter_TX == 1) {
			TX_need[count_main % 16] = NPRV_mod_total_2(my_buff_tx, size_loc);
		} else {
			TX_need[count_main % 16] = NPRV_mod_total(my_buff_tx, size_loc);
		}
		
		if ((TX_need[(count_main - 3) % 16] == 1) || (TX_need[(count_main - 2) % 16] == 1)) {
			TX_need_synth = 1;
		}
		else {
			TX_need_synth = 0;
		}
		if ((TX_need_synth != TX_need_synth_prev) && (TX_need_synth == 1)) {
			//SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 75);
		}
		if ((TX_need_synth != TX_need_synth_prev) && (TX_need_synth == 0)) {
			//SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, 45);
		}
		TX_need_synth_prev = TX_need_synth;
		if (data_from_file == 0) {
			NPRV_SDR_write_stream(my_buff_tx, size_loc);
		}
#endif /*windows*/
		
		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		//printf("deltaT: %i\r\n", delta_T_loc);
		if (data_from_file == 1) { sf::sleep(sf::milliseconds(32 - delta_T_loc)); }
		else if (delta_T_loc < 15) {
			sf::sleep(sf::milliseconds(25 - delta_T_loc)); 
			//printf("**SLEEP 32** %i\r\n", 25 - delta_T_loc);
		}
		
		elapsed_time = G_clock.getElapsedTime();
		T_capture = elapsed_time.asMilliseconds();
		delta_T_loc = T_capture - T_capture_prev;
		if (delta_T_loc > delta_T_max_loc) { delta_T_max_loc = delta_T_loc; }
		T_capture_prev = T_capture;
		
		if ((T_capture - T_capture_prev_TX_auto_gain) > 500) {
			if ((CONF_is_NPRV_HUB == 0) && (CONF_TX_test == 0)) {
				NPRV_SDR_client_autoGain_TX_periodic_call();
			}
			T_capture_prev_TX_auto_gain = T_capture;
		}
		if ((T_capture - T_capture_prev30) > 1000) {
			if (CONF_TX_test == 2) {
				NPRV_TX_test_periodic_call();
			}
			delta_T_30_loc = T_capture - T_capture_prev30;
			T_capture_prev30 = T_capture;
			delta_T_main_HMI_mean = (float)delta_T_30_loc / 30;
			delta_T_main_HMI_max = delta_T_max_loc;
			delta_T_max_loc = 0;

			NPRV_check_config_update(data_from_file);
			if (count_txt_lines != count_txt_lines_prev) { NPRV_do_not_overwrite_txt = 1; }
			count_txt_lines_prev = count_txt_lines;

			if ((i > 10) && (NPRV_do_not_overwrite_txt == 0)) {
				for (k = 0; k < count_txt_lines; k++) { printf("\x1b[A"); }
			}

			count_txt_lines = 6;//21
			count_txt_lines = count_txt_lines + print_demod_data(); // depend on chan count
			
			printf("%i DELTA_T READ %f Maxi %f\r\n", count_read_stream,delta_T_ReadStream_HMI_mean, delta_T_ReadStream_HMI_max);
			printf("%i DELTA_T DEMOD %f Maxi %f\r\n", count_demod,delta_T_Demod_HMI_mean, delta_T_Demod_HMI_max);
			printf("%i DELTA_T WRITE %f Maxi %f\r\n", count_write_stream, delta_T_WriteStream_HMI_mean, delta_T_WriteStream_HMI_max);
			printf("%i DELTA_T MAIN %f Maxi %f\r\n", count_main, delta_T_main_HMI_mean, delta_T_main_HMI_max);
			NPRV_get_BER_stat(&BER_BPSK, &BER_QPSK);

			NPRV_get_FrameErr_stat(&err_frame_BPSK, &err_frame_QPSK);

			delta_t_count = 0;
			delta_t_tot = 0;
			delta_t_max = 0;

			count_txt_lines = count_txt_lines + NPRV_HMI_print_radio_stats();
			printf("\r\n");

			NPRV_signaling_print_connection_status();
			count_txt_lines = count_txt_lines + DHCP_ARP_print_entries();
			DHCP_ARP_periodic_free_table();
			//if (CONF_TX_test != 0) {
			//	NPRV_print_pseudorand_err();
			//}
			count_txt_lines = count_txt_lines + NPRV_HMI_print_who();
			NPRV_signaling_periodic_call();
			NPRV_periodic_freq_retune();
			NPRV_do_not_overwrite_txt = 0;
		}
		
		if (Kill_requested == 1) {
			printf("\r\nkill requested\r\n");
#ifdef NPRV_windows
			mutex_main2demod[(count_main + 2) % 4][0].unlock();
#endif
			if (data_from_file == 1) {
			} else {
				sf::sleep(sf::milliseconds(500));
				if (CONF_TX_test == 0) {
					NPRV_SDR_TX_gain_capability_change(CONF_my_RFChan_TX_capab, 1, 0);
					NPRV_write_NVRAM();
				}
				NPRV_SDR_stop_stream();

				NPRV_SDR_close();
			}
			//NPRV_dipl_close_all_windows();
			sf::sleep(sf::milliseconds(500));
			if ((CONF_RX_SAVE_to_FILE == 1) && (FID_out_SDR_save_raw != 0)) {
				fclose(FID_out_SDR_save_raw); // fid_loc_out
				FID_out_SDR_save_raw = 0;
			}

			printf("Done\n");
			return 1;
		}
		i++;
		count_main++;
	}
	for (j = 0; j <= 4; j++) {
		printf("RX chan %i : %i\r\n", i, G_NPRV_RX_radio_frame_count[j]);
	}
	printf("delta T moy:%f maxi:%i\r\n", (float)delta_t_tot/(float)delta_t_count, delta_t_max);
	printf("TX frames: %i\r\n", G_TX_rframe_count);

	if (data_from_file == 1) {
		fclose(fid_loc_in);
	} else {
		NPRV_SDR_stop_stream();
		NPRV_SDR_close();
	}
	printf("Done\n");
	return 0;
}


//static float NPRV_current_RX_gain = 29;//30 52 34
static float NPRV_MaxAmpliRX_AGC_min = 800;
static float NPRV_MaxAmpliRX_AGC_max = 0;

void NPR_SDR_setautogainRX(void) {
	static int gainRX_int_prev = 0;
	int gainRX_int;
	gainRX_int = NPRV_current_RX_gain;
	if (gainRX_int != gainRX_int_prev) {
		SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_RX, 0, NPRV_current_RX_gain);//NPRV_current_RX_gain
		gainRX_int_prev = gainRX_int;
		//printf("set gain RX %f\r\n", NPRV_current_RX_gain);
		//NPRV_HMI_new_TXT_message();
	}
}

void NPR_SDR_setautogainTX(void) {
	static int gainTX_int_prev = 0;
	int gainTX_int;
	gainTX_int = NPRV_current_TX_gain;
	if (gainTX_int != gainTX_int_prev) {
		SoapySDRDevice_setGain(NPRV_glob_sdr, SOAPY_SDR_TX, 0, NPRV_current_TX_gain);
		gainTX_int_prev = gainTX_int;
		//printf("set gain TX %f\r\n", NPRV_current_TX_gain);
		//NPRV_HMI_new_TXT_message();
	}
}

void NPRV_SDR_autoGain_RX() {
	//short int* current_IQ;
	int i;
	float ampl_RX_max_loc = 0;
	for (i = 0; i < CONF_NPRV_channel_count; i++) {

		if (NPRV_MaxAmpliRX_AGC[i] > ampl_RX_max_loc) { ampl_RX_max_loc = NPRV_MaxAmpliRX_AGC[i]; }
	}

	if (ampl_RX_max_loc > 480) {//400 for 1 channel
		NPRV_current_RX_gain = NPRV_current_RX_gain - 1;//4
		if (NPRV_current_RX_gain < NPRV_SDR_gain_RX_mini) { NPRV_current_RX_gain = NPRV_SDR_gain_RX_mini; }
	} 
	if (ampl_RX_max_loc < 350) {//300 for 1 channel
		NPRV_current_RX_gain = NPRV_current_RX_gain + 0.1;//+0.03
		if (NPRV_current_RX_gain > NPRV_SDR_gain_RX_maxi) { NPRV_current_RX_gain = NPRV_SDR_gain_RX_maxi; }
	}
	//printf("ampl_RX_max_loc %f ", ampl_RX_max_loc);
	if ((ampl_RX_max_loc > 300) && (ampl_RX_max_loc < 600) && (G_NPRV_RX_gain_stabilised < 10) ) {
		G_NPRV_RX_gain_stabilised = 12;
	} 
	if (((ampl_RX_max_loc < 300) || (ampl_RX_max_loc > 600)) && (G_NPRV_RX_gain_stabilised > 0)) {
		G_NPRV_RX_gain_stabilised--;
	}
	//printf("G_NPRV_RX_gain_stabilised %i\r\n", G_NPRV_RX_gain_stabilised);
	if (ampl_RX_max_loc > NPRV_MaxAmpliRX_AGC_max) {
		NPRV_MaxAmpliRX_AGC_max = ampl_RX_max_loc;
	}
	if (ampl_RX_max_loc < NPRV_MaxAmpliRX_AGC_min) {
		NPRV_MaxAmpliRX_AGC_min = ampl_RX_max_loc;
	}
	//printf("max ampl:%f gain:%f\r\n", NPRV_MaxAmpliRX_AGC, current_gain);
	//NPRV_MaxAmpliRX_AGC = 0;
//#ifndef NPRV_MULTITHREAD
	for (i = 0; i < CONF_NPRV_channel_count; i++) {
		NPRV_MaxAmpliRX_AGC[i] = 0;
	}
//#endif
}

void NPRV_SDR_TX_gain_capability_change(int previous_capa, int new_capa, int apply_immediately) {
	float ratio, gain;
	ratio = (float)new_capa / (float)previous_capa;
	gain = 10 * log10(ratio);//10* because amplitude
	//printf("\r\nGAIN MODIF CAPA:%f (ratio:%f prev:%i new:%i)\r\n\r\n", gain, ratio, previous_capa, new_capa);
	//NPRV_HMI_new_TXT_message();
	NPRV_current_TX_gain = NPRV_current_TX_gain + gain;
	if (NPRV_current_TX_gain < NPRV_SDR_gain_TX_mini) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_mini; }
	if (NPRV_current_TX_gain > NPRV_SDR_gain_TX_maxi) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_maxi; }
	if (NPRV_current_TX_gain > CONF_Maxi_Gain_TX_user) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }
	if (apply_immediately == 1) {
		NPR_SDR_setautogainTX();
	}
}

sf::Mutex NPRV_SDR_stat_autoGain_TX;

static int NPRV_SDR_cycle_TX_gain = 0;
static int NPRV_SDR_TX_frame_count_me_for_gain[3];
static int NPRV_SDR_RX_frame_count_me_for_gain[3];

void NPRV_SDR_store_TX_event(void) {
	NPRV_SDR_stat_autoGain_TX.lock();
	NPRV_SDR_TX_frame_count_me_for_gain[NPRV_SDR_cycle_TX_gain] = 1;
	NPRV_SDR_stat_autoGain_TX.unlock();
}

void NPRV_SDR_store_RX_ME_event(void) {
	NPRV_SDR_stat_autoGain_TX.lock();
	NPRV_SDR_RX_frame_count_me_for_gain[NPRV_SDR_cycle_TX_gain] = 1;
	NPRV_SDR_stat_autoGain_TX.unlock();
}

void NPRV_SDR_client_autoGain_TX_periodic_call(void) {
	int RX_from_me = 0;
	int TX_from_me = 0;
	float local_ratio;
	NPRV_SDR_stat_autoGain_TX.lock();
	RX_from_me =
		NPRV_SDR_RX_frame_count_me_for_gain[NPRV_SDR_cycle_TX_gain] +
		NPRV_SDR_RX_frame_count_me_for_gain[(NPRV_SDR_cycle_TX_gain + 2) % 3];//-1
	TX_from_me =
		NPRV_SDR_TX_frame_count_me_for_gain[(NPRV_SDR_cycle_TX_gain + 2) % 3] + // -1
		NPRV_SDR_TX_frame_count_me_for_gain[(NPRV_SDR_cycle_TX_gain + 1) % 3]; // -2
	if (G_NPRV_RX_gain_stabilised > 1) {
		if (RX_from_me > 0) {
			local_ratio = G_NPRV_my_amplitude_client / G_NPRV_amplitude_ref_hub;
			if (local_ratio > 1.2) {
				NPRV_current_TX_gain = NPRV_current_TX_gain - 0.5;
				if (NPRV_current_TX_gain < NPRV_SDR_gain_TX_mini) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_mini; }
			}
			if (local_ratio < 0.8) {
				NPRV_current_TX_gain = NPRV_current_TX_gain + 0.5;
				if (NPRV_current_TX_gain > NPRV_SDR_gain_TX_maxi) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_maxi; }
				//if (NPRV_current_TX_gain > CONF_Maxi_Gain_TX_user) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }
			}
			//printf("local_ratio:%f\r\n", local_ratio);

		}
		else if ((TX_from_me > 0) && (RX_from_me == 0)) {
			NPRV_current_TX_gain = NPRV_current_TX_gain + 0.5;
			if (NPRV_current_TX_gain > NPRV_SDR_gain_TX_maxi) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_maxi; }
			//if (NPRV_current_TX_gain > CONF_Maxi_Gain_TX_user) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }
		}
	}
	if (NPRV_current_TX_gain > CONF_Maxi_Gain_TX_user) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }

	NPRV_SDR_cycle_TX_gain++;
	if (NPRV_SDR_cycle_TX_gain >= 3) { NPRV_SDR_cycle_TX_gain = 0; }
	NPRV_SDR_RX_frame_count_me_for_gain[NPRV_SDR_cycle_TX_gain] = 0;
	NPRV_SDR_TX_frame_count_me_for_gain[NPRV_SDR_cycle_TX_gain] = 0;
	NPRV_SDR_stat_autoGain_TX.unlock();
}

/*
void NPRV_SDR_autoGain_TX() {
	float local_ratio;
	if (G_NPRV_RX_gain_stabilised > 1) {
		local_ratio = G_NPRV_my_amplitude_client / G_NPRV_amplitude_ref_hub;
		if (local_ratio > 1.2) {
			NPRV_current_TX_gain = NPRV_current_TX_gain - 0.5;
			if (NPRV_current_TX_gain < NPRV_SDR_gain_TX_mini) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_mini; }
		}
		if (local_ratio < 0.8) {
			NPRV_current_TX_gain = NPRV_current_TX_gain + 0.5;
			if (NPRV_current_TX_gain > NPRV_SDR_gain_TX_maxi) { NPRV_current_TX_gain = NPRV_SDR_gain_TX_maxi; }
			if (NPRV_current_TX_gain > CONF_Maxi_Gain_TX_user) { NPRV_current_TX_gain = CONF_Maxi_Gain_TX_user; }
		}
		//printf("ampl_hub:%f ampl_moi:%f gaintx:%f\r\n", G_NPRV_amplitude_ref_hub, G_NPRV_my_amplitude_client, NPRV_current_TX_gain);
	}
}
*/

void NPRV_print_gain_RX(void) {
	printf("\r\n");
	printf("gainRX:%f gainTX:%f stabilised:%i\r\n", NPRV_current_RX_gain, NPRV_current_TX_gain, G_NPRV_RX_gain_stabilised);
	printf("SNR : %f ;; %f / %f\r\n", 20*log10(NPRV_MaxAmpliRX_AGC_max/ NPRV_MaxAmpliRX_AGC_min),NPRV_MaxAmpliRX_AGC_max, NPRV_MaxAmpliRX_AGC_min);
	printf("ampl Master:%f ampl_client:%f\r\n", G_NPRV_amplitude_ref_hub, G_NPRV_my_amplitude_client);
	NPRV_MaxAmpliRX_AGC_min = 800;
	NPRV_MaxAmpliRX_AGC_max = 0;
}

unsigned char NPRV_TX_test_seq_nb = 0;

int NPRV_TX_test_seq(int channel) {
	int answer;
	answer = (NPRV_TX_test_seq_nb >> channel) & 0x01;
	return answer;
}

void NPRV_TX_test_periodic_call(void) {
	unsigned char prev_seq_nb;
	int i;
	int chan_active_count = 0;
	int current_active, prev_active;
	prev_seq_nb = NPRV_TX_test_seq_nb;
	NPRV_TX_test_seq_nb++;
	for (i = 0; i < CONF_NPRV_channel_count; i++) {
		current_active = (NPRV_TX_test_seq_nb >> i) & 0x01;
		prev_active = (prev_seq_nb >> i) & 0x01;
		if ((current_active == 1) && (prev_active == 0)) {
			NPRV_tx_plan_new_burst(i, NPRV_G_tx_timer + 70000 +i * 6000, 50000, 0);
		}
		chan_active_count = chan_active_count + prev_active;
	}
	printf("TX_TEST: %i channels: ", chan_active_count);
	for (i = 0; i < CONF_NPRV_channel_count; i++) {
		prev_active = (prev_seq_nb >> i) & 0x01;
		printf("%i:", prev_active);
	}
	printf("\r\n");
}

static uint16_t NPRV_SDR_freq_offsetRX_conf = 0;
static uint16_t NPRV_SDR_freq_offsetTX_conf = 0;
static float NPRV_freq_offset_RX_obs = 0;
static float NPRV_freq_offset_TX_obs = 0;

void NPRV_SDR_set_freq_offset_conf(float freq_offset_TX_conf) {
	static float freq_offset_TX_conf_prev = 0;
	if (freq_offset_TX_conf != freq_offset_TX_conf_prev) {
		NPRV_freq_offset_TX_tuned = freq_offset_TX_conf;	// unit=Hz
		freq_offset_TX_conf_prev = freq_offset_TX_conf;
	}
}

void NPRV_collect_freq_offset_RX(int channel, float freq_in) {
	float freq_offset_loc;
	freq_offset_loc = (freq_in / 2 / M_PI) - NPRV_channels_freq[channel];
	//printf("FREQ_RX :  %f         \r\n", freq_offset_loc);
	NPRV_freq_offset_RX_obs = 0.1 * freq_offset_loc + 0.9 * NPRV_freq_offset_RX_obs;
}

void NPRV_collect_freq_offset_TX(int channel, float freq_in) {
	float freq_offset_loc;
	freq_offset_loc = (freq_in / 2 / M_PI) - NPRV_channels_freq[channel];
	freq_offset_loc = freq_offset_loc - NPRV_freq_offset_RX_obs;
	if (NPRV_freq_offset_TX_obs == 0) {
		NPRV_freq_offset_TX_obs = 0.6 * freq_offset_loc;
	}
	else {
		NPRV_freq_offset_TX_obs = 0.3 * freq_offset_loc + 0.7 * NPRV_freq_offset_TX_obs;
	}
}

void NPRV_collect_freq_offset_TX_my_discovery(int channel, float freq_in, uint32_t rand_nb_rx) {
	uint32_t my_rand_nb;
	my_rand_nb =
		(G_NPRV_my_random_nb[0] << 24) +
		(G_NPRV_my_random_nb[1] << 16) +
		(G_NPRV_my_random_nb[2] << 8) +
		(G_NPRV_my_random_nb[3]);
	if (rand_nb_rx == my_rand_nb) {
		NPRV_collect_freq_offset_TX(channel, freq_in);
	}
}

void NPRV_print_freq_offset(void) {
	printf("REQ_OFFSET_RX: %f %f\r\n", NPRV_freq_offset_RX_obs * 1000, NPRV_freq_offset_RX_tuned / 1000);
	printf("REQ_OFFSET_TX: %f %f\r\n", NPRV_freq_offset_TX_obs * 1000, NPRV_freq_offset_TX_tuned / 1000);
}

void NPRV_periodic_freq_retune(void) {
	if (abs(NPRV_freq_offset_RX_obs) > .001) { //2 kHz
		//1e6
		NPRV_freq_offset_RX_tuned = NPRV_freq_offset_RX_tuned + (NPRV_freq_offset_RX_obs * 1e6);
		NPRV_freq_offset_RX_obs = 0;
		NPRV_SDR_reconfigure_RX_Freq();
	}
	if (abs(NPRV_freq_offset_TX_obs) > .001) { //2 kHz
		//1e6
		NPRV_freq_offset_TX_tuned = NPRV_freq_offset_TX_tuned - (NPRV_freq_offset_TX_obs * 1e6);
		NPRV_freq_offset_TX_obs = 0;
		NPRV_SDR_reconfigure_TX_Freq();
	}
}
