// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include "NPRV_config.h" 
#include "Eth_IPv4.h"
#include "NPRV_save_to_file.h"
#include "NPRV_SDR_if.h"
#include "NPRV_mod_demod.h"
#include "NPRV_display.h"
#include "NPRV_coder_decoder.h"
#include "NPRV_L1_L2_radio.h"
#include "NPRV_sat_emul.h"
#include "NPRV_TDMA.h"
#include "NPRV_signaling.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#ifdef NPRV_linux
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#endif

#pragma warning(disable: 4996)

int NPRV_read_single_param(char* data_in, const char* label, float* data_out) {
	int answer = 1;
	char* found_str;
	char temp_char[20000];
	found_str = strstr(data_in, label);
	if (found_str == 0) { return 0; }
	strcpy(temp_char, found_str);
	found_str = strtok(temp_char, ":");
	if (found_str == 0) { return 0; }
	//printf("found temp: %s\r\n", found_str);
	found_str = strtok(NULL, " /\r\n");
	if (found_str == 0) { return 0; }
	//printf("found: %s\r\n", found_str);
	sscanf(found_str, "%f", data_out);
	return 1;
}

int NPRV_read_string_param(char* data_in, const char* label, char* found_str) {
	int answer = 1;
	char* found_str_loc;
	char temp_char[20000];
	//printf("read string param\r\n");
	found_str_loc = strstr(data_in, label);
	if (found_str_loc == 0) { return 0; }
	//else { printf("found str 0\r\n"); }
	strcpy(temp_char, found_str_loc);
	found_str_loc = strtok(temp_char, ":");
	if (found_str_loc == 0) { return 0; }
	//printf("char 1:%2X\r\n", found_str_loc[0]);
	//else { printf("found str 1\r\n"); }
	//printf("found temp: %s\r\n", found_str_loc);
	found_str_loc = strtok(NULL, "\r\n");
	if (found_str_loc == 0) { return 0; }
	//printf("char 2:%2X\r\n", found_str_loc[0]);
	if (found_str_loc[0] == '\r') { return 0; }
	if (found_str_loc[0] == '\n') { return 0; }
	printf("found4: %s\r\n", found_str_loc);
	found_str_loc[16] = 0;
	strcpy(found_str, found_str_loc);
	return 1;
}

//int NPRV_read_IP_param(char* data_in, const char* label, long unsigned int* IP_out) {
int NPRV_read_IP_param(char* data_in, const char* label, uint32_t* IP_out) {
	int answer = 1;
	char* found_str;
	char temp_char[20000];
	unsigned char IP_char[5];
	found_str = strstr(data_in, label);
	if (found_str == 0) { return 0; }
	strcpy(temp_char, found_str);
	found_str = strtok(temp_char, ":");
	if (found_str == 0) { return 0; }
	//printf("found temp: %s\r\n", found_str);
	found_str = strtok(NULL, " /\r\n");
	if (found_str == 0) { return 0; }
	//printf("found: \"%s\"\r\n", found_str);
	sscanf(found_str, "%u.%u.%u.%u", &IP_char[0], &IP_char[1], &IP_char[2], &IP_char[3]);
	//printf("scanf IP OK\r\n");
	(*IP_out) = IP_char2int(IP_char);
	return 1;
}

int NPRV_read_route_param(char* data_in, const char* label, unsigned char* route_field_out) {
	int answer = 1;
	char* found_str;
	char temp_char[20000];
	unsigned char route_field[50];
	unsigned int param_loc;
	int loc_index = 0;
	found_str = strstr(data_in, label);
	if (found_str == 0) { return 0; }
	strcpy(temp_char, found_str);
	found_str = strtok(temp_char, ":");
	if (found_str == 0) { return 0; }
	//printf("found temp: %s\r\n", found_str);
	printf("ROUTE_config:");
	while (1) {
		found_str = strtok(NULL, ".");
		if (found_str == 0) { 
			printf("\r\n");
			//exit(-1); 
			return 0; 
		}
		//printf("found: \"%s\"\r\n", found_str);
		sscanf(found_str, "%u.", &param_loc);
		route_field_out[loc_index] = param_loc;
		printf("%u:", route_field_out[loc_index]);
		loc_index++;
	}
	return 1;
}

float costas_QP_kP_loc, costas_QP_kI_loc, costas_BP_kP_loc, costas_BP_kI_loc, PSK_clk_BP_KP, PSK_clk_QP_KP, ampl_ratio_freq_comp, PSK_freq_comp_gain, demod_ampl_threshold, costas_QPins_kI_loc;
int costas_on_symbol = 0;
int displ_constellation_loc = 0;
int displ_eye_loc = 0;

void NPRV_init_config(void) {
	LAN_conf_applied.LAN_def_route = 0;
	LAN_conf_applied.LAN_def_route_activ = 2; // 2 means with DHCP option 121 "Classless Static Routes"
	
	CONF_SDR_HW_type = 0;
	CONF_INHIB_TUNTAP = 0;
	CONF_IPv4_SPY = 0;
	strcpy(CONF_SDR_PLUTO_IP, "192.168.2.1");
	strcpy(G_NPRV_tuntap_name, "tap_01");
	LAN_conf_applied.DHCP_server_active = 1;
	CONF_my_IP_size_requested = 1;
	CONF_NPRV_channel_count = 1;
	CONF_is_NPRV_HUB = 0;
	CONF_RX_SAVE_to_FILE = 0;
	NPRV_init_save_frame_to_file(0);
	CONF_Emul_Radio = 0;
	CONF_Emul_Radio_IP = 0xC0A81201; //192.168.18.1
	CONF_my_RFGlob_TX_capab = 3;
	CONF_my_RFChan_TX_capab = 1;
	CONF_center_freq_TX = 2409e6;
	CONF_center_freq_RX = 2409e6;
	NPRV_RX_LO_freq = 1660.450e6; 
	NPRV_SDR_set_freq_offset_conf(0); //TX_FREQ_OFFSET
	CONF_force_BPSK = 1;
	CONF_Maxi_Gain_TX_user = 89;
	CONF_Hub_TX_PWR = 88;
	CONF_TX_filter_active = 0;
	CONF_NPRV_filter_RX = 1;
	CONF_NPRV_insert_QPSK_symb = 7;
	CONF_NPRV_preamble_typ = 1;
	CONF_TX_test = 0;
	CONF_TXTEST_length = 128;
	CONF_TXTEST_salve_nb = 6;
	NPRV_channels_freq[0] = -0.260;
	NPRV_channels_freq[1] = 0.260;
	NPRV_channels_SR_std[0] = 2;
	NPRV_channels_SR_std[1] = 1;
	displ_constellation_loc = 0;
	displ_eye_loc = 0;
	
	costas_QP_kP_loc = 0.000;
	costas_QP_kI_loc = 0.15;
	costas_BP_kP_loc = 0.000;
	costas_BP_kI_loc = 0.08;
	costas_QPins_kI_loc = 0.6;
	PSK_clk_BP_KP = 0.025;
	PSK_clk_QP_KP = 0.005;
	ampl_ratio_freq_comp = 0.5;
	PSK_freq_comp_gain = 0.004;
	demod_ampl_threshold = 200;
	costas_on_symbol = 1;

	//TDMA 
	CONF_slot_duration = 23800;
	CONF_slot_T_margin = 1200;
	CONF_TDMA_Frame_slot_nb = 16;
	CONF_T_offset_channel = 170;

	NPRV_CONF_DNS_active_hub = 1;
	NPRV_CONF_DNS_active_client = 1;

	CONF_NPRV_filter_TX = 1;//!!! 2023_01_31 why 1 ??
	CONF_NPRV_filter_RX = 1;
	NPRV_CONF_filter_chan_zero = 1;
}

static float saved_TX_Test_RX_Freq = 0;

void NPRV_config_save_TX_test_RX_freq(float freq_loc) {
	saved_TX_Test_RX_Freq = freq_loc;
}

void NPRV_read_config_from_file(int all_data) {
	static char raw_data_from_file[40000] = { 0 };
	static int previous_save_to_file = 0;
	FILE* fid_in = 0;
	int nb_chars;
	char* found_str;
	char temp_char[40000];
	float data_A;
	int data_int;
	int param_present = 0;
	char raw_txt_loc[40];
	unsigned char IP_char_temp[20] = { 0 };
	//int displ_constellation_loc = 0;
	//int displ_eye_loc = 0;
	//float RX_LO_freq = 0;
	float TX_Freq_Offset = 0;
	double F_coup_test_filt = 0;
	static int tab_capab_conversion[20] = { 0, 1, 1, 1, 2, 3 };
	static int previous_chan_capab = 0;
	int i;
	char chan_param_name[20];
	NPRV_HMI_new_TXT_message();
	NPRV_emul_HMI_new_TXT_message();
	fid_in = fopen("config/NPRVSAT_glob_config.txt", "r");
	if (fid_in != 0) {
		nb_chars = fread(raw_data_from_file, 1, 39900, fid_in);
		fclose(fid_in);
	}
	else {
		printf("\r\n\r\nERROR config file missing!\r\n");
		printf("it should be at \"./config/NPRVSAT_glob_config.txt\"\r\n\r\n");
		exit(-1);
	}
	printf("FILE CONFIG, read%i bytes\r\n", nb_chars);

	if (all_data == 1) { // including no live data
		

		param_present = NPRV_read_single_param(raw_data_from_file, "IS_HUB:", &data_A);
		if (param_present) { CONF_is_NPRV_HUB = data_A; }
		printf("IS_HUB : %i\r\n", CONF_is_NPRV_HUB);
		if (CONF_is_NPRV_HUB == 1) {
			printf("ERROR: exec not compatible with hub.\r\n");
			exit(-1);
		}

		param_present = NPRV_read_single_param(raw_data_from_file, "EMUL_RADIO:", &data_A);
		if (param_present) { CONF_Emul_Radio = data_A; }
		printf("emul radio : %i\r\n", CONF_Emul_Radio);

		param_present = NPRV_read_IP_param(raw_data_from_file, "EMUL_IP:", &CONF_Emul_Radio_IP);
		printf("emul IP : %08X\r\n", CONF_Emul_Radio_IP);

		param_present = NPRV_read_single_param(raw_data_from_file, "FRAME_SAVE_TO_FILE:", &data_A);
		printf("Frame save to file : %i\r\n", (int)data_A);
		if (param_present) { NPRV_init_save_frame_to_file((int)data_A); }

		param_present = NPRV_read_single_param(raw_data_from_file, "CLIENT_IP_SIZE:", &data_A);
		if (param_present) { CONF_my_IP_size_requested = data_A; }
		printf("CLIENT_IP_SIZE : %i\r\n", CONF_my_IP_size_requested);

		//param_present = NPRV_read_single_param(raw_data_from_file, "TX_TEST:", &data_A);
		//if (param_present) { CONF_TX_test = data_A; }
		//printf("TX_TEST : %i\r\n", CONF_TX_test);

		param_present = NPRV_read_single_param(raw_data_from_file, "SDR_HW:", &data_A);//0:Pluto 1:Lime
		if (param_present) { CONF_SDR_HW_type = data_A; }
		printf("SDR_HW : %i\r\n", CONF_SDR_HW_type);
#ifdef NPRV_linux
		if (CONF_SDR_HW_type == 1) {
			printf("ERROR : Linux version is not (yet) compatible with Lime-SDR-Mini!\r\n");
			exit(0);
		}
#endif
		param_present = NPRV_read_string_param(raw_data_from_file, "CALLSIGN:", raw_txt_loc);
		if ((param_present) && (raw_txt_loc[0] != 0)) {
			strcpy(G_NPRV_my_Callsign, raw_txt_loc); 
			G_NPRV_my_Callsign[16] = 0;
			printf("CALLsign: %s\r\n", G_NPRV_my_Callsign);
		}
		else {
			printf("\r\nERROR: CALLSIGN not defined inside config file! \r\n");
			exit(0);
		}
		

		param_present = NPRV_read_string_param(raw_data_from_file, "PLUTO_IP:", raw_txt_loc);
		if (param_present) { strcpy(CONF_SDR_PLUTO_IP, raw_txt_loc); }
		printf("PLUTO_IP: %s\r\n", CONF_SDR_PLUTO_IP);

		param_present = NPRV_read_string_param(raw_data_from_file, "TUNTAP_NAME:", raw_txt_loc);
		if (param_present) {
			strcpy(G_NPRV_tuntap_name, raw_txt_loc);
			G_NPRV_tuntap_name[19] = 0;
		}
		printf("TUNTAP_NAME: %s\r\n", G_NPRV_tuntap_name);

		param_present = NPRV_read_single_param(raw_data_from_file, "DHCP_SERVER_ACTIVE:", &data_A);
		if (param_present == 1) { LAN_conf_applied.DHCP_server_active = data_A; }
		printf("DHCP_SERVER_ACTIVE %i: %i\r\n", param_present, LAN_conf_applied.DHCP_server_active);

		param_present = NPRV_read_single_param(raw_data_from_file, "DISPL_CONSTELLATION:", &data_A);
		if (param_present == 1) { displ_constellation_loc = data_A; }
		printf("DISPL_CONSTELLATION %i: %i\r\n", param_present, displ_constellation_loc);

		param_present = NPRV_read_single_param(raw_data_from_file, "DISPL_EYE:", &data_A);
		if (param_present == 1) { displ_eye_loc = data_A; }
		printf("DISPL_EYE %i: %i\r\n", param_present, displ_eye_loc);

		NPRV_display_config(displ_constellation_loc, displ_eye_loc);

		param_present = NPRV_read_single_param(raw_data_from_file, "DEF_ROUTE_ACTIVE:", &data_A);
		if (param_present) { LAN_conf_applied.LAN_def_route_activ = data_A; }
		printf("DEF_ROUTE_ACTIVE : %i\r\n", LAN_conf_applied.LAN_def_route_activ);

		param_present = NPRV_read_single_param(raw_data_from_file, "DHCP_DNS_ACTIVE:", &data_A);
		if (param_present) { NPRV_CONF_DNS_active_client = data_A; }
		printf("DHCP_DNS_ACTIVE : %i\r\n", NPRV_CONF_DNS_active_client);

		param_present = NPRV_read_single_param(raw_data_from_file, "INHIB_TUNTAP:", &data_A);
		if (param_present) { CONF_INHIB_TUNTAP = data_A; }
		printf("INHIB_TUNTAP %i: %i\r\n", param_present, CONF_INHIB_TUNTAP);

		//NPRV_CONF_filter_chan_zero
		param_present = NPRV_read_single_param(raw_data_from_file, "FILTER_CHAN_0:", &data_A);
		if (param_present) { NPRV_CONF_filter_chan_zero = data_A; }
		printf("FILTER_CHAN_0 %i: %i\r\n", param_present, NPRV_CONF_filter_chan_zero);
	}

	// *** beginning live data ***
	param_present = NPRV_read_single_param(raw_data_from_file, "SDR_SAVE_TO_FILE:", &data_A);
	if (param_present) {
		CONF_RX_SAVE_to_FILE = data_A;
		if (CONF_RX_SAVE_to_FILE != previous_save_to_file) {
			NPRV_SDR_Start_stop_recording();
			previous_save_to_file = CONF_RX_SAVE_to_FILE;
		}
	}
	printf("SDR_SAVE_TO_FILE %i : %i\r\n", param_present, CONF_RX_SAVE_to_FILE);

	param_present = NPRV_read_single_param(raw_data_from_file, "TX_MAX_GAIN:", &data_A);
	if (param_present) { CONF_Maxi_Gain_TX_user = data_A; }
	printf("TX Max Gain : %i\r\n", CONF_Maxi_Gain_TX_user);


	param_present = NPRV_read_single_param(raw_data_from_file, "RF_TX_CAPABILITY:", &data_A);
	if (param_present) { CONF_my_RFGlob_TX_capab = data_A; }
	if (CONF_is_NPRV_HUB == 0) {
		if (CONF_my_RFGlob_TX_capab < 1) { CONF_my_RFGlob_TX_capab = 1; }
		if (CONF_my_RFGlob_TX_capab > 5) { CONF_my_RFGlob_TX_capab = 5; }
		CONF_my_RFChan_TX_capab = tab_capab_conversion[CONF_my_RFGlob_TX_capab];
		//CONF_my_RFChan_TX_capab = 6;
	}
	printf("RF_TX_CAPABILITY : %i %i\r\n", CONF_my_RFGlob_TX_capab, CONF_my_RFChan_TX_capab);

	if (CONF_my_RFChan_TX_capab != previous_chan_capab) {
		if (all_data == 0) {
			if (G_NPRV_MyRadioState == 2) {
				NPRV_signal_prioritize_TX_ConReq();
			}
			if (CONF_Emul_Radio == 0) {
				NPRV_SDR_TX_gain_capability_change(previous_chan_capab, CONF_my_RFChan_TX_capab, 1);
				//exit(-1);
			}
		}
		previous_chan_capab = CONF_my_RFChan_TX_capab;
	}

	param_present = NPRV_read_single_param(raw_data_from_file, "TX_FREQ_OFFSET:", &data_A);
	if (param_present) { 
		TX_Freq_Offset = data_A * 1e6; 
		NPRV_SDR_set_freq_offset_conf(TX_Freq_Offset);
	}
	printf("TX_FREQ_OFFSET %i: %f\r\n", param_present, TX_Freq_Offset);

	param_present = NPRV_read_single_param(raw_data_from_file, "RX_LO_FREQ:", &data_A);
	if (param_present) { NPRV_RX_LO_freq = data_A * 1e6; }
	printf("RX_LO_FREQ %i: %f\r\n", param_present, NPRV_RX_LO_freq);
	
	if (CONF_TX_test == 1) {
		CONF_center_freq_RX = saved_TX_Test_RX_Freq - NPRV_RX_LO_freq;
	}

	if (CONF_TX_test == 0) {
		param_present = NPRV_read_single_param(raw_data_from_file, "CENTER_TX_FREQ:", &data_A);
		if (param_present == 1) {
			CONF_center_freq_TX = data_A * 1e6;
		}
		printf("CENTER_TX_FREQ %i: %f\r\n", param_present, CONF_center_freq_TX);

		param_present = NPRV_read_single_param(raw_data_from_file, "CENTER_RX_FREQ:", &data_A);
		if (param_present) {
			CONF_center_freq_RX = data_A * 1e6;
			CONF_center_freq_RX = CONF_center_freq_RX - NPRV_RX_LO_freq;
		}
		printf("CENTER_RX_FREQ %i: %f\r\n", param_present, CONF_center_freq_RX);
		

		/*
		param_present = NPRV_read_single_param(raw_data_from_file, "FREQ_CHAN_1:", &data_A);
		if (param_present) { NPRV_channels_freq[0] = data_A; }
		printf("FREQ_CHAN_1 %i: %f\r\n", param_present, NPRV_channels_freq[0]);

		param_present = NPRV_read_single_param(raw_data_from_file, "FREQ_CHAN_2:", &data_A);
		if (param_present) { NPRV_channels_freq[1] = data_A; }
		printf("FREQ_CHAN_2 %i: %f\r\n", param_present, NPRV_channels_freq[1]);

		param_present = NPRV_read_single_param(raw_data_from_file, "SR_CHAN_1:", &data_A);
		if (param_present) { NPRV_channels_SR_std[0] = data_A; }
		printf("SR_CHAN_1 %i: %f\r\n", param_present, NPRV_channels_SR_std[0]);

		param_present = NPRV_read_single_param(raw_data_from_file, "SR_CHAN_2:", &data_A);
		if (param_present) { NPRV_channels_SR_std[1] = data_A; }
		printf("SR_CHAN_2 %i: %f\r\n", param_present, NPRV_channels_SR_std[1]);
		*/

		
		if (CONF_is_NPRV_HUB == 0) {
			param_present = NPRV_read_single_param(raw_data_from_file, "SR_CHAN_1:", &data_A);
			if (param_present) { NPRV_channels_SR_std[0] = data_A; }
			printf("SR_CHAN_1 %i: %f\r\n", param_present, NPRV_channels_SR_std[0]);
		}
	}

	param_present = NPRV_read_single_param(raw_data_from_file, "TEST_LENGTH:", &data_A);
	if (param_present) { CONF_TXTEST_length = data_A; }
	printf("TEST_LENGTH %i: %i\r\n", param_present, CONF_TXTEST_length);

	param_present = NPRV_read_single_param(raw_data_from_file, "TEST_NB:", &data_A);
	if (param_present) { CONF_TXTEST_salve_nb = data_A; }
	printf("TEST_NB %i: %i\r\n", param_present, CONF_TXTEST_salve_nb);



	param_present = NPRV_read_single_param(raw_data_from_file, "IPV4_SPY:", &data_A);
	if (param_present) { CONF_IPv4_SPY = data_A; }
	printf("IPV4_SPY %i: %i\r\n", param_present, CONF_IPv4_SPY);



	int table_conversion_chan_loc[10] = {1, 1, 1, 0, 2, 2};
	if (CONF_TX_test >= 1) {
		CONF_NPRV_channel_count = 1;// remove if we want to test multiple chanels
		NPRV_channels_SR_std[0] = table_conversion_chan_loc[CONF_my_RFGlob_TX_capab];
		printf("\r\n\r\nNEW SR:%i\r\n\r\n", NPRV_channels_SR_std[0]);

	}

	NPRV_demod_set_channels();
	NPRV_mod_init();
	NPRV_coder_init();


	NPRV_demod_modif_param_live(
		costas_QP_kP_loc, 
		costas_QP_kI_loc, 
		costas_BP_kP_loc, 
		costas_BP_kI_loc, 
		PSK_clk_BP_KP, 
		PSK_clk_QP_KP, 
		ampl_ratio_freq_comp, 
		PSK_freq_comp_gain,
		demod_ampl_threshold, 
		costas_on_symbol, 
		costas_QPins_kI_loc);

}


void NPRV_init_client_config(void) {
	G_NPRV_MyClientID = 0x1FFE;
	G_NPRV_MyRadioState = 0;
	G_NPRV_MyIPConfigured = 0;
	CONF_radio_addr_table_status[0] = { 0 };
	//NPRV_CONF_DNS_active_client = 1;
}


void NPRV_display_config(void) {
	unsigned char IP_char[6];
	IP_int2char(LAN_conf_applied.DHCP_range_start, IP_char);
	printf("LAN_conf_applied.DHCP_range_start: %u.%u.%u.%u\r\n", IP_char[0], IP_char[1], IP_char[2], IP_char[3]);
	printf("LAN_conf_applied.DHCP_range_start: %16X\r\n", LAN_conf_applied.DHCP_range_start);
	IP_int2char(LAN_conf_applied.DHCP_range_size, IP_char);
	printf("LAN_conf_applied.DHCP_range_size: %u.%u.%u.%u\r\n", IP_char[0], IP_char[1], IP_char[2], IP_char[3]);
	IP_int2char(LAN_conf_applied.LAN_subnet_mask, IP_char);
	printf("LAN_conf_applied.LAN_subnet_mask: %u.%u.%u.%u\r\n", IP_char[0], IP_char[1], IP_char[2], IP_char[3]);
	printf("CONF_is_NPRV_HUB : %i\r\n", CONF_is_NPRV_HUB);
}

void NPRV_check_config_update(int SDR_inhibit) {
	static int32_t last_date_glob_config = 0;
	static int32_t last_date_TDMA_config = 0;
	int32_t current_date = 0;
	int32_t current_date2 = 0;
#ifdef NPRV_linux
	struct stat sb;
	struct stat sb2;
	stat("config/NPRVSAT_glob_config.txt", &sb);
	current_date = sb.st_mtim.tv_sec;//LINUX only
#endif
#ifdef NPRV_windows
	struct _stat sb;
	struct _stat sb2;
	_stat("config/NPRVSAT_glob_config.txt", &sb);
	current_date = sb.st_mtime;
#endif
	//printf("\r\n\r\nCONFIGtmtime: %s\r\n", ctime(&sb.st_mtim.tv_sec));
	//printf("\r\n\r\nCONFIGtmtime: %li\r\n", sb.st_mtim.tv_sec);
	if (last_date_glob_config == 0) { last_date_glob_config = current_date; }
	if (last_date_glob_config != current_date) {
		printf("\r\n******\r\n!!!!!! FICHIER CONFIG MODIFIE\r\n********\r\n");
		NPRV_read_config_from_file(0);
		if ((SDR_inhibit == 0) && (CONF_SDR_HW_type != 1)) {// !!! 2023_01_23 inhibition for LIME-SDR because of task slipping 
			NPRV_SDR_re_configure(); 
		}
		NPRV_mod_init();
		NPRV_demod_set_channels();
	}
	last_date_glob_config = current_date;

}