// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "stdio.h"
#include "string.h"
#include "ctype.h"
#include <stdlib.h>
#include <SFML/Graphics.hpp>
#include "NPRV_glob_variables.h"
#include "NPRV_NVRAM.h"
#pragma warning(disable: 4996)

sf::RenderWindow* window_IQ;
sf::RenderWindow* window_eye;
int NPRV_disp_activ_constellation = 0;
int NPRV_disp_activ_eye_diag = 0;

uint16_t DISP_posX_constel, DISP_posY_constel, DISP_posX_eye, DISP_posY_eye;

void NPRV_display_set_windows_pos(uint16_t posX_constel, uint16_t posY_constel, uint16_t posX_eye, uint16_t posY_eye) {
	DISP_posX_constel = posX_constel;
	DISP_posY_constel = posY_constel;
	DISP_posX_eye = posX_eye;
	DISP_posY_eye = posY_eye;
}

void NPRV_display_config(int activ_constellation_loc, int activ_eye_loc) {
	NPRV_disp_activ_constellation = activ_constellation_loc;
	NPRV_disp_activ_eye_diag = activ_eye_loc;
}

void NPRV_init_display(void) {
	if (NPRV_disp_activ_constellation == 1) {
		window_IQ = new(sf::RenderWindow);
		window_IQ->create(sf::VideoMode(600, 600), "IQ constellation");
		if (window_IQ->isOpen()) {
			window_IQ->setPosition(sf::Vector2i(DISP_posX_constel, DISP_posY_constel));
			window_IQ->clear();
			window_IQ->display();
		}
	}
	if (NPRV_disp_activ_eye_diag == 1) {
		window_eye = new(sf::RenderWindow);
		window_eye->create(sf::VideoMode(600, 400), "Eye diagram");
		if (window_eye->isOpen()) {
			window_eye->setPosition(sf::Vector2i(DISP_posX_eye, DISP_posY_eye));
			window_eye->clear();
			window_eye->display();
		}
	}
	if (NPRV_disp_activ_eye_diag == 3) {
		window_eye = new(sf::RenderWindow);
		window_eye->create(sf::VideoMode(600, 600), "Eye diagram");
		if (window_eye->isOpen()) {
			window_eye->clear();
			window_eye->display();
		}
	}
}

void NPRV_dipl_close_all_windows(void) {
	sf::Vector2i vect_pos;
	int posX_constel, posY_constel, constel_opened;
	int posX_eye, posY_eye, eye_opened;
	eye_opened = 0;
	constel_opened = 0;
	
	if (NPRV_disp_activ_constellation == 1) {
		if (window_IQ->isOpen()) {
			constel_opened = 1;
			vect_pos = window_IQ->getPosition();
			posX_constel = vect_pos.x;
			posY_constel = vect_pos.y;
			printf("POSITION WINDOW IQ : %i  %i\r\n", vect_pos.x, vect_pos.y);
			window_IQ->close();
		}
	}
	if (NPRV_disp_activ_eye_diag >= 1) {
		if (window_eye->isOpen()) {
			eye_opened = 1;
			vect_pos = window_eye->getPosition();
			posX_eye = vect_pos.x;
			posY_eye = vect_pos.y;
			window_eye->close();
		}
	}
	NPRV_save_wind_pos(posX_constel, posY_constel, constel_opened, posX_eye, posY_eye, eye_opened);
}

void NPRV_display_IQ(float* raw_IQ, char* color_loc) {
	sf::Event event;
	static sf::Vertex IQ_points[50000];
	static sf::Vertex lines_axes[20];
	sf::Color Mycolor_grey(160, 160, 160);
	sf::CircleShape my_circle(1);
	sf::RectangleShape my_rect(sf::Vector2f(1, 1));
	int i;
	float scale_factor = 0.5;
	float local_I, local_Q;

	if (NPRV_disp_activ_constellation == 1) {
		if (window_IQ->isOpen()) {
			lines_axes[0] = sf::Vertex(sf::Vector2f(0, 300), sf::Color::Yellow);
			lines_axes[1] = sf::Vertex(sf::Vector2f(600, 300), sf::Color::Yellow);
			lines_axes[2] = sf::Vertex(sf::Vector2f(300, 0), sf::Color::Yellow);
			lines_axes[3] = sf::Vertex(sf::Vector2f(300, 600), sf::Color::Yellow);
			lines_axes[4] = sf::Vertex(sf::Vector2f(0, 0), sf::Color::Blue);
			lines_axes[5] = sf::Vertex(sf::Vector2f(600, 600), sf::Color::Blue);
			lines_axes[6] = sf::Vertex(sf::Vector2f(600, 0), sf::Color::Blue);
			lines_axes[7] = sf::Vertex(sf::Vector2f(0, 600), sf::Color::Blue);
			window_IQ->clear();
			window_IQ->draw(lines_axes, 8, sf::Lines);
			// construction of the graphic
			for (i = 0; i < 3000; i++) {
				//IQ_points[i] = sf::Vertex(sf::Vector2f(i*3, i*3));
				local_I = (float)300 + scale_factor * raw_IQ[2 * i];
				local_Q = (float)300 - scale_factor * raw_IQ[2 * i + 1];
				//if (color_loc[i] == 1) {
				//	IQ_points[i] = sf::Vertex(sf::Vector2f(local_I, local_Q), sf::Color::White);
				//}
				//else {
				//	IQ_points[i] = sf::Vertex(sf::Vector2f(local_I, local_Q), Mycolor_grey);
				//}
				//if (color_loc[i] == 1) {
				if (1) {
					//my_circle.setPosition(sf::Vector2f(local_I, local_Q));
					//my_circle.setRadius(color_loc[i] + 0);
					my_rect.setSize(sf::Vector2f(2*color_loc[i] + 1, 2*color_loc[i] + 1));
					//my_rect.setSize(color_loc[i] + 1, color_loc[i] + 1);
					my_rect.setPosition(local_I, local_Q);
					window_IQ->draw(my_rect);
				}
				else {

				}
				//printf("I:%f Q:%f\r\n", local_I, local_Q);
			}
			//printf("\r\n");
			//displaying actually
			window_IQ->display();
			while (window_IQ->pollEvent(event))
			{
				if (event.type == sf::Event::Closed)
					window_IQ->close();
			}
		}
	}
}

void NPRV_display_eye(float* data_I, float* data_Q, float* data_BPSK) {
	sf::Event event;
	static sf::Vertex lines_BPSK[50000];
	static sf::Vertex lines_I[50000];
	static sf::Vertex lines_Q[50000];
	static sf::Vertex lines_axes[20];
	const int max_X = 30;
	float X_graph, Y_graph_BPSK, Y_graph_I, Y_graph_Q;
	int sampl_count, curv_count, index_in;
	int index_out = 0;
	int request_BPSK = 0;
	int request_QPSK = 0;
	float scale_factor;
	if ((NPRV_disp_activ_eye_diag & 0x01) == 0x01) { request_BPSK = 1; }
	if ((NPRV_disp_activ_eye_diag & 0x02) == 0x02) { request_QPSK = 1; }
	//scale_factor = (float)70 / NPRV_MaxAmpliRX_AGC[0];
	scale_factor = 0.25;
	if (NPRV_disp_activ_eye_diag > 0) {
		if (window_eye->isOpen()) {
			lines_axes[0] = sf::Vertex(sf::Vector2f(0, 100), sf::Color::Yellow);
			lines_axes[1] = sf::Vertex(sf::Vector2f(600, 100), sf::Color::Yellow);
			if (request_QPSK == 1) {
				lines_axes[2] = sf::Vertex(sf::Vector2f(0, 300), sf::Color::Yellow);
				lines_axes[3] = sf::Vertex(sf::Vector2f(600, 300), sf::Color::Yellow);
				lines_axes[4] = sf::Vertex(sf::Vector2f(0, 500), sf::Color::Yellow);
				lines_axes[5] = sf::Vertex(sf::Vector2f(600, 500), sf::Color::Yellow);
			}
			//lines_axes[1] = sf::Vertex(sf::Vector2f(0, 100), sf::Color::Yellow);
			for (curv_count = 0; curv_count < 40; curv_count++) {
				for (sampl_count = 0; sampl_count <= max_X; sampl_count++) {
					index_in = curv_count * 40 + sampl_count;
					X_graph = 20 + (20 * sampl_count);
					Y_graph_BPSK = 100 - (scale_factor * data_BPSK[index_in]);
					lines_BPSK[index_out] = sf::Vertex(sf::Vector2f(X_graph, Y_graph_BPSK));
					if (request_QPSK == 1) {
						Y_graph_I = 300 - (scale_factor * data_I[index_in]);
						Y_graph_Q = 500 - (scale_factor * data_Q[index_in]);
						lines_I[index_out] = sf::Vertex(sf::Vector2f(X_graph, Y_graph_I));
						lines_Q[index_out] = sf::Vertex(sf::Vector2f(X_graph, Y_graph_Q));
					}
					index_out++;
					if ((sampl_count != 0) && (sampl_count < max_X)) {
						lines_BPSK[index_out] = sf::Vertex(sf::Vector2f(X_graph, Y_graph_BPSK));
						if (request_QPSK == 1) {
							lines_I[index_out] = sf::Vertex(sf::Vector2f(X_graph, Y_graph_I));
							lines_Q[index_out] = sf::Vertex(sf::Vector2f(X_graph, Y_graph_Q));
						}
						index_out++;
					}
				}
			}

			window_eye->clear();
			window_eye->draw(lines_axes, 6, sf::Lines);
			window_eye->draw(lines_BPSK, 3200, sf::Lines);
			if (request_QPSK == 1) {
				window_eye->draw(lines_I, 3200, sf::Lines);
				window_eye->draw(lines_Q, 3200, sf::Lines);
			}
			window_eye->display();
			while (window_eye->pollEvent(event))
			{
				if (event.type == sf::Event::Closed) {
					window_eye->close();
					NPRV_disp_activ_eye_diag = 0;
				}
			}
		}
	}
}