// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include <math.h>
#include "NPRV_mod_demod.h"
#include "NPRV_coder_decoder.h"
#include "NPRV_L1_L2_radio.h"
#include "NPRV_display.h"
#include <SFML/System.hpp>
#pragma warning(disable: 4996)

#define M_PI 3.14159265358979

#define NPRV_SAMPLE_DEMOD 1
#define NPRV_SAMPLE_FREQ 1
#define NPRV_DEMOD_DATA_DISPLAY
//#define NPRV_MOD_EXTENDED

//*** QPSK param ****
static float NPRV_DEMOD_QPSK_costas_kP = 0;
static float NPRV_DEMOD_QPSK_costas_kI = 0;
static float NPRV_DEMOD_BPSK_costas_kP = 0;
static float NPRV_DEMOD_BPSK_costas_kI = 0;
static float NPRV_DEMOD_ClkRecov_BPSK_kP = 0;
static float NPRV_DEMOD_ClkRecov_QPSK_kP = 0;
static float NPRV_DEMOD_ampl_ratio_freq_comp = 0;
static float NPRV_DEMOD_PSK_freq_comp_gain = 0;
static float NPRV_demod_ampl_thrshld = 100;
static int NPRV_DEMOD_CLK_ESTIM_PREAM = 0;
static int NPRV_demod_costas_on_symb = 0;
static float NPRV_demod_QPSKins_costas_kI = 0;

static int NPRV_demod_curr_ModulTyp[NPRV_MAX_CHANNEL_COUNT] = { NPRV_BPSK, NPRV_BPSK, NPRV_BPSK, NPRV_BPSK, NPRV_BPSK, NPRV_BPSK, NPRV_BPSK };
static int NPRV_demod_post_sync_word[NPRV_MAX_CHANNEL_COUNT] = { 0,0,0,0,0,0,0,0,0,0};
static float NPRV_eye_data_BPSK[2000] = { 0 };
static float NPRV_eye_data_I[2000] = { 0 };
static float NPRV_eye_data_Q[2000] = { 0 };



void NPRV_demod_modif_param_live(
	float costas_QP_kP_loc,
	float costas_QP_kI_loc,
	float costas_BP_kP_loc,
	float costas_BP_kI_loc,
	float Clock_recov_BPSK_kP,
	float Clock_recov_QPSK_kP,
	float ampl_ratio_freq_comp,
	float PSK_freq_comp_gain,
	float demod_ampl_threshold,
	int costas_on_symb, 
	float QPSKins_costas_kI
) {
	NPRV_DEMOD_QPSK_costas_kP = costas_QP_kP_loc;
	NPRV_DEMOD_QPSK_costas_kI = costas_QP_kI_loc;
	NPRV_DEMOD_BPSK_costas_kP = costas_BP_kP_loc;
	NPRV_DEMOD_BPSK_costas_kI = costas_BP_kI_loc;
	NPRV_DEMOD_ClkRecov_BPSK_kP = Clock_recov_BPSK_kP;
	NPRV_DEMOD_ClkRecov_QPSK_kP = Clock_recov_QPSK_kP;
	NPRV_DEMOD_ampl_ratio_freq_comp = ampl_ratio_freq_comp;
	NPRV_DEMOD_PSK_freq_comp_gain = PSK_freq_comp_gain;
	NPRV_demod_ampl_thrshld = demod_ampl_threshold;
	NPRV_demod_costas_on_symb = costas_on_symb;
	NPRV_demod_QPSKins_costas_kI = QPSKins_costas_kI;
	//printf("DEMOD DEMOD\r\n NPRV_DEMOD_BPSK_costas_kI: %f\r\n", NPRV_DEMOD_BPSK_costas_kI);
}

void NPRV_demod_switch_PSK_ModulTyp(int new_ModulTyp, int post_sync_word, int chan) {
	NPRV_demod_curr_ModulTyp[chan] = new_ModulTyp;
	NPRV_demod_post_sync_word[chan] = post_sync_word;//0:before_sync 1:after_sync 2:after_header
	if (post_sync_word == 1) {//just after sync word

	}
}

//*** end QPSK   ****

const float CONST_deltaf_param[5] = {
	
	2 * M_PI * 0.00833333, //100kS
	2 * M_PI * 0.004166666, //50kS
	2 * M_PI * 0.01666666, //200kS
};

// DEMOD Parameters
static float NPRV_deltaf_demod[NPRV_MAX_CHANNEL_COUNT];

static int step_demod = 0;

static long unsigned int NPRV_G_rx_micro_timer[NPRV_MAX_CHANNEL_COUNT] = { 0 };

static float offset_freq_rx_init[NPRV_MAX_CHANNEL_COUNT] = { 0 };
static float NPRV_alpha_rx[2 * NPRV_MAX_CHANNEL_COUNT] = { 0 };
static float NPRV_alpha_costas[2 * NPRV_MAX_CHANNEL_COUNT] = { 0 };

static float freq_offset_rx[NPRV_MAX_CHANNEL_COUNT] = { 0 };
static int NPRV_demod_period_chan[NPRV_MAX_CHANNEL_COUNT] = { 1, 1, 1, 1 };
static int NPRV_demod_period_param[6] = { 1, 2 , 1};

static float freq_offset_rx_minlim[NPRV_MAX_CHANNEL_COUNT];
static float freq_offset_rx_maxlim[NPRV_MAX_CHANNEL_COUNT];
static float NPRV_demod_deriv_threshold_param[6] = { 0.075 , 0.075 / 2 };
static float NPRV_demod_deriv_threshold[NPRV_MAX_CHANNEL_COUNT] = { 0.075 };//0.22  !!!
void NPRV_demod_set_channels(void) {
	int i;
	for (i = 0; i < NPRV_MAX_CHANNEL_COUNT; i++) {
		offset_freq_rx_init[i] = 2* M_PI * NPRV_channels_freq[i];
		freq_offset_rx[i] = offset_freq_rx_init[i];
		if (NPRV_channels_SR_std[i] < 6) {
			NPRV_deltaf_demod[i] = CONST_deltaf_param[NPRV_channels_SR_std[i]];
			freq_offset_rx_minlim[i] = offset_freq_rx_init[i] - 3 * NPRV_deltaf_demod[i];
			freq_offset_rx_maxlim[i] = offset_freq_rx_init[i] + 3 * NPRV_deltaf_demod[i];
			NPRV_demod_period_chan[i] = NPRV_demod_period_param[NPRV_channels_SR_std[i]];
			NPRV_demod_deriv_threshold[i] = NPRV_demod_deriv_threshold_param[NPRV_channels_SR_std[i]];
		} else {
			printf("ERR unknown SR STD\r\n");
		}
	}
	NPRV_init_butterworth_filter();
}
void NPRV_demod_freq_offset_init(int channel) {
	
	int i;
	if (channel == 0xFF) {
		offset_freq_rx_init[0] = 2 * M_PI * (-0.360);// 0.120
		offset_freq_rx_init[1] = 2 * M_PI * (0.360); //0.150
		offset_freq_rx_init[2] = 2 * M_PI * (-0.120); //0.150
		offset_freq_rx_init[3] = 2 * M_PI * (-0.360); //0.150
		for (i = 0; i < 10; i++) {
			freq_offset_rx[i] = offset_freq_rx_init[i];
			freq_offset_rx_minlim[i] = offset_freq_rx_init[i] - 0.025 * 2 * M_PI;//0.157~=25kHz
			freq_offset_rx_maxlim[i] = offset_freq_rx_init[i] + 0.025 * 2 * M_PI;
		}
		
	}
	else {
		freq_offset_rx[channel] = offset_freq_rx_init[channel];
	}
}

void NPRV_demod_freq_offset(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chn) {
	static int i = 0;
	i++;
	NPRV_alpha_rx[chn] = NPRV_alpha_rx[chn] + freq_offset_rx[chn];//frequency 0.623 1.047 0.95 0.071
	(*data_out_I) = data_in_I * cos(NPRV_alpha_rx[chn]) + data_in_Q * sin(NPRV_alpha_rx[chn]);
	(*data_out_Q) = data_in_Q * cos(NPRV_alpha_rx[chn]) - data_in_I * sin(NPRV_alpha_rx[chn]);
	if (i >= 10) {
		if (NPRV_alpha_rx[chn] > (8*M_PI)) { NPRV_alpha_rx[chn] = NPRV_alpha_rx[chn] - (8*M_PI); }
		if (NPRV_alpha_rx[chn] < (-8*M_PI)) { NPRV_alpha_rx[chn] = NPRV_alpha_rx[chn] + (8*M_PI); }
		i = 0;
	}
}

//NPRV_alpha_costas
void NPRV_demod_phase_offset(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chn) {
	static int i = 0;
	i++;
	(*data_out_I) = data_in_I * cos(NPRV_alpha_costas[chn]) + data_in_Q * sin(NPRV_alpha_costas[chn]);
	(*data_out_Q) = data_in_Q * cos(NPRV_alpha_costas[chn]) - data_in_I * sin(NPRV_alpha_costas[chn]);
	if (i >= 10) {
		if (NPRV_alpha_costas[chn] > (8 * M_PI)) { NPRV_alpha_costas[chn] = NPRV_alpha_costas[chn] - (8 * M_PI); }
		if (NPRV_alpha_costas[chn] < (-8 * M_PI)) { NPRV_alpha_costas[chn] = NPRV_alpha_costas[chn] + (8 * M_PI); }
		i = 0;
	}
	//printf("alpha:%f Iin:%f Qin:%f Iout:%f Qout:%f\r\n", NPRV_alpha_costas[chn], data_in_I, data_in_Q, (*data_out_I), (*data_out_Q));
}

void NPRV_mod_freq_offset(short int data_in_I, short int data_in_Q, float* data_out_I, float* data_out_Q, int chn) {
	static float alpha_tx[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static int cnt[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	cnt[chn]++;
	alpha_tx[chn] = alpha_tx[chn] - offset_freq_rx_init[chn];//frequency 0.623 1.047 0.95 0.071
	(*data_out_I) = data_in_I * cos(alpha_tx[chn]) + data_in_Q * sin(alpha_tx[chn]);
	(*data_out_Q) = data_in_Q * cos(alpha_tx[chn]) - data_in_I * sin(alpha_tx[chn]);
	if (cnt[chn] >= 10) {
		if (alpha_tx[chn] > (32 * M_PI)) { alpha_tx[chn] = alpha_tx[chn] - (32 * M_PI); }//8
		if (alpha_tx[chn] < (-32 * M_PI)) { alpha_tx[chn] = alpha_tx[chn] + (32 * M_PI); }//8
		cnt[chn] = 0;
	}
}

int  print_demod_data(void) {
	int i;
	char char_loc;
	char_loc = 'N';
	for (i = 0; i < CONF_NPRV_channel_count; i++) {
		//printf("%c chan:%i freq:%f alpha:%f\r\n", char_loc, i, freq_offset_rx[i]*1000 /2 /M_PI, NPRV_alpha_costas[i]);
		printf("%c chan:%i freq:%f  \r\n", char_loc, i, freq_offset_rx[i] * 1000 / 2 / M_PI);
	}
	printf("\r\n");
	return (CONF_NPRV_channel_count + 1);
}


static float xv_I[2 * NPRV_MAX_CHANNEL_COUNT][7] = { 0.0 };
static float yv_I[2 * NPRV_MAX_CHANNEL_COUNT][7] = { 0.0 };
static float xv_Q[2 * NPRV_MAX_CHANNEL_COUNT][7] = { 0.0 };
static float yv_Q[2 * NPRV_MAX_CHANNEL_COUNT][7] = { 0.0 };

void NPRV_demod_butterw_filter_008(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {
	//const float GAIN = 9.339780497e+03;
	const float GAIN2 = 1.070688974244316e-4;
	xv_I[chan][0] = xv_I[chan][1]; xv_I[chan][1] = xv_I[chan][2]; xv_I[chan][2] = xv_I[chan][3]; 
	xv_I[chan][3] = xv_I[chan][4]; xv_I[chan][4] = xv_I[chan][5]; xv_I[chan][5] = xv_I[chan][6];
	//xv_I[chan][6] = data_in_I / GAIN;
	xv_I[chan][6] = data_in_I * GAIN2;
	yv_I[chan][0] = yv_I[chan][1]; yv_I[chan][1] = yv_I[chan][2]; yv_I[chan][2] = yv_I[chan][3]; 
	yv_I[chan][3] = yv_I[chan][4]; yv_I[chan][4] = yv_I[chan][5]; yv_I[chan][5] = yv_I[chan][6];
	yv_I[chan][6] = (xv_I[chan][0] + xv_I[chan][6]) + 6.0 * (xv_I[chan][1] + xv_I[chan][5]) + 15.0 * (xv_I[chan][2] + xv_I[chan][4])
		+ 20.0 * xv_I[chan][3]
		+ (-0.1396600417 * yv_I[chan][0]) + (1.1086708553 * yv_I[chan][1])
		+ (-3.7230194289 * yv_I[chan][2]) + (6.7850160254 * yv_I[chan][3])
		+ (-7.0995038188 * yv_I[chan][4]) + (4.0616439992 * yv_I[chan][5]);
	(*data_out_I) = yv_I[chan][6];

	xv_Q[chan][0] = xv_Q[chan][1]; xv_Q[chan][1] = xv_Q[chan][2]; xv_Q[chan][2] = xv_Q[chan][3]; 
	xv_Q[chan][3] = xv_Q[chan][4]; xv_Q[chan][4] = xv_Q[chan][5]; xv_Q[chan][5] = xv_Q[chan][6];
	//xv_Q[chan][6] = data_in_Q / GAIN;
	xv_Q[chan][6] = data_in_Q * GAIN2;
	yv_Q[chan][0] = yv_Q[chan][1]; yv_Q[chan][1] = yv_Q[chan][2]; yv_Q[chan][2] = yv_Q[chan][3]; 
	yv_Q[chan][3] = yv_Q[chan][4]; yv_Q[chan][4] = yv_Q[chan][5]; yv_Q[chan][5] = yv_Q[chan][6];
	yv_Q[chan][6] = (xv_Q[chan][0] + xv_Q[chan][6]) + 6.0 * (xv_Q[chan][1] + xv_Q[chan][5]) + 15.0 * (xv_Q[chan][2] + xv_Q[chan][4])
		+ 20.0 * xv_Q[chan][3]
		+ (-0.1396600417 * yv_Q[chan][0]) + (1.1086708553 * yv_Q[chan][1])
		+ (-3.7230194289 * yv_Q[chan][2]) + (6.7850160254 * yv_Q[chan][3])
		+ (-7.0995038188 * yv_Q[chan][4]) + (4.0616439992 * yv_Q[chan][5]);
	(*data_out_Q) = yv_Q[chan][6];
}

void NPRV_demod_butterw_filter_005(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {
	//const float GAIN = 1.165969038e+05;
	const float GAIN2 = 8.5765570732076334946e-6;
	xv_I[chan][0] = xv_I[chan][1]; xv_I[chan][1] = xv_I[chan][2]; xv_I[chan][2] = xv_I[chan][3];
	xv_I[chan][3] = xv_I[chan][4]; xv_I[chan][4] = xv_I[chan][5]; xv_I[chan][5] = xv_I[chan][6];
	//xv_I[chan][6] = data_in_I / GAIN;
	xv_I[chan][6] = data_in_I * GAIN2;
	yv_I[chan][0] = yv_I[chan][1]; yv_I[chan][1] = yv_I[chan][2]; yv_I[chan][2] = yv_I[chan][3];
	yv_I[chan][3] = yv_I[chan][4]; yv_I[chan][4] = yv_I[chan][5]; yv_I[chan][5] = yv_I[chan][6];
	yv_I[chan][6] = (xv_I[chan][0] + xv_I[chan][6]) + 6.0 * (xv_I[chan][1] + xv_I[chan][5]) + 15.0 * (xv_I[chan][2] + xv_I[chan][4])
		+ 20.0 * xv_I[chan][3]
		+ (-0.2951724313 * yv_I[chan][0]) + (2.1290387500 * yv_I[chan][1])
		+ (-6.4411118810 * yv_I[chan][2]) + (10.4690788930 * yv_I[chan][3])
		+ (-9.6495177287 * yv_I[chan][4]) + (4.7871354989 * yv_I[chan][5]);
	(*data_out_I) = yv_I[chan][6];

	xv_Q[chan][0] = xv_Q[chan][1]; xv_Q[chan][1] = xv_Q[chan][2]; xv_Q[chan][2] = xv_Q[chan][3];
	xv_Q[chan][3] = xv_Q[chan][4]; xv_Q[chan][4] = xv_Q[chan][5]; xv_Q[chan][5] = xv_Q[chan][6];
	//xv_Q[chan][6] = data_in_Q / GAIN;
	xv_Q[chan][6] = data_in_Q * GAIN2;
	yv_Q[chan][0] = yv_Q[chan][1]; yv_Q[chan][1] = yv_Q[chan][2]; yv_Q[chan][2] = yv_Q[chan][3];
	yv_Q[chan][3] = yv_Q[chan][4]; yv_Q[chan][4] = yv_Q[chan][5]; yv_Q[chan][5] = yv_Q[chan][6];
	yv_Q[chan][6] = (xv_Q[chan][0] + xv_Q[chan][6]) + 6.0 * (xv_Q[chan][1] + xv_Q[chan][5]) + 15.0 * (xv_Q[chan][2] + xv_Q[chan][4])
		+ 20.0 * xv_Q[chan][3]
		+ (-0.2951724313 * yv_Q[chan][0]) + (2.1290387500 * yv_Q[chan][1])
		+ (-6.4411118810 * yv_Q[chan][2]) + (10.4690788930 * yv_Q[chan][3])
		+ (-9.6495177287 * yv_Q[chan][4]) + (4.7871354989 * yv_Q[chan][5]);
	(*data_out_Q) = yv_Q[chan][6];
}

void NPRV_demod_butterw_filter_004(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {
	//const float GAIN = 4.004448900e+05;
	const float GAIN2 = 2.497222526675268e-6;
	xv_I[chan][0] = xv_I[chan][1]; xv_I[chan][1] = xv_I[chan][2]; xv_I[chan][2] = xv_I[chan][3];
	xv_I[chan][3] = xv_I[chan][4]; xv_I[chan][4] = xv_I[chan][5]; xv_I[chan][5] = xv_I[chan][6];
	//xv_I[chan][6] = data_in_I / GAIN;
	xv_I[chan][6] = data_in_I * GAIN2;
	yv_I[chan][0] = yv_I[chan][1]; yv_I[chan][1] = yv_I[chan][2]; yv_I[chan][2] = yv_I[chan][3];
	yv_I[chan][3] = yv_I[chan][4]; yv_I[chan][4] = yv_I[chan][5]; yv_I[chan][5] = yv_I[chan][6];
	yv_I[chan][6] = (xv_I[chan][0] + xv_I[chan][6]) + 6.0 * (xv_I[chan][1] + xv_I[chan][5]) + 15.0 * (xv_I[chan][2] + xv_I[chan][4])
		+ 20.0 * xv_I[chan][3]
		+ (-0.3774523864 * yv_I[chan][0]) + (2.63105512850 * yv_I[chan][1])
		+ (-7.6754745482 * yv_I[chan][2]) + (11.9993158160 * yv_I[chan][3])
		+ (-10.6070421840 * yv_I[chan][4]) + (5.0294383514 * yv_I[chan][5]);
	(*data_out_I) = yv_I[chan][6];

	xv_Q[chan][0] = xv_Q[chan][1]; xv_Q[chan][1] = xv_Q[chan][2]; xv_Q[chan][2] = xv_Q[chan][3];
	xv_Q[chan][3] = xv_Q[chan][4]; xv_Q[chan][4] = xv_Q[chan][5]; xv_Q[chan][5] = xv_Q[chan][6];
	//xv_Q[chan][6] = data_in_Q / GAIN;
	xv_Q[chan][6] = data_in_Q * GAIN2;
	yv_Q[chan][0] = yv_Q[chan][1]; yv_Q[chan][1] = yv_Q[chan][2]; yv_Q[chan][2] = yv_Q[chan][3];
	yv_Q[chan][3] = yv_Q[chan][4]; yv_Q[chan][4] = yv_Q[chan][5]; yv_Q[chan][5] = yv_Q[chan][6];
	yv_Q[chan][6] = (xv_Q[chan][0] + xv_Q[chan][6]) + 6.0 * (xv_Q[chan][1] + xv_Q[chan][5]) + 15.0 * (xv_Q[chan][2] + xv_Q[chan][4])
		+ 20.0 * xv_Q[chan][3]
		+ (-0.3774523864 * yv_Q[chan][0]) + (2.6310551285 * yv_Q[chan][1])
		+ (-7.6754745482 * yv_Q[chan][2]) + (11.9993158160 * yv_Q[chan][3])
		+ (-10.6070421840 * yv_Q[chan][4]) + (5.0294383514 * yv_Q[chan][5]);
	(*data_out_Q) = yv_Q[chan][6];
}

void NPRV_demod_butterw_filter_010(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {
	//const float GAIN = 2.936532839e+03;
	const float GAIN2 = 3.405376526763e-4;
	xv_I[chan][0] = xv_I[chan][1]; xv_I[chan][1] = xv_I[chan][2]; xv_I[chan][2] = xv_I[chan][3];
	xv_I[chan][3] = xv_I[chan][4]; xv_I[chan][4] = xv_I[chan][5]; xv_I[chan][5] = xv_I[chan][6];
	//xv_I[chan][6] = data_in_I / GAIN;
	xv_I[chan][6] = data_in_I * GAIN2;
	yv_I[chan][0] = yv_I[chan][1]; yv_I[chan][1] = yv_I[chan][2]; yv_I[chan][2] = yv_I[chan][3];
	yv_I[chan][3] = yv_I[chan][4]; yv_I[chan][4] = yv_I[chan][5]; yv_I[chan][5] = yv_I[chan][6];
	yv_I[chan][6] = (xv_I[chan][0] + xv_I[chan][6]) + 6.0 * (xv_I[chan][1] + xv_I[chan][5]) + 15.0 * (xv_I[chan][2] + xv_I[chan][4])
		+ 20.0 * xv_I[chan][3]
		+ (-0.0837564796 * yv_I[chan][0]) + (0.7052741145 * yv_I[chan][1])
		+ (-2.5294949058 * yv_I[chan][2]) + (4.9654152288 * yv_I[chan][3])
		+ (-5.6586671659 * yv_I[chan][4]) + (3.5794347983 * yv_I[chan][5]);
	(*data_out_I) = yv_I[chan][6];

	xv_Q[chan][0] = xv_Q[chan][1]; xv_Q[chan][1] = xv_Q[chan][2]; xv_Q[chan][2] = xv_Q[chan][3];
	xv_Q[chan][3] = xv_Q[chan][4]; xv_Q[chan][4] = xv_Q[chan][5]; xv_Q[chan][5] = xv_Q[chan][6];
	//xv_Q[chan][6] = data_in_Q / GAIN;
	xv_Q[chan][6] = data_in_Q * GAIN2;
	yv_Q[chan][0] = yv_Q[chan][1]; yv_Q[chan][1] = yv_Q[chan][2]; yv_Q[chan][2] = yv_Q[chan][3];
	yv_Q[chan][3] = yv_Q[chan][4]; yv_Q[chan][4] = yv_Q[chan][5]; yv_Q[chan][5] = yv_Q[chan][6];
	yv_Q[chan][6] = (xv_Q[chan][0] + xv_Q[chan][6]) + 6.0 * (xv_Q[chan][1] + xv_Q[chan][5]) + 15.0 * (xv_Q[chan][2] + xv_Q[chan][4])
		+ 20.0 * xv_Q[chan][3]
		+ (-0.0837564796 * yv_Q[chan][0]) + (0.7052741145 * yv_Q[chan][1])
		+ (-2.5294949058 * yv_Q[chan][2]) + (4.9654152288 * yv_Q[chan][3])
		+ (-5.6586671659 * yv_Q[chan][4]) + (3.5794347983 * yv_Q[chan][5]);
	(*data_out_Q) = yv_Q[chan][6];
}

void NPRV_demod_butterw_filter_008bis(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {
	//float A[5];
	//float d1[5];
	//float d2[5];
	//0080 N=6
	//const float A[6] = { 5.49901097478124947782e-02, 4.61318040520628247503e-02, 4.22064051641058007869e-02, };
	//const float d1[6] = { 1.55831204238596288114e+00, 1.30728500308510020567e+00, 1.19604688433384165336e+00, };
	//const float d2[6] = { -7.78272481377212610454e-01, -4.91812219293351393645e-01, -3.64872504990264800995e-01, };
	//0075 N=6
	//const float A[6] = { 0.048766597682, 0.041253539018, 0.037883864513, };
	//const float d1[6] = { 1.594640549248, 1.348967720638, 1.238781243457, };
	//const float d2[6] = { -0.789706939975, -0.513981876710, -0.390316701510, };
	//0075 N=12
	//const float A[7] = { 0.051448050110, 0.046430205329, 0.042696599689, 0.040065980524, 0.038393329075, 0.037581197347, };
	//const float d1[7] = { 1.682322548313, 1.518241822207, 1.396154999865, 1.310135220143, 1.255440450511, 1.228884195907, };
	//const float d2[7] = { -0.888114748751, -0.703962643522, -0.566941398623, -0.470399142240, -0.409013766811, -0.379208985296, };
	//0080 N=12
	//const float A[12] = { 0.058187732822, 0.052219517088, 0.047821813356, 0.044745060707, 0.042798019234, 0.041855248241, };
	//const float d1[12] = { 1.648926419522, 1.479798871080, 1.355176557793, 1.267987411857, 1.212812068717, 1.186095784668, };
	//const float d2[12] = { -0.881677350810, -0.688676939431, -0.546463811215, -0.446967654685, -0.384004145652, -0.353516777630, };
	//0090 N=12
	//const double A[12] = { 0.072748081738, 0.064591435385, 0.058691439619, 0.054617950781, 0.052062868721, 0.050831941606, };
	//const double d1[12] = { 1.578272357338, 1.401313609295, 1.273312980220, 1.184938487349, 1.129505886389, 1.102800876533, };
	//const double d2[12] = { -0.869264684288, -0.659679350835, -0.508078738695, -0.403410290471, -0.337757361274, -0.306128642956, };

	//0090 N=12
	//const float A[12] = { 7.27480817375793231694e-02, 6.45914353850161743198e-02, 5.86914396187050227405e-02, 5.46179507807042080580e-02, 5.20628687210344665015e-02, 5.08319416058265957647e-02, };
	//const float d1[12] = { 1.57827235733814563723e+00, 1.40131360929484927169e+00, 1.27331298022024341776e+00, 1.18493848734861773053e+00, 1.12950588638940563335e+00, 1.10280087653308367379e+00, };
	//const float d2[12] = { -8.69264684288462818884e-01, -6.59679350834913913459e-01, -5.08078738695063591990e-01, -4.03410290471434396231e-01, -3.37757361273543388336e-01, -3.06128642956390084606e-01, };
	//0100 N=12
	//const float A[12] = { 8.86872966271187723741e-02, 7.79563439892779902474e-02, 7.03270220757633707231e-02, 6.51231763130007273999e-02, 6.18851978100506405767e-02, 6.03323990677552893236e-02, };
	//const float d1[12] = { 1.50274156905748812463e+00, 1.32091340180292893258e+00, 1.19164010541006182287e+00, 1.10346473369311759427e+00, 1.04859954914976838403e+00, 1.02228850678889560122e+00, };
	//const float d2[12] = { -8.57490755565963325147e-01, -6.32738777760040949083e-01, -4.72948193713115416781e-01, -3.63957438945120614893e-01, -2.96140340389970946333e-01, -2.63618103059916897291e-01, };
	//0080 N=8
	//const float A[8] = { 5.65333491440992916277e-02, 4.87885143565325585246e-02, 4.41584125105671035372e-02, 4.20012172339315828506e-02, };
	//const float d1[8] = { 1.60204442529166923848e+00, 1.38257097140871776375e+00, 1.25136295060042201932e+00, 1.19023226014034388953e+00, };
	//const float d2[8] = { -8.28177821868066321720e-01, -5.77725028834847886827e-01, -4.27996600642690239180e-01, -3.58237129076070082156e-01, };
	//0090 N=8
	//const float A[8] = { 7.04695348893204093210e-02, 5.99804849633386663599e-02, 5.38462569432033646977e-02, 5.10222597206315131579e-02, };
	//const float d1[8] = { 1.52883919814531243908e+00, 1.30127886724018648756e+00, 1.16819656064807309015e+00, 1.10692983516021059565e+00, };
	//const float d2[8] = { -8.10717337702593909832e-01, -5.41200807093541125248e-01, -3.83581588420886521185e-01, -3.11018874042736648278e-01, };
	//0085 N=8
	const float A[8] = { 6.33388670040658935401e-02, 5.42785635234388635917e-02, 4.89224441330757056634e-02, 4.64422277011591611018e-02, };
	const float d1[8] = { 1.56596911930690607662e+00, 1.34196518407232057157e+00, 1.20954226649643126201e+00, 1.14822221886549469261e+00, };
	const float d2[8] = { -8.19324587323169817310e-01, -5.59079438166076081451e-01, -4.05232043028734056911e-01, -3.33991129670131392526e-01, };
	int i;
	static double w0_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w1_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w2_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w0_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w1_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w2_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	float data_loc;

	data_loc = data_in_I;
	for (i = 0; i < 4; i++) {
		w0_I[chan][i] = d1[i] * w1_I[chan][i] + d2[i] * w2_I[chan][i] + data_loc;
		data_loc = A[i] * (w0_I[chan][i] + 2.0*w1_I[chan][i] + w2_I[chan][i]);
		w2_I[chan][i] = w1_I[chan][i];
		w1_I[chan][i] = w0_I[chan][i];
	}
	(*data_out_I) = data_loc;

	data_loc = data_in_Q;
	for (i = 0; i < 4; i++) {
		w0_Q[chan][i] = d1[i] * w1_Q[chan][i] + d2[i] * w2_Q[chan][i] + data_loc;
		data_loc = A[i] * (w0_Q[chan][i] + 2.0*w1_Q[chan][i] + w2_Q[chan][i]);
		w2_Q[chan][i] = w1_Q[chan][i];
		w1_Q[chan][i] = w0_Q[chan][i];
	}
	(*data_out_Q) = data_loc;
	//printf("I%f Q%f\r\n", (*data_out_I), (*data_out_Q));
}

void NPRV_demod_butterw_filter_016bis(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {
	//0160 N=8
	const float A[8] = { 1.99263833036215975980e-01, 1.57980548512291651964e-01, 1.36358466092187630148e-01, 1.26954790714553594766e-01, };
	const float d1[8] = { 9.20095255100158104788e-01, 7.29470827040958336163e-01, 6.29631457613057832212e-01, 5.86210245827521458928e-01, };
	const float d2[8] = { -7.17150587245021897687e-01, -3.61393021090124888506e-01, -1.75065321981808297291e-01, -9.40294086857358241138e-02, };
	int i;
	static double w0_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w1_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w2_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w0_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w1_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w2_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	float data_loc;

	data_loc = data_in_I;
	for (i = 0; i < 4; i++) {
		w0_I[chan][i] = d1[i] * w1_I[chan][i] + d2[i] * w2_I[chan][i] + data_loc;
		data_loc = A[i] * (w0_I[chan][i] + 2.0*w1_I[chan][i] + w2_I[chan][i]);
		w2_I[chan][i] = w1_I[chan][i];
		w1_I[chan][i] = w0_I[chan][i];
	}
	(*data_out_I) = data_loc;

	data_loc = data_in_Q;
	for (i = 0; i < 4; i++) {
		w0_Q[chan][i] = d1[i] * w1_Q[chan][i] + d2[i] * w2_Q[chan][i] + data_loc;
		data_loc = A[i] * (w0_Q[chan][i] + 2.0*w1_Q[chan][i] + w2_Q[chan][i]);
		w2_Q[chan][i] = w1_Q[chan][i];
		w1_Q[chan][i] = w0_Q[chan][i];
	}
	(*data_out_Q) = data_loc;
	//printf("I%f Q%f\r\n", (*data_out_I), (*data_out_Q));
}

static float A_FiltTest[8] = { 0 };
static float d1_FiltTest[8] = { 0 };
static float d2_FiltTest[8] = { 0 };

void NPRV_demod_BFiltTest_init(double FCoup) {
	int i;
	int n = 8;
	n = n / 2;
	double s = 1000;
	double f = FCoup;
	double a = tan(M_PI*f / s);
	double a2 = a * a;
	double r;

	for (i = 0; i < n; ++i) {
		r = sin(M_PI*(2.0*i + 1.0) / (4.0*n));
		s = a2 + 2.0*a*r + 1.0;
		A_FiltTest[i] = a2 / s;
		d1_FiltTest[i] = 2.0*(1 - a2) / s;
		d2_FiltTest[i] = -(a2 - 2.0*a*r + 1.0) / s;
		//printf("n=%i r:%f s:%f A:%f d1:%f d2:%f\r\n", i, r, s, A[i], d1[i], d2[i]);
		printf("A[%i] = %f; d1[%i] = %f; d2[%i] = %f;\r\n", i, A_FiltTest[i], i, d1_FiltTest[i], i, d2_FiltTest[i]);
	}
}

void NPRV_demod_butterw_filter_test(float data_in_I, float data_in_Q, float* data_out_I, float* data_out_Q, int chan) {

	int i;
	static double w0_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w1_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w2_I[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w0_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w1_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	static double w2_Q[NPRV_MAX_CHANNEL_COUNT][7] = { 0 };
	float data_loc;

	data_loc = data_in_I;
	for (i = 0; i < 4; i++) {
		w0_I[chan][i] = d1_FiltTest[i] * w1_I[chan][i] + d2_FiltTest[i] * w2_I[chan][i] + data_loc;
		data_loc = A_FiltTest[i] * (w0_I[chan][i] + 2.0*w1_I[chan][i] + w2_I[chan][i]);
		w2_I[chan][i] = w1_I[chan][i];
		w1_I[chan][i] = w0_I[chan][i];
	}
	(*data_out_I) = data_loc;

	data_loc = data_in_Q;
	for (i = 0; i < 4; i++) {
		w0_Q[chan][i] = d1_FiltTest[i] * w1_Q[chan][i] + d2_FiltTest[i] * w2_Q[chan][i] + data_loc;
		data_loc = A_FiltTest[i] * (w0_Q[chan][i] + 2.0*w1_Q[chan][i] + w2_Q[chan][i]);
		w2_Q[chan][i] = w1_Q[chan][i];
		w1_Q[chan][i] = w0_Q[chan][i];
	}
	(*data_out_Q) = data_loc;
	//printf("I%f Q%f\r\n", (*data_out_I), (*data_out_Q));
}

void(*NPRV_demod_butterw_filter_tab[2*NPRV_MAX_CHANNEL_COUNT])(float, float, float*, float*, int) = { NPRV_demod_butterw_filter_008 , NPRV_demod_butterw_filter_004 };

void NPRV_init_butterworth_filter(void) {
	int i;
	for (i = 0; i < NPRV_MAX_CHANNEL_COUNT; i++) {
		if (NPRV_channels_SR_std[i] == 0) { // 100kS
			NPRV_demod_butterw_filter_tab[i] = &NPRV_demod_butterw_filter_008;//008 NPRV_demod_butterw_filter_008 !!!
			NPRV_demod_butterw_filter_tab[i + NPRV_MAX_CHANNEL_COUNT] = &NPRV_demod_butterw_filter_010;//010 RX part
		}
		if (NPRV_channels_SR_std[i] == 1) { // 50kS
			NPRV_demod_butterw_filter_tab[i] = &NPRV_demod_butterw_filter_004;//004 !!!
			NPRV_demod_butterw_filter_tab[i + NPRV_MAX_CHANNEL_COUNT] = &NPRV_demod_butterw_filter_005;//005 RX part
		}
		if (NPRV_channels_SR_std[i] == 2) { // 200kS
			NPRV_demod_butterw_filter_tab[i] = &NPRV_demod_butterw_filter_016bis;//004 !!!  NPRV_demod_butterw_filter_test NPRV_demod_butterw_filter_016bis
			NPRV_demod_butterw_filter_tab[i + NPRV_MAX_CHANNEL_COUNT] = &NPRV_demod_butterw_filter_016bis;//005 RX part
		}
	}
}

static float freq_filtered[NPRV_MAX_CHANNEL_COUNT] = { 0 };

float NPRV_freq_extract(float sample_I, float sample_Q, float* amplitude, int chan) {
	static float phase_demod_prev[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	float freq_out, phase_loc, delta_phase;
	(*amplitude) = sqrt((sample_I*sample_I) + (sample_Q*sample_Q));
	if (sample_Q > 0) {
		phase_loc = atan(sample_I / sample_Q);
	}
	else if (sample_Q < 0) {
		phase_loc = atan(sample_I / sample_Q) + M_PI;
	}
	else {//Q = 0
		if (sample_I > 0) { phase_loc = M_PI / 4; }
		else { phase_loc = -M_PI / 4; }
	}

	delta_phase = (phase_demod_prev[chan] - phase_loc);
	phase_demod_prev[chan] = phase_loc;
	if (delta_phase > M_PI) { delta_phase = delta_phase - 2 * M_PI; }
	if (delta_phase < -M_PI) { delta_phase = delta_phase + 2 * M_PI; }
	freq_out = delta_phase / NPRV_demod_period_chan[chan];
	return freq_out;
}

float NPRV_freq_extract_large(float sample_I, float sample_Q, float ampl_inst, float ampl_mean, int chan) {
	static float phase_stored[NPRV_MAX_CHANNEL_COUNT][22] = { 0 };
	static int ampl_OK_stored[NPRV_MAX_CHANNEL_COUNT][22] = { 0 };
	static int index_phase_st[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	float freq_out, phase_loc, phase_prev, delta_phase;
	int ampl_OK, index_loc;
	int period_loc;

	if (NPRV_channels_SR_std[chan] == 2) { period_loc = 5; }
	else { period_loc = 10; }

	index_phase_st[chan]++;
	index_loc = index_phase_st[chan] % period_loc; //%10
	if ((ampl_mean > NPRV_demod_ampl_thrshld) && (ampl_inst > (NPRV_DEMOD_ampl_ratio_freq_comp * ampl_mean))) {
		ampl_OK = 1;
	}
	else { ampl_OK = 0; }
	//phase extraction
	if (ampl_OK == 1) {
		if (sample_Q > 0) {
			phase_loc = atan(sample_I / sample_Q);
		}
		else if (sample_Q < 0) {
			phase_loc = atan(sample_I / sample_Q) + M_PI;
		}
		else {//Q = 0
			if (sample_I > 0) { phase_loc = M_PI / 4; }
			else { phase_loc = -M_PI / 4; }
		}
		//printf("%i PH:%f ", index_loc, phase_loc);
	}
	else { phase_loc = 0; }//useless
	//extracting freq
	
	if ((ampl_OK == 1) && (ampl_OK_stored[chan][index_loc] == 1)) {
		delta_phase = (phase_stored[chan][index_loc] - phase_loc);
		if (delta_phase > M_PI) { delta_phase = delta_phase - 2 * M_PI; }
		if (delta_phase < -M_PI) { delta_phase = delta_phase + 2 * M_PI; }
		if (delta_phase > (M_PI/2)) { delta_phase = delta_phase - M_PI; }
		if (delta_phase < (-M_PI/2)) { delta_phase = delta_phase + M_PI; }
		//printf("PHold:%f delta:%f\r\n", phase_stored[chan][index_loc], delta_phase);
		freq_out = delta_phase / period_loc; // /10
	}
	else { 
		freq_out = 0; 
		//if (ampl_OK == 1) { printf("\r\n"); }
	}
	//storing previous values
	ampl_OK_stored[chan][index_loc] = ampl_OK;
	phase_stored[chan][index_loc] = phase_loc;

	//printf("F:%f\r\n", freq_out);
	return freq_out;
}


void NPRV_demod_PSK_freq_offset_comp(float freq_in, float ampl_inst, float ampl_mean, int chan) {
	static int high_ampl_count[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	if (NPRV_demod_curr_ModulTyp[chan] == NPRV_BPSK) {
		if (ampl_mean > NPRV_demod_ampl_thrshld) {
			if (ampl_inst > (NPRV_DEMOD_ampl_ratio_freq_comp * ampl_mean)) {
				high_ampl_count[chan]++;
				if (high_ampl_count[chan] >= 2) {
					//printf("%f\r\n", freq_in);
					freq_offset_rx[chan] = freq_offset_rx[chan] + (NPRV_DEMOD_PSK_freq_comp_gain * freq_in);
					if (freq_offset_rx[chan] < freq_offset_rx_minlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_minlim[chan]; }
					if (freq_offset_rx[chan] > freq_offset_rx_maxlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_maxlim[chan]; }
				}
			}
			else { high_ampl_count[chan] = 0; }
		} 
		else {// ampl mean too low
			high_ampl_count[chan] = 0;
			freq_offset_rx[chan] = (float)0.95 * freq_offset_rx[chan] + (float)0.05 * offset_freq_rx_init[chan];
		}
	}
}

void NPRV_demod_PSK_freq_offset_comp2(float freq_in, float ampl_mean, int chan) {
	if (NPRV_channels_SR_std[chan] == 2) {
		if ((NPRV_demod_curr_ModulTyp[chan] == NPRV_BPSK) && (freq_in != 0)) {
			//printf("comp\r\n");
			freq_offset_rx[chan] = freq_offset_rx[chan] + (2 * NPRV_DEMOD_PSK_freq_comp_gain * freq_in);
			if (freq_offset_rx[chan] < freq_offset_rx_minlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_minlim[chan]; }
			if (freq_offset_rx[chan] > freq_offset_rx_maxlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_maxlim[chan]; }
		}
		if (ampl_mean < NPRV_demod_ampl_thrshld) {
			freq_offset_rx[chan] = (float)0.9 * freq_offset_rx[chan] + (float)0.1 * offset_freq_rx_init[chan];
		}
	}
	else {
		if ((NPRV_demod_curr_ModulTyp[chan] == NPRV_BPSK) && (freq_in != 0)) {
			//printf("comp\r\n");
			freq_offset_rx[chan] = freq_offset_rx[chan] + (NPRV_DEMOD_PSK_freq_comp_gain * freq_in);
			if (freq_offset_rx[chan] < freq_offset_rx_minlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_minlim[chan]; }
			if (freq_offset_rx[chan] > freq_offset_rx_maxlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_maxlim[chan]; }
		}
		if (ampl_mean < NPRV_demod_ampl_thrshld) {
			freq_offset_rx[chan] = (float)0.95 * freq_offset_rx[chan] + (float)0.05 * offset_freq_rx_init[chan];
		}
	}
		
}

void NPRV_demod_costas(float sampl_I, float sampl_Q, float ampl_inst, float ampl_mean, int chan, int option_loc) {
	float phase_loc, freq_loc, ampl_loc;
	float err_phase_loc = 0;
	//ampl_loc = sqrt((sampl_I*sampl_I) + (sampl_Q*sampl_Q));
	//if (ampl_loc > NPRV_MaxAmpliRX_AGC[chan]) { NPRV_MaxAmpliRX_AGC[chan] = ampl_loc; }
	if (option_loc >= 2) {
		if (sampl_I > 0) {
			phase_loc = atan(sampl_Q / sampl_I);
		}
		else if (sampl_I < 0) {
			phase_loc = atan(sampl_Q / sampl_I) + M_PI;
		}
		else {//I = 0
			if (sampl_Q > 0) { phase_loc = M_PI / 2; }
			else { phase_loc = -M_PI / 2; }
		}
		if (option_loc == 3) { err_phase_loc = phase_loc - (M_PI / 4); }
		if (option_loc == 2) { err_phase_loc = phase_loc - (5 * M_PI / 4); }
		if (err_phase_loc < (-M_PI)) { err_phase_loc = err_phase_loc + 2 * M_PI; }
		if (err_phase_loc > (M_PI)) { err_phase_loc = err_phase_loc - 2 * M_PI; }
		if (err_phase_loc < (-M_PI)) { printf("err too small %f\r\n", err_phase_loc); }
		if (err_phase_loc > (M_PI)) { printf("err too big %f\r\n", err_phase_loc); }
		
		//freq_offset_rx[chan] = freq_offset_rx[chan] + (NPRV_DEMOD_BPSK_costas_kP * err_phase_loc);
		//NPRV_alpha_rx[chan] = NPRV_alpha_rx[chan] + (NPRV_DEMOD_BPSK_costas_kI * err_phase_loc);
		//if (ampl_inst > (NPRV_DEMOD_ampl_ratio_freq_comp * ampl_mean)) {
		NPRV_alpha_costas[chan] = NPRV_alpha_costas[chan] + (NPRV_demod_QPSKins_costas_kI * err_phase_loc);
		//}
	}
	else if (NPRV_demod_curr_ModulTyp[chan] == NPRV_QPSK) { 
		if (sampl_I > 0) {
			phase_loc = atan(sampl_Q / sampl_I);
			if (sampl_Q >= 0) {
				err_phase_loc = phase_loc - (M_PI / 4);
			}
			else {
				err_phase_loc = phase_loc + (M_PI / 4);
			}
		}
		else if (sampl_I < 0) {
			phase_loc = atan(sampl_Q / sampl_I);
			if (sampl_Q >= 0) {
				err_phase_loc = phase_loc + (M_PI / 4);
			}
			else {
				err_phase_loc = phase_loc - (M_PI / 4);
			}
		}
		else {
			err_phase_loc = 0;
		}
		/*
		if (amplitude > 150) {
			freq_offset_rx[chan] = freq_offset_rx[chan] + (NPRV_DEMOD_QPSK_costas_kP * err_phase_loc);
		}
		else {
			freq_offset_rx[chan] = (float)0.95 * freq_offset_rx[chan] + (float)0.05 * offset_freq_rx_init[chan];
		}
		*/
		//NPRV_alpha_rx[chan] = NPRV_alpha_rx[chan] + (NPRV_DEMOD_QPSK_costas_kI * err_phase_loc);
		if (ampl_inst > (NPRV_DEMOD_ampl_ratio_freq_comp * ampl_mean)) {
			if (NPRV_channels_SR_std[chan] == 2) {
				NPRV_alpha_costas[chan] = NPRV_alpha_costas[chan] + (2 * NPRV_DEMOD_QPSK_costas_kI * err_phase_loc);
			}
			else {
				NPRV_alpha_costas[chan] = NPRV_alpha_costas[chan] + (NPRV_DEMOD_QPSK_costas_kI * err_phase_loc);
			}
		}
	}
	else if (NPRV_demod_curr_ModulTyp[chan] == NPRV_BPSK) {
		if (sampl_I > 0) {
			phase_loc = atan(sampl_Q / sampl_I);
		}
		else if (sampl_I < 0) {
			phase_loc = atan(sampl_Q / sampl_I) + M_PI;
		}
		else {//I = 0
			if (sampl_Q > 0) { phase_loc = M_PI / 2; }
			else { phase_loc = -M_PI / 2; }
		}
		if (sampl_Q > (-sampl_I)) {
			err_phase_loc = phase_loc - (M_PI / 4);
		}
		else {
			err_phase_loc = phase_loc - (5 * M_PI / 4);
			//if (err_phase_loc < (-M_PI)) { err_phase_loc = err_phase_loc + 2 * M_PI; }
			if (err_phase_loc < (-M_PI/2)) { err_phase_loc = err_phase_loc + 2 * M_PI; }
		}
		//freq_offset_rx[chan] = freq_offset_rx[chan] + (NPRV_DEMOD_BPSK_costas_kP * err_phase_loc);
		//NPRV_alpha_rx[chan] = NPRV_alpha_rx[chan] + (NPRV_DEMOD_BPSK_costas_kI * err_phase_loc);
		if (ampl_inst > (NPRV_DEMOD_ampl_ratio_freq_comp * ampl_mean)) {
			if (NPRV_channels_SR_std[chan] == 2) {
				NPRV_alpha_costas[chan] = NPRV_alpha_costas[chan] + (2 * NPRV_DEMOD_BPSK_costas_kI * err_phase_loc);
			}
			else {
				NPRV_alpha_costas[chan] = NPRV_alpha_costas[chan] + (NPRV_DEMOD_BPSK_costas_kI * err_phase_loc);
			}
		}
	}
	//printf("err phase %e new phase%e\r\n", err_phase_loc, NPRV_alpha_costas[chan]);
	//if (freq_offset_rx[chan] < freq_offset_rx_minlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_minlim[chan]; }
	//if (freq_offset_rx[chan] > freq_offset_rx_maxlim[chan]) { freq_offset_rx[chan] = freq_offset_rx_maxlim[chan]; }
	//ampl_loc = sqrt((sampl_I*sampl_I) + (sampl_Q*sampl_Q));
	//if (ampl_loc > NPRV_MaxAmpliRX_AGC[chan]) { NPRV_MaxAmpliRX_AGC[chan] = ampl_loc; }
}

void NPRV_demod_store_eye_data(float sample_I, float sample_Q, float sample_BPSK, int top_symbol, float amplitude) {
	static int sample_count = 0;
	static int image_count = 0;
	int table_index;
	if (amplitude > NPRV_demod_ampl_thrshld) {
		sample_count++;
		if ((top_symbol == 1) && (sample_count > 25)) {
			image_count++;
			sample_count = 0;
		}
		table_index = (40 * (image_count % 40)) + (sample_count % 40);
		NPRV_eye_data_BPSK[table_index] = sample_BPSK;
		NPRV_eye_data_I[table_index] = sample_I;
		NPRV_eye_data_Q[table_index] = sample_Q;
	}
}

float NPRV_demod_PSK_amplitude(float data_I, float data_Q, float* ampl_instant, int chan) {
	static float ampl_mean[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	float ampl_loc;
	ampl_loc = sqrt((data_I*data_I) + (data_Q*data_Q)); //!!!remove sqrt in the future. 
	if (ampl_loc > NPRV_MaxAmpliRX_AGC[chan]) { NPRV_MaxAmpliRX_AGC[chan] = ampl_loc; }
	if (ampl_loc > ampl_mean[chan]) { ampl_mean[chan] = (0.5 * ampl_mean[chan]) + (0.5 * ampl_loc); }
	else { ampl_mean[chan] = 0.998 * ampl_mean[chan] + 0.002 * ampl_loc; }

	(*ampl_instant) = ampl_loc;
	return ampl_mean[chan];
}

static uint32_t Symb_count_tot_test_MER = 0;
static uint32_t Symb_count_err_test_MER = 0;

float NPRV_demod_count_MER_test(void) {
	float answer;
	if (Symb_count_tot_test_MER > 0) { answer = (100 * (float)Symb_count_err_test_MER) / (float)Symb_count_tot_test_MER; }
	else {answer = 200; }
	Symb_count_tot_test_MER = 0;
	Symb_count_err_test_MER = 0;
	return answer;
}

void NPRV_demod_PSK_Symbol_extract(float sampl_I, float sampl_Q, float amplitude, char* color_loc, float* data_display, int chan) {
	static int prev_I[NPRV_MAX_CHANNEL_COUNT];
	static int prev_Q[NPRV_MAX_CHANNEL_COUNT];
	static int prev_BPSK[NPRV_MAX_CHANNEL_COUNT];
	static float rx_symb_clk[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static float rx_symb_period_mes[NPRV_MAX_CHANNEL_COUNT][256] = { 0 };
	static uint8_t rx_symb_period_indx[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static uint32_t rx_symb_instants[NPRV_MAX_CHANNEL_COUNT][256] = { 0 };
	static uint8_t rx_symb_instants_indx[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	float symbol_duration = 10.0;
	float half_symbol_duration = 5.00;
	static float real_symb_duration[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static int demod_post_sync_word_prev[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static uint8_t dumy_symb_val[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	float phase_err, clk_correction;
	float sampl_BPSK;
	int top_symbol = 0;
	unsigned char decoded_symbol;
	unsigned long int loc_timer;
	uint8_t symb_temp;
	int symb_decod_answ;
	uint8_t index_loc;
	uint8_t i;
	float correction_loc;
	
	if ( NPRV_channels_SR_std[chan] == 2 ) {
		symbol_duration = 5;
		half_symbol_duration = 2.5;
	}

	sampl_BPSK = sampl_I + sampl_Q;
	data_display[0] = sampl_I;
	data_display[8] = sampl_I;
	data_display[2] = sampl_Q;
	data_display[10] = sampl_Q;
	rx_symb_clk[chan] = rx_symb_clk[chan] + 1.0; 
	if ((NPRV_demod_post_sync_word[chan] == 0) || (NPRV_DEMOD_CLK_ESTIM_PREAM == 0)) {//transition new symbol
		if (rx_symb_clk[chan] > (symbol_duration - 0.5)) {
			rx_symb_clk[chan] = rx_symb_clk[chan] - symbol_duration;
			top_symbol = 1;
		}
	}
	else {
		if (rx_symb_clk[chan] > (symbol_duration - 0.5)) {
		//if (rx_symb_clk[chan] > (real_symb_duration[chan] - 0.5)) {
			rx_symb_clk[chan] = rx_symb_clk[chan] - symbol_duration;
			//rx_symb_clk[chan] = rx_symb_clk[chan] - real_symb_duration[chan];
			top_symbol = 1;
		}
	}
	if (top_symbol == 1) {//transition new symbol
		(*color_loc) = 1;
		data_display[1] = sampl_I;
		data_display[9] = sampl_I;
		data_display[3] = sampl_Q;
		data_display[11] = sampl_Q;
		//rx_symb_clk[chan] = rx_symb_clk[chan] - symbol_duration;
		top_symbol = 1;
		// period computation
		rx_symb_period_indx[chan]++;
		rx_symb_period_mes[chan][rx_symb_period_indx[chan]] = 0;
		//appel a symbol extract
		loc_timer = NPRV_G_rx_macro_timer[chan] + NPRV_G_rx_micro_timer[chan];

		rx_symb_instants[chan][rx_symb_instants_indx[chan]] = loc_timer;
		rx_symb_instants_indx[chan]++;
		if ((NPRV_demod_post_sync_word[chan] == 1) && (demod_post_sync_word_prev[chan] == 0)) {//transition sync to header
			index_loc = (rx_symb_instants_indx[chan] - 16) % 256;
			//printf("T:%i %i\r\n", loc_timer - rx_symb_instants[chan][index_loc], index_loc);
			dumy_symb_val[chan] = 0;
		}
		demod_post_sync_word_prev[chan] = NPRV_demod_post_sync_word[chan];

		decoded_symbol = 0;
		if ((sampl_Q + prev_Q[chan]) >= 0) { decoded_symbol = 0x01; }
		if ((sampl_I + prev_I[chan]) >= 0) { decoded_symbol = decoded_symbol + 0x02; }
		if ((sampl_BPSK + prev_BPSK[chan]) >= 0) { decoded_symbol = decoded_symbol + 0x04; }
		//test MER
		if (NPRV_demod_post_sync_word[chan] == 2) {
			Symb_count_tot_test_MER++;
			symb_temp = decoded_symbol & 0x03;
			if ((symb_temp == 0x01) || (symb_temp == 0x02)) { Symb_count_err_test_MER++; }
		}
		//fin test MER
		if ((NPRV_demod_curr_ModulTyp[chan] == NPRV_QPSK) && (NPRV_demod_costas_on_symb == 1)) {
			NPRV_demod_costas(sampl_I, sampl_Q, 400, 300, chan, 0);//!!! test
		}
		symb_decod_answ = NPRV_rx_symb_decode(decoded_symbol, loc_timer, amplitude, freq_offset_rx[chan], chan);

		if (symb_decod_answ == 1) {//save measured period
			dumy_symb_val[chan] = 0;
			//uint8_t i;
			correction_loc = 0;
			index_loc = rx_symb_period_indx[chan];
			//printf("index_loc:%i\r\n", rx_symb_period_indx[chan]);
			for (i = 0; i < 50; i++) {
				correction_loc = correction_loc + rx_symb_period_mes[chan][(index_loc - 0 - i)%256];
				//printf("%i:%f ", ((index_loc - i) % 256), rx_symb_period_mes[chan][(index_loc - i)%256]);
			}
			real_symb_duration[chan] = symbol_duration + correction_loc / 50;
			//printf("SYNC period:%f\r\n", real_symb_duration[chan]);
		}
		if (symb_decod_answ >= 2) {// synchro on dummy QPSK symbol
			
			if (dumy_symb_val[chan] == 1) {
				//if (symb_decod_answ == 2) { symb_decod_answ = 3; }
				//else if (symb_decod_answ == 3) { symb_decod_answ = 2; }
				dumy_symb_val[chan] = 0;
			}
			else { 
				//if (symb_decod_answ == 2) { symb_decod_answ = 3; }
				//else if (symb_decod_answ == 3) { symb_decod_answ = 2; }
				dumy_symb_val[chan] = 1; 
			}
			NPRV_demod_costas(sampl_I, sampl_Q, 400, 300, chan, symb_decod_answ);
			//printf("decod %i\r\n", symb_decod_answ);
		}
	}
	else {
		(*color_loc) = 0;
	}
	if ((NPRV_demod_curr_ModulTyp[chan] == NPRV_BPSK) && ((NPRV_demod_post_sync_word[chan] == 0) || (NPRV_DEMOD_CLK_ESTIM_PREAM == 0))) {
		if (((sampl_BPSK > 0) && (prev_BPSK[chan] < 0)) || ((sampl_BPSK < 0) && (prev_BPSK[chan] > 0))) {//Zero crossing detection
			phase_err = rx_symb_clk[chan] - half_symbol_duration;
			clk_correction = NPRV_DEMOD_ClkRecov_BPSK_kP * phase_err;
			rx_symb_clk[chan] = rx_symb_clk[chan] - clk_correction;
			rx_symb_period_mes[chan][rx_symb_period_indx[chan]] = clk_correction;
			//if (clk_correction > 3) { printf("Clk_corr:%f\r\n", clk_correction); }
		}
	}
	if ((NPRV_demod_curr_ModulTyp[chan] == NPRV_QPSK) && (NPRV_DEMOD_CLK_ESTIM_PREAM == 0)) {
		if (((sampl_I > 0) && (prev_I[chan] < 0)) || ((sampl_I < 0) && (prev_I[chan] > 0))) {//zero crossing detected
			phase_err = rx_symb_clk[chan] - half_symbol_duration;
			clk_correction = NPRV_DEMOD_ClkRecov_QPSK_kP * phase_err;
			rx_symb_clk[chan] = rx_symb_clk[chan] - clk_correction;
			rx_symb_period_mes[chan][rx_symb_period_indx[chan]] = clk_correction;
		}
		if (((sampl_Q > 0) && (prev_Q[chan] < 0)) || ((sampl_Q < 0) && (prev_Q[chan] > 0))) {//zero crossing detected
			phase_err = rx_symb_clk[chan] - half_symbol_duration;
			clk_correction = NPRV_DEMOD_ClkRecov_QPSK_kP * phase_err;
			rx_symb_clk[chan] = rx_symb_clk[chan] - clk_correction;
			rx_symb_period_mes[chan][rx_symb_period_indx[chan]] = rx_symb_period_mes[chan][rx_symb_period_indx[chan]] + clk_correction;
		}
	}
	prev_BPSK[chan] = sampl_BPSK;
	prev_I[chan] = sampl_I;
	prev_Q[chan] = sampl_Q;
	if (chan == 0) {
		NPRV_demod_store_eye_data(sampl_I, sampl_Q, sampl_BPSK, top_symbol, amplitude); 
		//printf("chan:%i\r\n", chan);
	}
}

static float NPRV_demod_saved_IQ[70000] = { 0 };
static char NPRV_demod_saved_IQ_color[40000] = { 0 };

void NPRV_demod_indiv(short int* IQ_samples_in, short int* IQ_samples_out, float* data_for_display, size_t length, int chan) {
	int j;
	float current_I, current_Q;
	float modified_I, modified_Q,modified2_I, modified2_Q, modified3_I, modified3_Q;
	char color_loc;
	short int* current_datain;
	short int* current_dataout;
	float* current_data_displ;
	float* current_saved_IQ;
	char* current_saved_IQ_color;
	float freq_A;
	float amplitude_A;
	float ampl_mean, ampl_instant;
	
	current_datain = IQ_samples_in;
	current_dataout = IQ_samples_out;
	current_data_displ = data_for_display;
	current_saved_IQ = NPRV_demod_saved_IQ;
	current_saved_IQ_color = NPRV_demod_saved_IQ_color;
	for (NPRV_G_rx_micro_timer[chan] = 0; NPRV_G_rx_micro_timer[chan] < length; NPRV_G_rx_micro_timer[chan]++) {

		current_I = current_datain[0];
		current_Q = current_datain[1];
		NPRV_demod_freq_offset(current_I, current_Q, &modified_I, &modified_Q, chan);
		if (CONF_NPRV_filter_RX == 1) {
			NPRV_demod_butterw_filter_tab[chan](modified_I, modified_Q, &modified2_I, &modified2_Q, chan);
		}
		else { 
			modified2_I = modified_I; 
			modified2_Q = modified_Q;
		}


		if ((NPRV_G_rx_micro_timer[chan] % NPRV_demod_period_chan[chan]) == 0) {

			ampl_mean = NPRV_demod_PSK_amplitude(modified2_I, modified2_Q, &ampl_instant, chan);
			//freq_A = NPRV_freq_extract(modified2_I, modified2_Q, &amplitude_A, chan);//selection filtered IQ or not
			freq_A = NPRV_freq_extract_large(modified2_I, modified2_Q, ampl_instant, ampl_mean, chan);//selection filtered IQ or not
			NPRV_demod_PSK_freq_offset_comp2(freq_A, ampl_mean, chan);
			//modified3_I = modified2_I;
			//modified3_Q = modified2_Q;
			NPRV_demod_phase_offset(modified2_I, modified2_Q, &modified3_I, &modified3_Q, chan);
			if ((NPRV_demod_curr_ModulTyp[chan] == NPRV_BPSK) || (NPRV_demod_costas_on_symb == 0)) {
				NPRV_demod_costas(modified3_I, modified3_Q, ampl_instant, ampl_mean, chan, 0);
			}
			NPRV_demod_PSK_Symbol_extract(modified3_I, modified3_Q, ampl_mean, &color_loc, current_data_displ, chan);
		}
#ifdef NPRV_DEMOD_DATA_DISPLAY
		current_data_displ = current_data_displ + 8;
#endif /* end demod data display */
		if (chan == 0) {
			current_dataout[0] = modified2_I;
			current_dataout[1] = modified2_Q;
			if (ampl_mean > NPRV_demod_ampl_thrshld) {
				current_saved_IQ[0] = modified3_I;//normaly 3_IQ
				current_saved_IQ[1] = modified3_Q;
				current_saved_IQ_color[0] = color_loc;
				current_saved_IQ = current_saved_IQ + 2;
				current_saved_IQ_color++;
			}
		}
		current_datain = current_datain + 2;
		current_dataout = current_dataout + 2;
	}
}

void NPRV_demod_total(short int* IQ_samples_in, short int* IQ_samples_out, float* data_for_display, size_t length) {
	int channel;
	static int count_demod = 0;
	float* toto;
	for (channel = 0; channel < CONF_NPRV_channel_count; channel++) {
		NPRV_demod_indiv(IQ_samples_in, IQ_samples_out, data_for_display, length, channel);
		NPRV_G_rx_macro_timer[channel] = NPRV_G_rx_macro_timer[channel] + length;
	}
	for (channel = CONF_NPRV_channel_count; channel < NPRV_MAX_CHANNEL_COUNT; channel++) {
		NPRV_G_rx_macro_timer[channel] = NPRV_G_rx_macro_timer[channel] + length;
	}
	count_demod++;
	if (count_demod % 3 == 0) {
		NPRV_display_IQ(NPRV_demod_saved_IQ, NPRV_demod_saved_IQ_color);
		NPRV_display_eye(NPRV_eye_data_I, NPRV_eye_data_Q, NPRV_eye_data_BPSK);
	}
}

static float NPRV_deltaf_mod[NPRV_MAX_CHANNEL_COUNT] = { 0 };



static float NPRV_mod_cos_table[20] = { 0, 0.024471742, 0.095491503, 0.206107374, 0.345491503, 0.5,
										0.654508497, 0.793892626, 0.904508497, 0.975528258, 1 };

static float NPRV_mod_cos_table2[10][30] = {
	{ 0, 0.024471742, 0.095491503, 0.206107374, 0.345491503, 0.5,0.654508497, 0.793892626, 0.904508497, 0.975528258, 1 } ,
	{0, 0.00615583, 0.024471742, 0.054496738, 0.095491503, 0.146446609, 0.206107374, 0.27300475, 0.345491503, 0.421782767, 0.5, 0.578217233, 0.654508497, 0.72699525, 0.793892626, 0.853553391, 0.904508497, 0.945503262, 0.975528258, 0.99384417, 1}, 
	{0, 0.095491503, 0.345491503, 0.654508497, 0.904508497, 1}
};

static float *NPRV_mod_cos_table3[NPRV_MAX_CHANNEL_COUNT];
const float CONST_mod_symb_duration_param[5] = { 10, 20 , 5};
static int NPRV_mod_symb_duration[NPRV_MAX_CHANNEL_COUNT] = { 10 };

void NPRV_mod_init(void) {
	int i;
	for (i = 0; i < NPRV_MAX_CHANNEL_COUNT; i++) {
		//mod_center_freq[i] = 2 * M_PI * NPRV_channels_freq[i];
		if (NPRV_channels_SR_std[i] < 6) {
			NPRV_deltaf_mod[i] = CONST_deltaf_param[NPRV_channels_SR_std[i]];
			NPRV_mod_cos_table3[i] = NPRV_mod_cos_table2[NPRV_channels_SR_std[i]];
			NPRV_mod_symb_duration[i] = CONST_mod_symb_duration_param[NPRV_channels_SR_std[i]];
		}
	}

}



int NPRV_mod_total(short int* IQ_samples_out, size_t length) {
	//static int sample_count = 0;
	static uint32_t sample_count = 0;
	static float phase[NPRV_MAX_CHANNEL_COUNT] = { 0, 0, 0, 0, 0, 0 };
	static float mod_ampl[20] = { 0, 0, 0, 0, 0, 0 };
	static float mod_ampl_var[20] = { 0, 0, 0, 0, 0, 0 };
	static float new_I_raw[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static float new_Q_raw[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static float prev_I_raw[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static float prev_Q_raw[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	float curr_I, curr_Q;
	float ampl_loc = -2;
	float k_prev_symbol, k_new_symbol, current_freq;
	short int* current_IQ;
	short int current_I, current_Q;
	float chan_I, chan_Q;
	int i, j, chan, chan_mod;
	unsigned char loc_symbol;
	int TX_need = 0;
	current_IQ = IQ_samples_out;
	for (i = 0; i < length; i++) {
		sample_count++;
		if ((sample_count % 5) == 0) {//transition between 2 symbols previously 10
			NPRV_G_tx_timer = NPRV_G_tx_timer + 5;//10 microsecond step
			for (chan = 0; chan < CONF_NPRV_channel_count; chan++) {
				if ((sample_count % NPRV_mod_symb_duration[chan]) == 0) {
					prev_I_raw[chan] = new_I_raw[chan];
					prev_Q_raw[chan] = new_Q_raw[chan];
					//printf("get new symb chan %i\r\n", chan);
					loc_symbol = NPRV_mod_get_new_symbol(&ampl_loc, mod_ampl_var + chan, chan);
					if (ampl_loc > -2) {
						mod_ampl[chan] = ampl_loc;
					}
					if ((loc_symbol & 0x01) == 1) { new_Q_raw[chan] = 1; }
					else { new_Q_raw[chan] = -1; }
					if ((loc_symbol & 0x02) == 2) { new_I_raw[chan] = 1; }
					else { new_I_raw[chan] = -1; }
					if (mod_ampl[chan] > 10) { TX_need = 1; }
				}
			}
		}
		current_I = 0;
		current_Q = 0;

		for (chan = 0; chan < CONF_NPRV_channel_count; chan++) {
			if (TX_curr_status[chan] >= 1) {
				k_new_symbol = NPRV_mod_cos_table3[chan][sample_count % NPRV_mod_symb_duration[chan]];
				k_prev_symbol = 1 - k_new_symbol;
				curr_I = k_prev_symbol * prev_I_raw[chan] + k_new_symbol * new_I_raw[chan];
				curr_Q = k_prev_symbol * prev_Q_raw[chan] + k_new_symbol * new_Q_raw[chan];
				//curr_I = new_I_raw[chan];
				//curr_Q = new_Q_raw[chan];
				mod_ampl[chan] = mod_ampl[chan] + mod_ampl_var[chan];
				curr_I = mod_ampl[chan] * curr_I;
				curr_Q = mod_ampl[chan] * curr_Q;
				NPRV_mod_freq_offset(curr_I, curr_Q, &chan_I, &chan_Q, chan);
				current_I = current_I + chan_I;
				current_Q = current_Q + chan_Q;
				//if ((TX_curr_status[chan] >= 2) &&(NPRV_G_tx_timer < 809000)) {
					//printf("%i;%i;%i\r\n", NPRV_G_tx_timer, current_I, current_Q);
					//printf("%i;%i;%i\r\n", NPRV_G_tx_timer, current_I, current_Q);
				//}
				//if ((mod_ampl[chan] > 100) && (mod_ampl[chan] < 12000)) {
					//printf("T:%i AMP:%f\r\n", NPRV_G_tx_timer, mod_ampl[chan]);
					//NPRV_HMI_new_TXT_message();
				//}
			}
		}
		//if (TX_need == 1) {
		//	printf("I:%i Q:%i\r\n", (int)current_I, (int)current_Q);
		//}
		current_IQ[0] = current_I;
		current_IQ[1] = current_Q;
		current_IQ = current_IQ + 2; 
	}
	return TX_need;
}



static float raised_cosin_coef[3][128] = {
	{0, 0.0007161, 0.0030188, 0.006884, 0.0118817, 0.0171489, 0.0214386, 0.0232543, 0.021063, 0.0135668, 0, -0.0195911, -0.0441321, -0.0714131, -0.0981227, -0.1200422, -0.1323963, -0.1303353, -0.1095008, -0.0666121, 0, 0.0899847, 0.2007514, 0.3274812, 0.4633871, 0.6002109, 0.7289115, 0.8404773, 0.9267741, 0.9813349, 1, 0.9813349, 0.9267741, 0.8404773, 0.7289115, 0.6002109, 0.4633871, 0.3274812, 0.2007514, 0.0899847, 0, -0.0666121, -0.1095008, -0.1303353, -0.1323963, -0.1200422, -0.0981227, -0.0714131, -0.0441321, -0.0195911, 0, 0.0135668, 0.021063, 0.0232543, 0.0214386, 0.0171489, 0.0118817, 0.006884, 0.0030188, 0.0007161, 0} ,
	{0,0.0001719,0.0007161,0.0016619,0.0030188,0.0047728,0.006884,0.0092852,0.0118817,0.0145517,0.0171489,0.0195057,0.0214386,0.0227536,0.0232543,0.0227496,0.021063,0.0180419,0.0135668,0.0075613,0,-0.0090837,-0.0195911,-0.031354,-0.0441321,-0.0576127,-0.0714131,-0.0850851,-0.0981227,-0.1099717,-0.1200422,-0.1277226,-0.1323963,-0.1334586,-0.1303353,-0.1225017,-0.1095008,-0.0909609,-0.0666121,-0.0363005,0,0.0421791,0.0899847,0.1430211,0.2007514,0.2625037,0.3274812,0.394776,0.4633871,0.5322401,0.6002109,0.6661505,0.7289115,0.787375,0.8404773,0.8872361,0.9267741,0.9583417,0.9813349,0.995311,1,0.995311,0.9813349,0.9583417,0.9267741,0.8872361,0.8404773,0.787375,0.7289115,0.6661505,0.6002109,0.5322401,0.4633871,0.394776,0.3274812,0.2625037,0.2007514,0.1430211,0.0899847,0.0421791,0,-0.0363005,-0.0666121,-0.0909609,-0.1095008,-0.1225017,-0.1303353,-0.1334586,-0.1323963,-0.1277226,-0.1200422,-0.1099717,-0.0981227,-0.0850851,-0.0714131,-0.0576127,-0.0441321,-0.031354,-0.0195911,-0.0090837,0,0.0075613,0.0135668,0.0180419,0.021063,0.0227496,0.0232543,0.0227536,0.0214386,0.0195057,0.0171489,0.0145517,0.0118817,0.0092852,0.006884,0.0047728,0.0030188,0.0016619,0.0007161,0.0001719,0} ,
	{0,0.0030188,0.0118817,0.0214386,0.021063,0,-0.0441321,-0.0981227,-0.1323963,-0.1095008,0,0.2007514,0.4633871,0.7289115,0.9267741,1,0.9267741,0.7289115,0.4633871,0.2007514,0,-0.1095008,-0.1323963,-0.0981227,-0.0441321,0,0.021063,0.0214386,0.0118817,0.0030188,0}
};



int NPRV_mod_total_2(short int* IQ_samples_out, size_t length) {// with raised cosine filter
	static float phase[NPRV_MAX_CHANNEL_COUNT] = { 0, 0, 0, 0, 0, 0 };

	static uint8_t n_symb[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static uint8_t n_sampl[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	static char I_symb[NPRV_MAX_CHANNEL_COUNT][256] = { 0 };//-1 or +1
	static char Q_symb[NPRV_MAX_CHANNEL_COUNT][256] = { 0 };
	static float mod_ampl_saved[NPRV_MAX_CHANNEL_COUNT][256] = { -20, -20, -20, -20, -20, -20 };
	static float mod_ampl_var_saved[NPRV_MAX_CHANNEL_COUNT][256] = { 0, 0, 0, 0, 0, 0 };
	static float mod_ampl[NPRV_MAX_CHANNEL_COUNT] = { 0, 0, 0, 0, 0, 0 };
	static float mod_ampl_var[NPRV_MAX_CHANNEL_COUNT] = { 0, 0, 0, 0, 0, 0 };

	int chan_std_loc;
	uint8_t n_symb_loc = 0;
	uint8_t n_sampl_loc = 0;
	int loc_period = 0;
	float loc_ampl = -2;
	float curr_I, curr_Q;
	float k_prev_symbol, k_new_symbol, current_freq;
	short int* current_IQ;
	short int current_I, current_Q;
	float chan_I, chan_Q;
	int i, j, chan, chan_mod;
	unsigned char loc_symbol;
	int TX_need = 0;
	current_IQ = IQ_samples_out;
	for (i = 0; i < length; i++) {

		NPRV_G_tx_timer = NPRV_G_tx_timer + 1;// +10;//10 microsecond step
		for (chan = 0; chan < CONF_NPRV_channel_count; chan++) {
			n_sampl[chan]++;
			if (n_sampl[chan] == NPRV_mod_symb_duration[chan]) {
				n_symb[chan]++;
				n_sampl[chan] = 0;//!!! 0
				n_symb_loc = n_symb[chan];

				loc_symbol = NPRV_mod_get_new_symbol(&mod_ampl_saved[chan][n_symb_loc], &mod_ampl_var_saved[chan][n_symb_loc], chan);

				if ((loc_symbol & 0x01) == 1) { Q_symb[chan][n_symb_loc] = 1; }
				else { Q_symb[chan][n_symb_loc] = -1; }
				if ((loc_symbol & 0x02) == 2) { I_symb[chan][n_symb_loc] = 1; }
				else { I_symb[chan][n_symb_loc] = -1; }
				loc_ampl = mod_ampl_saved[chan][uint8_t(n_symb_loc - 3)];
				if (loc_ampl > -2) { mod_ampl[chan] = loc_ampl; }
				mod_ampl_var[chan] = mod_ampl_var_saved[chan][uint8_t(n_symb_loc - 3)];
				if (mod_ampl[chan] > 10) { TX_need = 1; }//??? comment faire ??? 
			}
		}

		current_I = 0;
		current_Q = 0;

		for (chan = 0; chan < CONF_NPRV_channel_count; chan++) {
			if (TX_curr_status[chan] >= 1) { // ???
				chan_std_loc = NPRV_channels_SR_std[chan];
				n_symb_loc = n_symb[chan];
				n_sampl_loc = n_sampl[chan];
				loc_period = NPRV_mod_symb_duration[chan];
				curr_I =
					I_symb[chan][n_symb_loc] * raised_cosin_coef[chan_std_loc][n_sampl_loc] +
					I_symb[chan][uint8_t(n_symb_loc - 1)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + loc_period] +
					I_symb[chan][uint8_t(n_symb_loc - 2)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 2 * loc_period] +
					I_symb[chan][uint8_t(n_symb_loc - 3)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 3 * loc_period] +
					I_symb[chan][uint8_t(n_symb_loc - 4)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 4 * loc_period] +
					I_symb[chan][uint8_t(n_symb_loc - 5)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 5 * loc_period];
				curr_Q =
					Q_symb[chan][n_symb_loc] * raised_cosin_coef[chan_std_loc][n_sampl_loc] +
					Q_symb[chan][uint8_t(n_symb_loc - 1)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + loc_period] +
					Q_symb[chan][uint8_t(n_symb_loc - 2)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 2 * loc_period] +
					Q_symb[chan][uint8_t(n_symb_loc - 3)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 3 * loc_period] +
					Q_symb[chan][uint8_t(n_symb_loc - 4)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 4 * loc_period] +
					Q_symb[chan][uint8_t(n_symb_loc - 5)] * raised_cosin_coef[chan_std_loc][n_sampl_loc + 5 * loc_period];
				
				mod_ampl[chan] = mod_ampl[chan] + mod_ampl_var[chan];
				curr_I = mod_ampl[chan] * curr_I;
				curr_Q = mod_ampl[chan] * curr_Q;
				NPRV_mod_freq_offset(curr_I, curr_Q, &chan_I, &chan_Q, chan);
				current_I = current_I + chan_I;
				current_Q = current_Q + chan_Q;
				//if ((NPRV_G_tx_timer > 800000) &&(NPRV_G_tx_timer < 809000)) {
					//printf("%i;%i;%i\r\n", NPRV_G_tx_timer, current_I, current_Q);
					//printf("%i;%i;%f;%f\r\n", NPRV_G_tx_timer, current_I, mod_ampl_saved[chan][n_symb_loc], mod_ampl[chan]);
				//}
			}
		}
		//if (TX_need == 1) {
		//	printf("I:%i Q:%i\r\n", (int)current_I, (int)current_Q);
		//}
		current_IQ[0] = current_I;
		current_IQ[1] = current_Q;
		current_IQ = current_IQ + 2;
	}
	return TX_need;
}


