// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include <SFML/System.hpp>
#include <SoapySDR/Device.h>
#include <SoapySDR/Formats.h>
#include <SFML/Network.hpp>
#include "NPRV_coder_decoder.h"
#include "NPRV_L1_L2_radio.h"
#include "Eth_IPv4.h"
#include "DHCP_ARP.h"
#include "NPRV_sat_emul.h"
#include "NPRV_HMI.h"
#include "NPRV_signaling.h"
#include "NPRV_config.h"
#pragma warning(disable: 4996)

#ifdef NPRV_linux
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#endif

#ifdef NPRV_windows
#include <windows.h>
#include <stdio.h>
#endif

const int emul_chan_nb = 6;//10

static unsigned char RX_emul_symb[12][6100] = { 0 };
static unsigned char TX_emul_symb[12][6100] = { 0 };
static long unsigned int NPRV_emul_rx_macro_timer = 0;
static int datagram_size = 0;
static float tx_emul_ampl[12] = { 0, 0, 0, 0, 0 };
static float tx_emul_ampl_var[12] = { 0, 0, 0, 0, 0 };
sf::TcpSocket socket_loc;
static int Kill_requested_emu = 0;

void termination_handler_emu(int signum)
{
	Kill_requested_emu = 1;
}

#ifdef NPRV_windows
BOOL WINAPI CtrlHandler_emu(DWORD fdwCtrlType)
{
	switch (fdwCtrlType)
	{
		// Handle the CTRL-C signal.
	case CTRL_C_EVENT:
		printf("Ctrl-C event\n\n");
		Kill_requested_emu = 1;
		return TRUE;

		// CTRL-CLOSE: confirm that the user wants to exit.
	case CTRL_CLOSE_EVENT:
		printf("Ctrl-Close event\n\n");
		Kill_requested_emu = 1;
		return TRUE;

		// Pass other signals to the next handler.
	case CTRL_BREAK_EVENT:
		printf("Ctrl-Break event\n\n");
		Kill_requested_emu = 1;
		return FALSE;

	case CTRL_LOGOFF_EVENT:
		printf("Ctrl-Logoff event\n\n");
		Kill_requested_emu = 1;
		return FALSE;

	case CTRL_SHUTDOWN_EVENT:
		printf("Ctrl-Shutdown event\n\n");
		Kill_requested_emu = 1;
		return FALSE;

	default:
		return FALSE;
	}
}
#endif

void NPRV_emul_TX_tot(int size_in) {
	static unsigned char total_TX_data[70000] = { 0 };
	int lpos = 0;
	int i, chan;
	int size_loc = 1;
	total_TX_data[0] = 0xF5;
	total_TX_data[1] = 0xF6;
	lpos = 2;
	//for (chan = 0; chan < CONF_NPRV_channel_count; chan++) {
	for (chan = 0; chan < emul_chan_nb; chan++) {
		if (NPRV_channels_SR_std[chan] == 0) { // 100kS
			size_loc = size_in;
		}
		if (NPRV_channels_SR_std[chan] == 1) { //50kS
			size_loc = size_in / 2;
		}
		memcpy(&total_TX_data[lpos], TX_emul_symb[chan], size_in);
		lpos = lpos + size_in;
	}
	socket_loc.send(total_TX_data, lpos);
}

int NPRV_emul_RX_tot(int size_in) {
	static unsigned char total_RX_data[70000] = { 0 };
	int lpos = 0;
	//int size_loc = 1;
	int i, chan;
	std::size_t received;
	lpos = 0;
	//printf("trying to RX TCP\r\n");
	//printf("before TCP received:%i datagram_size:%i\r\n", lpos, datagram_size);
	while (lpos < datagram_size) {
		socket_loc.receive(&total_RX_data[lpos], datagram_size, received);
		lpos = lpos + received;
		//printf("received temp : %i\r\n", received);
	}
	//printf("after TCP received:%i datagram_size:%i\r\n", lpos, datagram_size);
	if ((total_RX_data[0] != 0xF5) || (total_RX_data[1] != 0xF6)) {
		printf("TCP RX ERROR WRONG KEYWORD\r\n");
		return 0;
	}
	lpos = 2;
	for (chan = 0; chan < emul_chan_nb; chan++) {
		/*
		if (NPRV_channels_SR_std[chan] == 0) { // 100kS
			size_loc = size_in;
		}
		if (NPRV_channels_SR_std[chan] == 1) { //50kS
			size_loc = size_in / 2;
		}
		*/
		memcpy(RX_emul_symb[chan], &total_RX_data[lpos], size_in);
		lpos = lpos + size_in;
	}
	return 1;
}

unsigned char NPRV_emul_TX_raw2TCP_symbol(unsigned char raw_symb, float ampl_float) {
	unsigned char symb_out;
	unsigned char raw_symb_bis;
	unsigned int ampl_out;
	ampl_out = (unsigned int)(ampl_float);
	ampl_out = ampl_out / 350;
	raw_symb_bis = raw_symb + ((raw_symb & 0x01) << 2);
	symb_out = raw_symb_bis + ((ampl_out & 0x1F) << 3);
	return symb_out;
}


unsigned char NPRV_emul_RX_TCP2raw_symbol(unsigned char raw_symb, float* ampl_float) {
	unsigned char real_symb, symb_out, ampl_raw;
	real_symb = raw_symb & 0x7;
	ampl_raw = (raw_symb & 0xF8) >> 3;
	(*ampl_float) = ampl_raw * 350;
	//if (real_symb < 0x4) { symb_out = real_symb; }
	//else { symb_out = 0x7; }
	symb_out = real_symb;
	return symb_out;
}

void NPRV_emul_demod_tot(int size_in) {
	int chan, i;
	uint32_t loc_timer;
	int symb_size = 1;
	unsigned char raw_symb, real_symb;
	float ampl_loc;
	for (chan = 0; chan < CONF_NPRV_channel_count; chan++) {
		if (NPRV_channels_SR_std[chan] == 0) { // 100kS
			symb_size = 2;
		}
		if (NPRV_channels_SR_std[chan] == 1) { //50kS
			symb_size = 4;
		}
		if (NPRV_channels_SR_std[chan] == 2) { //200kS
			symb_size = 1;
		}
		for (i = 0; i < size_in; i++) {
			loc_timer = NPRV_emul_rx_macro_timer + (5 * i);// 5micro sec
			raw_symb = RX_emul_symb[chan][i];
			if ((i%symb_size) == 0) {
				real_symb = NPRV_emul_RX_TCP2raw_symbol(raw_symb, &ampl_loc);
				NPRV_rx_symb_decode(real_symb, loc_timer, ampl_loc, 0, chan);
			}
		}
	}
	NPRV_emul_rx_macro_timer = NPRV_emul_rx_macro_timer + (5 * size_in);//
}

void NPRV_emul_mod_tot(int size_loc) {
	int chan, i;
	unsigned char raw_symb, symb_out;
	//float loc_ampl_var;
	for (i = 0; i < size_loc; i++) {
		for (chan = 0; chan < CONF_NPRV_channel_count; chan++) { 
			if ((NPRV_channels_SR_std[chan] == 0) && ((i % 2) == 0)) { // 100kS
				tx_emul_ampl[chan] = tx_emul_ampl[chan] + (10 * tx_emul_ampl_var[chan]);
				raw_symb = NPRV_mod_get_new_symbol(&tx_emul_ampl[chan], &tx_emul_ampl_var[chan], chan);
				symb_out = NPRV_emul_TX_raw2TCP_symbol(raw_symb, tx_emul_ampl[chan]);
				TX_emul_symb[chan][i] = symb_out;
				TX_emul_symb[chan][i+1] = symb_out;
			}
			if ((NPRV_channels_SR_std[chan] == 1) && ((i % 4) == 0)) { // 50kS
				tx_emul_ampl[chan] = tx_emul_ampl[chan] + (20 * tx_emul_ampl_var[chan]); //20
				raw_symb = NPRV_mod_get_new_symbol(&tx_emul_ampl[chan], &tx_emul_ampl_var[chan], chan);
				symb_out = NPRV_emul_TX_raw2TCP_symbol(raw_symb, tx_emul_ampl[chan]);
				TX_emul_symb[chan][i] = symb_out;
				TX_emul_symb[chan][i + 1] = symb_out;
				TX_emul_symb[chan][i + 2] = symb_out;
				TX_emul_symb[chan][i + 3] = symb_out;
				//printf("i/2:%i\r\n", i / 2);
			}
			if (NPRV_channels_SR_std[chan] == 2) {				// 200kS
				tx_emul_ampl[chan] = tx_emul_ampl[chan] + (5 * tx_emul_ampl_var[chan]);
				raw_symb = NPRV_mod_get_new_symbol(&tx_emul_ampl[chan], &tx_emul_ampl_var[chan], chan);
				symb_out = NPRV_emul_TX_raw2TCP_symbol(raw_symb, tx_emul_ampl[chan]);
				TX_emul_symb[chan][i] = symb_out;
			}
		}
		NPRV_G_tx_timer = NPRV_G_tx_timer + 5;//+10
	}
}

static int NPRV_emul_do_not_overwrite_txt = 0;

void NPRV_emul_HMI_new_TXT_message(void) {
	NPRV_emul_do_not_overwrite_txt = 1;
}

void NPRV_sat_emul_glob_loop(void) {
	int size_loc;
	int i = -1;
	int j;
	sf::Time elapsed_time;
	int current_time, previous_time;
	int delta_t_mean, delta_t_count = 0, delta_t_tot = 0;
	int delta_t_loc, delta_t_max = 0;
	float err_frame_BPSK, err_frame_QPSK;
	uint32_t Tstart2;
	int count_txt_lines = 0;
	int count_txt_lines_prev = 0;
	int k = 0;
	static int count_frame_prev[30];
	float BER_BPSK, BER_QPSK;
	int loc_error;
	//static char status_text[5][20] = { "NOT-CONNECTED", " ", "CONNECTED        " };
	size_loc = 6000;//32768 16384 3000

	datagram_size = 2 + emul_chan_nb * size_loc;
	printf("emul datagram size: %i chan_nb: %i\r\n", datagram_size, CONF_NPRV_channel_count);

	
	NPRV_coder_init();
	NPRV_IPv4_init_TX_struct();

	if ((CONF_TX_test == 1) || (CONF_TX_test == 3)) {
		for (i = 0; i < CONF_NPRV_channel_count; i++) {
			NPRV_tx_plan_new_burst(i, 800000 + i * 6000, 50000, 0);
		}
	}
	else if (CONF_TX_test == 2) {
		//NPRV_TX_test_periodic_call();
	}

#ifdef NPRV_linux 
	struct sigaction new_action, old_action;
	new_action.sa_handler = termination_handler_emu;
	sigemptyset(&new_action.sa_mask);
	new_action.sa_flags = 0;

	sigaction(SIGINT, NULL, &old_action);
	if (old_action.sa_handler != SIG_IGN)
		sigaction(SIGINT, &new_action, NULL);
	sigaction(SIGHUP, NULL, &old_action);
	if (old_action.sa_handler != SIG_IGN)
		sigaction(SIGHUP, &new_action, NULL);
	sigaction(SIGTERM, NULL, &old_action);
	if (old_action.sa_handler != SIG_IGN)
		sigaction(SIGTERM, &new_action, NULL);
#endif /*linux*/

#ifdef NPRV_windows
	SetConsoleCtrlHandler(CtrlHandler_emu, TRUE); //catch events

	// terminal mode windows
	loc_error = 0;
	HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
	if (hOut == INVALID_HANDLE_VALUE) { loc_error = 1; }
	DWORD dwMode = 0;
	if (loc_error == 0) {
		if (!GetConsoleMode(hOut, &dwMode)) { loc_error = 1; }
	}
	dwMode |= ENABLE_VIRTUAL_TERMINAL_PROCESSING;
	if (loc_error == 0) {
		if (!SetConsoleMode(hOut, dwMode)) { loc_error = 1; }
	}
#endif /*windows*/

	//open TCP connection
	sf::IpAddress loc_IP_address(CONF_Emul_Radio_IP);
	sf::Socket::Status status = socket_loc.connect(loc_IP_address, 9435);
	if (status != sf::Socket::Done) {
		printf("ERR UNABLE TO OPEN TCP connection\r\n");
	}
	else {
		printf("TCP opened!\r\n");
	}

	current_time = elapsed_time.asMilliseconds();
	previous_time = current_time;

	NPRV_emul_TX_tot(size_loc);
	
	while (Kill_requested_emu == 0) {//big loop
		i++;
		Eth_RX_dequeue();
		
		//mod
		NPRV_emul_mod_tot(size_loc);
		//TX
		NPRV_emul_TX_tot(size_loc);
		//RX
		NPRV_emul_RX_tot(size_loc);
		//demod
		NPRV_emul_demod_tot(size_loc);

		elapsed_time = G_clock.getElapsedTime();
		current_time = elapsed_time.asMilliseconds();//!!!
		delta_t_loc = current_time - previous_time;
		
		if (delta_t_loc > delta_t_max) { delta_t_max = delta_t_loc; }
		previous_time = current_time;

		delta_t_count++;
		delta_t_tot = delta_t_tot + delta_t_loc;

		//sf::sleep(sf::milliseconds(35));

		if ((i % 30) == 0) {
			NPRV_check_config_update(1);
			if (count_txt_lines != count_txt_lines_prev) { NPRV_emul_do_not_overwrite_txt = 1; }
			count_txt_lines_prev = count_txt_lines;
			if ((i > 10) && (NPRV_emul_do_not_overwrite_txt == 0)) {
				for (k = 0; k < count_txt_lines; k++) { printf("\x1b[A"); }
			}
			count_txt_lines = 4;
			if (CONF_TX_test != 0) {
				//count_txt_lines = count_txt_lines + 2;
			}
			printf("            \r\ndelta T moy:%f maxi:%i\r\n", (float)delta_t_tot / (float)delta_t_count, delta_t_max);
			delta_t_count = 0;
			delta_t_tot = 0;
			delta_t_max = 0;
			NPRV_get_BER_stat(&BER_BPSK, &BER_QPSK);
			//printf("BER - BPSK:%f QPSK:%f  \r\n", BER_BPSK, BER_QPSK);
			//NPRV_FEC_stat_display();
			NPRV_get_FrameErr_stat(&err_frame_BPSK, &err_frame_QPSK);
			//printf("FRAME ERR BPSK:%.3f QPSK:%.3f\r\n", err_frame_BPSK, err_frame_QPSK);
			//printf("TA : %u T:%u\r\n", G_NPRV_my_TimeAdvance, NPRV_G_tx_timer);

			for (j = 0; j < CONF_NPRV_channel_count; j++) {
				printf("RX chan %i %i : %i %i  \r\n", j, i, G_NPRV_RX_radio_frame_count[j], G_NPRV_RX_radio_frame_count[j] - count_frame_prev[j]);
				count_frame_prev[j] = G_NPRV_RX_radio_frame_count[j];
				count_txt_lines++;
			}
			count_txt_lines = count_txt_lines + NPRV_HMI_print_radio_stats();
			//NPRV_L1L2_print_TX_frames();
			//NPRV_L1L2_print_RX_frame_count();
			printf("\r\n");
			//printf("STATUS: %s\r\n", status_text[G_NPRV_MyRadioState]);
			NPRV_signaling_print_connection_status();
			count_txt_lines = count_txt_lines + DHCP_ARP_print_entries();
			DHCP_ARP_periodic_free_table();
			if (CONF_TX_test != 0) {
				//NPRV_print_pseudorand_err();
			}
			count_txt_lines = count_txt_lines + NPRV_HMI_print_who();
			NPRV_signaling_periodic_call();
			NPRV_emul_do_not_overwrite_txt = 0;
		}
	}
}