// This file is part of "NPR-VSAT SDR Modem" software
// (New-Packet-Radio, via VSAT, software modem for PC with duplex SDR)
// (A virtual/SDR data IPv4 modem for amateur radio geostationary satellite, 
// Mostly designed for Hamnet connectivity over QO-100)
// Copyright (c) 2020-2023 Guillaume F. F4HDK (amateur radio callsign)
// contact e-mail : f4hdk [at] free [dot] fr
// 
// "NPR-VSAT Software Modem" is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// "NPR-VSAT Software Modem" is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with "NPR-VSAT Software Modem".  If not, see <http://www.gnu.org/licenses/>

#include "pch.h"
#include "NPRV_glob_variables.h"
#include "NPRV_signaling.h"
#include "NPRV_L1_L2_radio.h"
#include "Eth_IPv4.h"
#include "NPRV_SDR_if.h"
#include "NPRV_sat_emul.h"
#include "NPRV_mod_demod.h"
#include "NPRV_coder_decoder.h"
#include "NPRV_log_file.h"

#ifdef NPRV_windows
#include "Win_tuntap_if.h"
#endif

#ifdef NPRV_linux
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "Linux_tuntap_if.h"
#endif

#pragma warning(disable: 4996)

#define NPRV_SIGNAL_PERIOD_RFCHAN	12400 /*milliseconds*/
#define NPRV_SIGNAL_PERIOD_TIMING	15200 /*milliseconds*/
#define NPRV_SIGNAL_PERIOD_IPGEN	14100 /*milliseconds*/
#define NPRV_SIGNAL_PERIOD_IP_ROUT	11000 /*milliseconds*/
#define NPRV_SIGNAL_PERIOD_IPRANGES	 9700 /*milliseconds*/
#define NPRV_SIGNAL_PERIOD_WHO		24300 /*milliseconds*/
#define NPRV_SIGNAL_PERIOD_CONREQ	21700 /*milliseconds 21700*/

#define NPRV_SIGNAL_CLIENT_TIMEOUT	50		/*seconds*/
#define NPRV_SIGNAL_MY_TIMEOUT		40		/*seconds*/



uint32_t T_last_client_seen[NPRV_MAX_CLIENT_NB] = { 0 };
uint32_t T_my_client_last_ConAck = 0;

int NPRV_sig_temp_ConAck_TXneed[NPRV_MAX_CLIENT_NB] = { 0 };
unsigned char NPRV_sig_temp_ConReqNACK_answers[20][25] = { 0 };

int NPRV_IPRout_config_valid = 0;
int NPRV_Timing_config_valid = 0;


//Who TX
static uint32_t T_last_Who_signal_TX_complete = 0;
int NPRV_signal_TX_Who_data_avail(uint32_t current_time) {
	uint32_t delta_T;
	delta_T = current_time - T_last_Who_signal_TX_complete;
	//printf("delta_t signaling ")
	
	if ((delta_T > NPRV_SIGNAL_PERIOD_WHO) && (G_NPRV_MyClientID < NPRV_MAX_CLIENT_NB)) {
		//printf("data Who available\r\n");
		return 1;
	}
	else {
		return 0;
	}
}

int NPRV_get_signal_TX_Who(unsigned char* content) {
	uint32_t current_time, delta_T;
	sf::Time elapsed_time;
	unsigned char IP_char_loc[6];
	int size = 23;
	content[0] = 0x01;//
	content[1] = size;
	content[2] = (G_NPRV_MyClientID & 0x1F00) >> 8;	//client ID MSB
	content[3] = (G_NPRV_MyClientID & 0x00FF);		//client ID LSB
	memcpy((char*)(content + 4), G_NPRV_my_Callsign, 16);//callsign
	content[20] = NPRV_my_IP_static;
	IP_int2char(LAN_conf_applied.DHCP_range_start, IP_char_loc);
	content[21] = IP_char_loc[2];//IP Start 2nd LSB
	content[22] = IP_char_loc[3];//IP Start LSB
	IP_int2char(LAN_conf_applied.DHCP_range_size, IP_char_loc);
	content[23] = IP_char_loc[2];//IP Start 2nd LSB
	content[24] = IP_char_loc[3];//IP Start LSB
	elapsed_time = G_clock.getElapsedTime();
	current_time = elapsed_time.asMilliseconds();
	T_last_Who_signal_TX_complete = current_time;
	return size + 2;
}

static uint32_t T_last_ConReq_signal_TX_complete = 0;

void NPRV_signal_prioritize_TX_ConReq(void) {
	T_last_ConReq_signal_TX_complete = T_last_ConReq_signal_TX_complete - NPRV_SIGNAL_PERIOD_CONREQ;
}

int NPRV_signal_TX_ConReq_data_avail(uint32_t current_time) {
	uint32_t delta_T;
	int rand_loc;
	delta_T = current_time - T_last_ConReq_signal_TX_complete;
	//srand(time(NULL));
	//srand(current_time);
	rand_loc = rand() % 1000;
	//printf("delta_t signaling %u RAND:%i\r\n", delta_T, rand_loc);
	if ((CONF_is_NPRV_HUB == 0) && 
		(delta_T > NPRV_SIGNAL_PERIOD_CONREQ) &&
		(NPRV_IPgen_config_valid == 1) && 
		(NPRV_chan_config_valid == 1) &&
		(NPRV_IPRout_config_valid == 1) &&
		(NPRV_Timing_config_valid == 1) &&
		(rand_loc > 500)) 
	{ // affiner en fonction du status
		//printf("data CON REQ available\r\n");
		return 1;
	}
	else {
		return 0;
	}
}

int NPRV_get_signal_TX_ConReq(unsigned char* content) {
	uint32_t current_time, delta_T;
	sf::Time elapsed_time;
	unsigned char IP_char_loc[6];
	int size = 25;
	//printf("SIGNALIN: Con Req TX\r\n");
	content[0] = 0x05;
	content[1] = size;
	memcpy(content + 2, G_NPRV_my_random_nb, 4);
	memcpy((char*)(content + 6), G_NPRV_my_Callsign, 16);//callsign
	content[22] = CONF_my_IP_static_requested & 0xFF;
	content[23] = (CONF_my_IP_size_requested & 0xFF00) >> 8;
	content[24] = (CONF_my_IP_size_requested & 0x00FF);
	content[25] = CONF_my_RFGlob_TX_capab & 0xFF;
	content[26] = CONF_my_radio_RX_capab & 0xFF;
	elapsed_time = G_clock.getElapsedTime();
	current_time = elapsed_time.asMilliseconds();
	T_last_ConReq_signal_TX_complete = current_time;
	return size + 2;
}



// synthesis TX functions
int NPRV_signaling_TX_data_available(void) {
	//synthesis for all types of messages
	uint32_t current_time;
	sf::Time elapsed_time;
	elapsed_time = G_clock.getElapsedTime();
	current_time = elapsed_time.asMilliseconds();
	int data_avail_synth = 0;
	if (CONF_is_NPRV_HUB == 0) {
		if (NPRV_signal_TX_Who_data_avail(current_time) == 1) { data_avail_synth = 1; }
		if (NPRV_signal_TX_ConReq_data_avail(current_time) == 1) { data_avail_synth = 1; }
	}
	return data_avail_synth;
}

int NPRV_signaling_TX_get_frame(unsigned char* tx_frame, int avail_size) {
	uint32_t current_time;
	sf::Time elapsed_time;
	int pos, size;
	int message_size = 0;
	int size_left_avail;
	tx_frame[0] = 0x1F;//dest ID
	tx_frame[1] = 0xFF;
	tx_frame[2] = 0xD0;//protocol signaling
	pos = 3;
	avail_size = avail_size - 4; //for checksum plus FF FF end signaling message
	elapsed_time = G_clock.getElapsedTime();
	current_time = elapsed_time.asMilliseconds();
	if ((CONF_is_NPRV_HUB == 0) && (NPRV_signal_TX_ConReq_data_avail(current_time) == 1) && ((pos + 27) < avail_size)) {
		message_size = NPRV_get_signal_TX_ConReq(tx_frame + pos);
		pos = pos + message_size;
		if (pos > avail_size) {
			printf("** ERR signaling CON REQ too big ** \r\n\r\n");
			NPRV_HMI_new_TXT_message();
			NPRV_emul_HMI_new_TXT_message();
		}
	}

	if ((NPRV_signal_TX_Who_data_avail(current_time) == 1) && ((pos + 25) < avail_size)) {
		message_size = NPRV_get_signal_TX_Who(tx_frame + pos);
		pos = pos + message_size;
		if (pos > avail_size) {
			printf("** ERR signaling WHO too big ** \r\n\r\n");
			NPRV_HMI_new_TXT_message();
			NPRV_emul_HMI_new_TXT_message();
		}
	}

	tx_frame[pos] = 0xFF;
	tx_frame[pos+1] = 0xFF;
	size = pos + 2;
	size = size + (size % 2);
	NPRV_code_checksum(tx_frame, size);
	if (size > 100) {
		//printf("!!!!!size enormous SIGNALING!!!! %i\r\n", size);
		//return 30;
	}
	size = size + 2;//chksum size

	return size;
}



//RX part

void NPRV_signaling_RX_IPRoutingTabl(unsigned char* message) {
	int size, i, matching;
	static int matching_old = 0;
	static unsigned char routing_data_prev[100] = { 0 };
	size = message[1];
	if (size > 50) { size = 50; }
	matching = 1;
	for (i = 0; i <= size; i++) {
		if (message[i + 1] != routing_data_prev[i]) { matching = 0; }
	}
	if ((matching == 1) && (matching_old == 0)) {
		//printf("CONF IP routing OK\r\n");
		for (i = 0; i <= size; i++) {
			CONF_DHCP_route_table[i] = message[i + 1];
			//printf("%02X:", CONF_DHCP_route_table[i]);			
		}
		LAN_conf_applied.LAN_def_route = IP_char2int(CONF_DHCP_route_table + 1);
		NPRV_IPRout_config_valid = 1;
		//printf("\r\n");
		//NPRV_HMI_new_TXT_message();
		//NPRV_emul_HMI_new_TXT_message();
	}
	else { 
		//printf("CONF IP rouging 1st\r\n"); 
	}
	matching_old = matching;
	for (i = 0; i <= size; i++) {
		routing_data_prev[i] = message[i + 1];
	}
}

void NPRV_signaling_RX_IPGenConf(unsigned char* message) {
	static uint32_t IP_modem_prev = 0xFFFFFFFF;
	static uint32_t IP_subnet_prev = 0xFFFFFFFF;
	static uint32_t IP_DNS_prev = 0xFFFFFFFF;
	static int DNS_active_prev = 9;
	uint32_t IP_modem_rx = 0xFFFFFFFF;
	uint32_t IP_subnet_rx = 0xFFFFFFFF;
	uint32_t IP_DNS_rx = 0xFFFFFFFF;
	int DNS_active_rx;
	int valid_DNS_loc = 0;
	int valid_subnet_loc = 0;
	int matching_loc;
	static int matching_old = 0;
	IP_modem_rx = IP_char2int(message + 0);
	IP_subnet_rx = IP_char2int(message + 4);
	DNS_active_rx = message[8];
	IP_DNS_rx = IP_char2int(message + 9);

	if ((DNS_active_rx == 1) || (DNS_active_rx == 0)) {
		valid_DNS_loc = 1;
	}
	if ((IP_subnet_rx & 0xFFFF0000) == 0xFFFF0000) {// subnet smaller than /16
		valid_subnet_loc = 1;
	}
	if ((valid_DNS_loc == 1) && (valid_subnet_loc == 1)) {
		//printf("valid loc IP config:\r\n");
		if ((IP_modem_rx == IP_modem_prev) && (IP_subnet_rx == IP_subnet_prev) && (DNS_active_rx == DNS_active_prev) && (IP_DNS_rx == IP_DNS_prev)) {
			matching_loc = 1;
		}
		else {
			matching_loc = 0;
			// modif planned: Ethernet disconnect !!!
		}
		if ((matching_loc == 1) && (matching_old == 0)) {
			LAN_conf_applied.LAN_modem_IP = IP_modem_rx;
			LAN_conf_applied.LAN_subnet_mask = IP_subnet_rx;
			//LAN_conf_applied.LAN_DNS_activ = DNS_active_rx;
			NPRV_DNS_activ_from_Hub = DNS_active_rx;
			LAN_conf_applied.LAN_DNS_value = IP_DNS_rx; 
			NPRV_IPgen_config_valid = 1;
			//printf("\r\n\r\n ******  CONFIG IPv4 OK *******\r\n\r\n");
			//NPRV_HMI_new_TXT_message();
			//NPRV_emul_HMI_new_TXT_message();
		}
		matching_old = matching_loc;
		IP_modem_prev = IP_modem_rx;
		IP_subnet_prev = IP_subnet_rx;
		DNS_active_prev = DNS_active_rx;
		IP_DNS_prev = IP_DNS_rx;
	}
}

void NPRV_signaling_RX_TimingConf(unsigned char* message) {
	static uint32_t slot_duration_prev = 0;
	static uint32_t slot_T_margin_prev = 0;
	static uint32_t TDMA_Frame_slot_nb_prev = 0;
	static uint32_t T_offset_channel_prev = 0;
	uint32_t slot_duration_loc = 0;
	uint32_t slot_T_margin_loc = 0;
	uint32_t TDMA_Frame_slot_nb_loc = 0;
	uint32_t T_offset_channel_loc = 0;
	int all_matching;
	static int all_matching_prev = 0;
	slot_duration_loc =
		(message[0] << 16) +
		(message[1] << 8) +
		message[2];
	slot_T_margin_loc =
		(message[3] << 8) +
		message[4];
	TDMA_Frame_slot_nb_loc = message[5];
	T_offset_channel_loc =
		(message[6] << 8) +
		message[7];
	all_matching = 1;
	if (slot_duration_loc != slot_duration_prev) { all_matching = 0; }
	if (slot_T_margin_loc != slot_T_margin_prev) { all_matching = 0; }
	if (TDMA_Frame_slot_nb_loc != TDMA_Frame_slot_nb_prev) { all_matching = 0; }
	if (T_offset_channel_loc!= T_offset_channel_prev) { all_matching = 0; }
	if ((all_matching == 1) && (all_matching_prev == 0)) {
		//printf("SIGNALING valid timing conf received; Tslot:%u Margin:%u slot_nb:%u T_offset:%u *****************************\r\n", slot_duration_loc, slot_T_margin_loc, TDMA_Frame_slot_nb_loc, T_offset_channel_loc);
		CONF_slot_duration = slot_duration_loc;
		CONF_slot_T_margin = slot_T_margin_loc;
		CONF_TDMA_Frame_slot_nb = TDMA_Frame_slot_nb_loc;
		CONF_T_offset_channel = T_offset_channel_loc;
		NPRV_Timing_config_valid = 1;
		//printf("\r\n\r\n ****** TIMING CONFIG OK *******\r\n\r\n");
		//NPRV_HMI_new_TXT_message();
		//NPRV_emul_HMI_new_TXT_message();
	}
	all_matching_prev = all_matching;
	slot_duration_prev = slot_duration_loc;
	slot_T_margin_prev = slot_T_margin_loc;
	TDMA_Frame_slot_nb_prev = TDMA_Frame_slot_nb_loc;
	T_offset_channel_prev = T_offset_channel_loc;


}

void NPRV_signaling_RX_RFChan(unsigned char* message) {
	static int channels_SR_std_prev[NPRV_MAX_CHANNEL_COUNT] = { 9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9 };
	static float channels_freq_prev[NPRV_MAX_CHANNEL_COUNT] = { 99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99,99, };
	static int tot_chan_prev = 99;
	int valid_chan_nb = 0;
	int valid_chan_config_freq = 0;
	int valid_chan_config_SR = 0;
	static int valid_chan_config[NPRV_MAX_CHANNEL_COUNT] = { 0 };
	int i;
	int all_valid_received;
	static int all_valid_received_prev = 0;
	unsigned char channel_rx;
	int Symbol_rate_rx, tot_chan_rx; 
	int16_t chan_freq_int16;
	float chan_freq_float_rx;
	channel_rx = message[0];
	Symbol_rate_rx = message[1];
	chan_freq_int16 = (message[2] << 8) + message[3];
	chan_freq_float_rx = ((float)chan_freq_int16) / 1000;
	tot_chan_rx = message[4];
	//printf("received channel %i SR %i freq %f tot_chan %i\r\n", channel_rx, Symbol_rate_rx, chan_freq_float_rx, tot_chan_rx);
	if (tot_chan_rx <= NPRV_MAX_CHANNEL_COUNT) {
		if (tot_chan_rx == tot_chan_prev) {
			CONF_NPRV_channel_count = tot_chan_rx;
			valid_chan_nb = 1;
			//printf("valid chan nb: %i\r\n", CONF_NPRV_channel_count); 
			//NPRV_HMI_new_TXT_message();
			//NPRV_emul_HMI_new_TXT_message();
		}
		tot_chan_prev = tot_chan_rx;	
	}
	if (channel_rx < CONF_NPRV_channel_count) {
		//symbol rate
		if ((Symbol_rate_rx >= 0) && (Symbol_rate_rx <= 2) && (Symbol_rate_rx == channels_SR_std_prev[channel_rx])) {
			valid_chan_config_SR = 1;
			//printf("valid SR: %i\r\n", NPRV_channels_SR_std[channel_rx]);
			//NPRV_HMI_new_TXT_message();
			//NPRV_emul_HMI_new_TXT_message();
		}
		channels_SR_std_prev[channel_rx] = Symbol_rate_rx;
		//frequency
		if ((chan_freq_float_rx >= -0.5) && (chan_freq_float_rx <= 0.5) && (chan_freq_float_rx == channels_freq_prev[channel_rx])) {
			valid_chan_config_freq = 1;
			//printf("valid freq: %f\r\n", channels_freq_prev[channel_rx]);
			//NPRV_HMI_new_TXT_message();
			//NPRV_emul_HMI_new_TXT_message();
		}
		channels_freq_prev[channel_rx] = chan_freq_float_rx;
		if ((valid_chan_config_SR == 1) && (valid_chan_config_freq == 1)) {
			NPRV_channels_SR_std[channel_rx] = Symbol_rate_rx;
			NPRV_channels_freq[channel_rx] = chan_freq_float_rx;
			valid_chan_config[channel_rx] = 1;
			//printf("\r\n auto set chan %i ; f:%f SR:%i\r\n\r\n", channel_rx, chan_freq_float_rx, Symbol_rate_rx);
			//NPRV_HMI_new_TXT_message();
			//NPRV_emul_HMI_new_TXT_message();
		}
		else {
			valid_chan_config[channel_rx] = 0;
		}
		all_valid_received = 1;
		if (valid_chan_nb == 0) { all_valid_received = 0; }
		for (i = 0; i < CONF_NPRV_channel_count; i++) {
			if (valid_chan_config[i] == 0) { 
				all_valid_received = 0; 
				//printf("FREQ_config_modif\r\n");
				//NPRV_HMI_new_TXT_message();
				//NPRV_emul_HMI_new_TXT_message();
			}
		}
		if ((all_valid_received == 1) && (all_valid_received_prev == 0)) {
			NPRV_chan_config_valid = 1;
			NPRV_demod_set_channels();
			NPRV_mod_init();
			NPRV_coder_init();
			if (CONF_Emul_Radio == 0) {
				NPRV_SDR_re_configure();
			}
			//printf("\r\n!! full FREQ config received\r\n\r\n");
			//NPRV_HMI_new_TXT_message();
			//NPRV_emul_HMI_new_TXT_message();
		}
		all_valid_received_prev = all_valid_received;
	}
}



void NPRV_signaling_RX_IPRanges(unsigned char* message) {
	uint32_t current_time;
	sf::Time elapsed_time;
	elapsed_time = G_clock.getElapsedTime();
	current_time = elapsed_time.asSeconds();
	int pos = 2;
	int size;
	int loc_client_ID;
	uint32_t IP_MSBytes, IP_loc, IP_LSBytes;
	size = message[1];
	IP_MSBytes = LAN_conf_applied.LAN_modem_IP & 0xFFFF0000;
	for (pos = 2; pos < size; pos = pos + 6) {
		loc_client_ID = ((message[pos] & 0x1F) << 8) +
			(message[pos + 1] & 0xFF);
		//if ((loc_client_ID < NPRV_MAX_CLIENT_NB) && (NPRV_IPgen_config_valid == 1) && (loc_client_ID != G_NPRV_MyClientID)) {
		if ((loc_client_ID < NPRV_MAX_CLIENT_NB) && (NPRV_IPgen_config_valid == 1)) {
			//printf("valid IPRange: %i ", loc_client_ID);
			IP_LSBytes = IP_char2int(message + pos);//pos : +2 and -2because 2 LSB
			IP_LSBytes = IP_LSBytes & 0x0000FFFF;
			IP_loc = IP_MSBytes + IP_LSBytes;
			CONF_radio_addr_table_IP_begin[loc_client_ID] = IP_loc;
			//printf("IP begin:%08X", CONF_radio_addr_table_IP_begin[loc_client_ID]);
			IP_LSBytes = IP_char2int(message + pos + 2);//pos : +4 and -2because 2 LSB
			IP_loc = IP_LSBytes & 0x0000FFFF;
			CONF_radio_addr_table_IP_size[loc_client_ID] = IP_loc;
			//printf("IP size:%08X\r\n", CONF_radio_addr_table_IP_size[loc_client_ID]);
			T_last_client_seen[loc_client_ID] = current_time;
			CONF_radio_addr_table_status[loc_client_ID] = 1;
		}
	}
}

void NPRV_signaling_RX_who(unsigned char* message, unsigned int TX_ID) {
	unsigned char IP_char_loc[6] = { 0 };
	unsigned int loc_client_ID;
	uint32_t IP_32_loc;
	uint32_t IP_size_loc;
	char loc_callsign[18];
	int loc_static_IP;

	
	loc_client_ID = ((message[0] & 0x1F) << 8) +
		(message[1] & 0xFF);
	memcpy(loc_callsign,(char*)(message + 2), 16);
	loc_callsign[16] = 0;
	//printf("ID:%u callsign:%s\r\n", loc_client_ID, loc_callsign);
	loc_static_IP = message[18];
	IP_char_loc[2] = message[19];
	IP_char_loc[3] = message[20];
	IP_32_loc = IP_char2int(IP_char_loc);
	IP_32_loc = (LAN_conf_applied.LAN_modem_IP & 0xFFFF0000) + IP_32_loc;
	//printf("IP start:%08X\r\n", IP_32_loc);
	IP_char_loc[2] = message[21];
	IP_char_loc[3] = message[22];
	IP_size_loc = IP_char2int(IP_char_loc);
	//printf("IP size:%u\r\n", IP_size_loc);
	if ((loc_client_ID == TX_ID) 
		&& (loc_client_ID < NPRV_MAX_CLIENT_NB))
	//	&& (loc_client_ID != G_NPRV_MyClientID)) 
	{// && (loc_client_ID != G_NPRV_MyClientID)) {
		strcpy(RX_who_callsigns[loc_client_ID], loc_callsign);
		RX_who_static_IP[loc_client_ID] = loc_static_IP;
		RX_who_IP_range_start[loc_client_ID] = IP_32_loc;
		RX_who_IP_range_size[loc_client_ID] = IP_size_loc;
	}
}

int compare_callsigns(char* callsign_a, unsigned char* rand_a, char* callsign_b, unsigned char* rand_b) {
	int i, answer;
	answer = 1;
	for (i = 0; i < 4; i++) {
		if (rand_a[i] != rand_b[i]) { answer = 0; }
	}
	if (strcmp(callsign_a, callsign_b) != 0) { answer = 0 ; }
	return answer;
}


void NPRV_signaling_RX_ConACK(unsigned char* message) {
	unsigned char rand_RX[6] = { 0 };
	char callsign_RX[20] = { 0 };
	unsigned char temp_IP_char[6] = { 0 };
	uint32_t temp_IP_uint;
	int temp_client_ID = 0;
	int i;
	uint32_t current_time, delta_T;
	sf::Time elapsed_time;
	
	memcpy(rand_RX, message, 4);
	memcpy(callsign_RX, (char*)(message + 4), 16);
	callsign_RX[16] = 0;
	if (compare_callsigns(callsign_RX, rand_RX, G_NPRV_my_Callsign, G_NPRV_my_random_nb) == 1) {
		//printf("SIGNALING : ACK! %s\r\n", callsign_RX);
		//printf("Ack match me\r\n");
		for (i = 0; i < 27; i++) {
			//printf("%02X:", message[i]);
		}
		//printf("\r\n");
		temp_client_ID = (message[20] << 8) +
			message[21];
		//truc = message[20]; //static IP !!! not yet managed
		if ((temp_client_ID > 0) && (temp_client_ID < NPRV_MAX_CLIENT_NB)) {
			if (G_NPRV_MyClientID = 0x1FFE) {//previous state unconnected
				//!!! a affiner
				if (CONF_INHIB_TUNTAP == 0) {
					connect_tuntap();
				}
			}
			elapsed_time = G_clock.getElapsedTime();
			current_time = elapsed_time.asSeconds();
			//printf("client IP proposed:%i\r\n", temp_client_ID);
			temp_IP_char[2] = message[23];
			temp_IP_char[3] = message[24];
			temp_IP_uint = IP_char2int(temp_IP_char);
			temp_IP_uint = temp_IP_uint + ((LAN_conf_applied.LAN_modem_IP) & 0xFFFF0000);
			LAN_conf_applied.DHCP_range_start = temp_IP_uint;
			//printf("IP start:%08X\r\n", temp_IP_uint);
			temp_IP_char[2] = message[25];
			temp_IP_char[3] = message[26];
			temp_IP_uint = IP_char2int(temp_IP_char);
			LAN_conf_applied.DHCP_range_size = temp_IP_uint;
			//printf("IP size:%08X\r\n", temp_IP_uint);
			G_NPRV_MyClientID = temp_client_ID;
			G_NPRV_MyIPConfigured = 1;
			G_NPRV_MyRadioState = 2;
			T_my_client_last_ConAck = current_time;
		}
	}
}

static unsigned char NACK_reason_stored;

void NPRV_signaling_RX_ConNACK(unsigned char* message) {
	unsigned char rand_RX[6] = { 0 };
	char callsign_RX[20] = { 0 };
	unsigned char NACK_reason = 0;
	char nack_reason_text[40];
	memcpy(rand_RX, message, 4);
	memcpy(callsign_RX, (char*)(message + 4), 16);
	callsign_RX[16] = 0;
	NACK_reason = message[20];
	//printf("SIGNALING : NACK! %s\r\n", callsign_RX);
	if (compare_callsigns(callsign_RX, rand_RX, G_NPRV_my_Callsign, G_NPRV_my_random_nb) == 1) {
		//printf("NAck match me\r\n");
		G_NPRV_MyRadioState = 4;
		if (NACK_reason < 5) { NACK_reason_stored = NACK_reason; }
		else { NACK_reason_stored = 0; }
		if (NACK_reason == 0x02) { sprintf(nack_reason_text, "NO IP address resource left..."); }
		if (NACK_reason == 0x03) { sprintf(nack_reason_text, "Too many clients connected..."); }
		if (NACK_reason == 0x04) { sprintf(nack_reason_text, "IP size requested too big..."); }
		printf("!!! CONNECTION REFUSED: %s \r\n", nack_reason_text);
		NPRV_HMI_new_TXT_message();
		NPRV_emul_HMI_new_TXT_message();
	}
	//!!! Action a prendre en compte
}


uint32_t NPRV_signaling_get_RX_RandNb(unsigned char* message) {
	uint32_t rand_nb_loc = 0;
	rand_nb_loc = 
		(message[0] << 24) + 
		(message[1] << 16) + 
		(message[2] << 8) + 
		(message[3]);
	return rand_nb_loc;
}

uint32_t NPRV_signaling_RX_frame_interpret(unsigned char* raw_frame, int size) {
	int i;
	unsigned int dest_ID, TX_ID;
	unsigned char field_type;
	unsigned char field_length;
	int data_pos;// = 11;
	uint32_t RX_rand_nb = 0;
	unsigned char* data_field;
#ifdef NPRV_WITH_HEADER_FEC_RS
	data_field = raw_frame + 12;
	data_pos = 15;
#else
	data_field = raw_frame + 8;
	data_pos = 11;
#endif
	/*
	printf("\r\nSIGNALING RX FRAME: ");
	for (i = 0; i < size; i++) {
		printf("%02X:", raw_frame[i]);
	}
	printf("\r\n");
	*/
	dest_ID = ((data_field[0] & 0x1F) << 8) + data_field[1];
	TX_ID = ((raw_frame[3] & 0x1F) << 8) + raw_frame[4];
	//printf("dest ID: %04X\r\n", dest_ID);
	do {
		field_type = raw_frame[data_pos];
		field_length = raw_frame[data_pos + 1];
		//printf("message: type %02X lenght %02X content", field_type, field_length);
		//for (i = 0; i < field_length; i++) {
		//	printf("%02X:", raw_frame[data_pos + 2 + i]);
		//}
		//printf("\r\n");
		switch (field_type) {
			case 0x01: // Who
				NPRV_signaling_RX_who(raw_frame + data_pos + 2, TX_ID);
			break;
			case 0x05: // Connection Request
				if ((CONF_is_NPRV_HUB == 0) && (G_NPRV_MyClientID == 0x1FFE) && (TX_ID == 0x1FFE)) {
					RX_rand_nb = NPRV_signaling_get_RX_RandNb(raw_frame + data_pos + 2);
				}
			case 0x06: // Connection ACK
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_ConACK(raw_frame + data_pos + 2);
				}
				break;
			case 0x07: // Connection ACK
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_ConNACK(raw_frame + data_pos + 2);
				}
				break;
			case 0x11 : // RF Channel characteristics
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_RFChan(raw_frame + data_pos + 2);
				}
				break;
			case 0x12: // Timing config
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_TimingConf(raw_frame + data_pos + 2);
				}
				break;
			case 0x21 : //IP general configuration
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_IPGenConf(raw_frame + data_pos + 2);
				}
				break;
			case 0x22: //IPv4 static routing table
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_IPRoutingTabl(raw_frame + data_pos);
				}
				break;

			case 0x24: //IP ranges of each station
				if ((CONF_is_NPRV_HUB == 0) && (TX_ID == 0) && (dest_ID == 0x1FFF)) {
					NPRV_signaling_RX_IPRanges(raw_frame + data_pos);
				}
				break;
		}
		data_pos = data_pos + field_length + 2;
	} while ((field_type != 0xFF) && (data_pos < size));
	return RX_rand_nb;
}

void NPRV_signaling_periodic_call(void) {
	int i;
	uint32_t current_time, delta_T;
	sf::Time elapsed_time;
	elapsed_time = G_clock.getElapsedTime();
	current_time = elapsed_time.asSeconds();
	if (CONF_is_NPRV_HUB == 0) {
		//current_time = elapsed_time.asMilliseconds();
		for (i = 0; i < NPRV_MAX_CLIENT_NB; i++) {
			if ((CONF_radio_addr_table_status[i] == 1) && (i != G_NPRV_MyClientID)) {
				delta_T = current_time - T_last_client_seen[i];


				if (delta_T > NPRV_SIGNAL_CLIENT_TIMEOUT) {//20 seconds
					CONF_radio_addr_table_status[i] = 0;
					//printf("client %i timed out!\r\n", i);
					//NPRV_HMI_new_TXT_message();
					//NPRV_emul_HMI_new_TXT_message();
				}
			}
		}
		//disconnection for me
		delta_T = current_time - T_my_client_last_ConAck;
		if ((delta_T > NPRV_SIGNAL_MY_TIMEOUT) && (G_NPRV_MyClientID < NPRV_MAX_CLIENT_NB)) {
			disconnect_tuntap();
			//CONF_radio_addr_table_status[G_NPRV_MyClientID] = 0;
			G_NPRV_MyClientID = 0x1FFE;
			G_NPRV_MyIPConfigured = 0;
			G_NPRV_MyRadioState = 3; //disconnected
			NPRV_IPv4_init_TX_struct();// flush IPv4 TX buffer
		}
	}
}

void NPRV_signaling_print_connection_status(void) {
	static char status_text[5][30] = { 
		"NOT-CONNECTED (yet)", 
		" ", 
		"CONNECTED        ",
		"DISCONNECTED (timeout)", 
		"CONNECTION REFUSED; "
	};
	static char additional_table[5][30] = {
		"    ",
		"    ",
		"No IP address resource left",
		"Too many clients connected",
		"IP size requested too big"
	};
	char* additional_text;
	if (G_NPRV_MyRadioState == 4) {//connection refused
		additional_text = additional_table[NACK_reason_stored];
	} else {
		additional_text = additional_table[0];
	}
	if (G_NPRV_MyRadioState <= 4) {
		printf("STATUS: %s %s   \r\n", status_text[G_NPRV_MyRadioState], additional_text);
	}
}